/************************************************************************/
/*                                                                      */
/*      DosOut -- Format search output using DOS-specific calls         */
/*                                                                      */
/*      In the interests of keeping the code portability high, the      */
/*      core files have been carefully cut to use only standard         */
/*      ANSI C calls.  However, some special effects require            */
/*      non-standard calls, such as setting character colour            */
/*      attributes.  This is a non-portable module to provide           */
/*      match highlighting output.                                      */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#define _CLIP_WIN_

#include "ascii.h"
#include <compdef.h>
#include <conio.h>
#include <dos.h>
#include "dosout.h"
#include "linefile.h"
#include "matcheng.h"
#include <stdio.h>
#include <string.h>

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
DosOut_Init(void)
{
} /*Init*/

/************************************************************************/
/*                                                                      */
/*      Display -- Report line and include details as requested         */
/*                                                                      */
/************************************************************************/
public_scope BOOL
DosOut_Display(MATCHENG_DETAILS *pDetails)
{
        /*Display extra identifying info requested by client*/
        if (pDetails->ReportingOptions & DETAILS_FILENAME) {
                fputs(pDetails->pFilename, stdout);
                putchar(':');
        }
        if (pDetails->ReportingOptions & DETAILS_LINENUMBER) {
                printf("%lu:", pDetails->LineNr);
        }
        if (pDetails->ReportingOptions & DETAILS_BYTEOFFSET) {
                printf("%lu:", LineFile_Position(pDetails->pLineStart));

        }

        /*Write line (plus trailing LF) to stdout as binary data*/
        DosOut_putsBinary(pDetails->pLineStart);

        /*Tell search engine to keep going*/
        return TRUE;

} /*Display*/

/************************************************************************/
/*                                                                      */
/*      putsBinary -- Write text but don't allow special characters     */
/*                                                                      */
/*      This function acts in the same way as puts, except that         */
/*      special characters that disrupt the output libraries are        */
/*      censored.  This is being done merely so that the function       */
/*      is not poisoned by binary files.                                */
/*                                                                      */
/*      Note that MANY parts of ggrep assume that the search is         */
/*      being performed on text files and behave differently when       */
/*      presented with binary files.  This is intentional; a true       */
/*      binary search utility is a completely different beast to        */
/*      grep.  behoffski has deliberately ignored binary files.         */
/*      A binary search utility would be extremely useful, and          */
/*      could possibly be integrated into a more sophisticated          */
/*      version of ggrep.  However, binary has implications for         */
/*      output (like the CTRL/Z case handled here), has implications    */
/*      for input (especially the absence of LF) and has implications   */
/*      for searching (especially the treatment of NUL).                */
/*      behoffski is not prepared to provide a utility that sort-of     */
/*      handles these cases: these functions need to be engineered      */
/*      properly.  Contributions aimed at providing binary searches     */
/*      would be welcomed, but the result might best fit into a         */
/*      utility with a different name than ggrep.                       */
/*                                                                      */
/************************************************************************/
public_scope void
DosOut_putsBinary(CHAR *pStr)
{
        CHAR ch;

        /*Loop through string until terminator found*/
        for (;;) {

                ch = *pStr++;

                /*Decode character and handle special cases*/
                switch (ch) {
                case NUL:
                        /*End of string found: write LF and return*/
                        putchar(LF);
                        return;
                        break;

                case ASCII_SUB:
                        /*CTRL/Z: write "^Z" so stdout isn't closed*/
                        putchar('^');
                        putchar('Z');
                        break;

                default:
                        /*Character is not special: display it as it is*/
                        putchar(ch);
                        break;

                }

        }

} /*putsBinary*/

/************************************************************************/
/*                                                                      */
/*      DisplayHighlighted -- Show line, including highlighting match   */
/*                                                                      */
/************************************************************************/
public_scope BOOL
DosOut_DisplayHighlighted(MATCHENG_DETAILS *pDetails)
{
        /*Display extra identifying info requested by client*/
        if (pDetails->ReportingOptions & DETAILS_FILENAME) {
                printf("%s:", pDetails->pFilename);
        }
        if (pDetails->ReportingOptions & DETAILS_LINENUMBER) {
                printf("%lu:", pDetails->LineNr);
        }
        if (pDetails->ReportingOptions & DETAILS_BYTEOFFSET) {
                printf("%lu:", LineFile_Position(pDetails->pLineStart));

        }

        /*Is the highlighting the insertion of marker chars?*/
        if (pDetails->ReportingOptions & DETAILS_MARKER_FLAG) {
                /*Display line up to beginning of match with marker*/
                printf("%c%-.*s", pDetails->MarkerChar,
                                pDetails->pMatchStart - pDetails->pLineStart,
                                pDetails->pLineStart);

                /*?? Should encode binary characters*/

                /*Write matching part of line with markers*/
                printf("%c%.*s%c", pDetails->MarkerChar,
                                pDetails->pMatchEnd -
                                pDetails->pMatchStart,
                                pDetails->pMatchStart,
                                pDetails->MarkerChar);

                /*Write remainder of line (but no ending delimiter)*/
                printf("%s\n", pDetails->pMatchEnd);

        } else {
                /*Format output and write directly to console*/
                cprintf("%-.*s", pDetails->pMatchStart - pDetails->pLineStart,
                                pDetails->pLineStart);

                /*Yeah, I know the colour selection is terribly crude here...*/
                /*Also, should encode binary characters*/

                /*Display match with "highlight"*/
                textattr(0x1b);
                cprintf("%.*s", pDetails->pMatchEnd -
                                pDetails->pMatchStart,
                                pDetails->pMatchStart);

                /*Display remainder of line in "normal" mode*/
                textattr(0x07);
                cputs(pDetails->pMatchEnd);
                putch(LF);

        }

        /*Count lines in case search engine cares*/
        pDetails->LineMatchCount++;

        /*Search should continue*/
        return TRUE;

} /*DisplayHighlighted*/

/************************************************************************/
/*                                                                      */
/*      FilenameOut -- Write filename and/or match count summary        */
/*                                                                      */
/************************************************************************/
public_scope BOOL
DosOut_FilenameOut(MATCHENG_DETAILS *pDetails)
{
        /*Has filename output been selected?*/
        if (pDetails->ReportingOptions & DETAILS_FILENAME) {
                /*Yes, is line counting enabled as well?*/
                if (pDetails->ReportingOptions & DETAILS_LINECOUNT) {
                        /*Yes, report it now as well*/
                        printf("%s:%lu\n", pDetails->pFilename,
                                         pDetails->LineMatchCount);

                } else {
                        /*No, report file name only (plus LF)*/
                        printf("%s\n", pDetails->pFilename);

                }

        } else {
                /*Caller must want line count display without filename*/
                printf("%lu\n", pDetails->LineMatchCount);

        }

        /*Search should abandon this file as it has been reported*/
        return FALSE;

} /*FilenameOut*/
