/************************************************************************/
/*                                                                      */
/*      ggrep -- Regular expression search using Grouse architecture    */
/*                                                                      */
/*      behoffski is seeing if his way of counting words using an       */
/*      optimised switch statement in assembly can be extended          */
/*      somewhat dramatically to include regular expressions.  If       */
/*      so, the performance should be rather astonishing.  This         */
/*      is the first step: to get a similar, slower table-driven        */
/*      version running in C.  This is to teach behoffski the full      */
/*      joys of regular expression searching.                           */
/*                                                                      */
/*      Started at 10pm on Saturday, 13-May-95.                         */
/*                                                                      */
/*      If you want to find the program's entry point (main),           */
/*      look in Platform.  The code is cut this way so that             */
/*      this module remains portable.                                   */
/*                                                                      */
/*      Copyright (C) 1995-7 Grouse Software.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include "ascii.h"
#include "bakslash.h"
#include <compdef.h>
#include "glob.h"
#include "main.h"
#include "regexp.h"
#include "scanfile.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*Maximum RE text that can be read from file*/
#define RE_TEXT_SIZE_MAX                256

/*Parameter for module diagnosis strings (?? should be in compdef?)*/
#define DIAGNOSIS_LENGTH_MAX            128

/*Codes to select RE syntax and search behaviour*/
typedef enum {
        SEARCH_AS_FGREP,
        SEARCH_AS_GREP,
        SEARCH_AS_EGREP
} SEARCH_STYLE;

/*Program options*/
typedef struct {
        BOOL FullLine;

        SEARCH_STYLE SearchStyle;

} OPT_STRUCT;

/*Permanent variables referenced within this module*/
typedef struct {
        /*Options specified by client*/
        OPT_STRUCT Opt;

        /*Search and display options used to configure ScanFile*/
        LWORD ScanOptions;
        LWORD ReportingOptions;

        /*Program return code*/
        UINT ReturnCode;

} GREP_MODULE_CONTEXT;

module_scope GREP_MODULE_CONTEXT gGgrep;

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
ggrep_Init(void)
{
        /*Program defaults to successful run*/
        gGgrep.ReturnCode = MAIN_RETURN_MATCHED;

} /*Init*/

/************************************************************************/
/*                                                                      */
/*      ReadREFromFile -- Use first line of file as RE                  */
/*                                                                      */
/*      This function is provided to support the -f option of           */
/*      ggrep.  It attempts to open the file specified, then            */
/*      read the first line as RE text.  Returns a pointer to the       */
/*      start of the buffer if successful (with NUL terminator),        */
/*      or else NULL if not successful.                                 */
/*                                                                      */
/************************************************************************/
module_scope CHAR *
ggrep_ReadREFromFile(CHAR *pFilename)
{
        FILE *f;
        CHAR *pREBuf;
        UINT len;

        /*Open named file for read*/
        f = fopen(pFilename, "r");
        if (f == NULL) {
                /*Sorry, unable to open file*/
                fprintf(stdout, "ggrep: Unable to open file %s", pFilename);
                return NULL;

        }

        /*Acquire buffer to store RE text*/
        pREBuf = malloc(RE_TEXT_SIZE_MAX);
        if (pREBuf == NULL) {
                /*Sorry, unable to acquire memory for line buffer*/
                fprintf(stdout, "ggrep: No memory for line buffer");
                return NULL;

        }

        /*Read line into buffer*/
        (void) fgets(pREBuf, RE_TEXT_SIZE_MAX, f);

        /*Does the line end with a LF?*/
        len = strlen(pREBuf);
        if (pREBuf[len - 1] != LF) {
                /*No, line is too big for buffer*/
                fprintf(stdout, "ggrep: line too long in file %s", pFilename);
                free(pREBuf);
                return NULL;

        } else {
                /*Yes, remove it*/
                pREBuf[len - 1] = NUL;
        }

        /*Finished with specifier file*/
        fclose(f);

        /*Report RE to caller*/
        return pREBuf;

} /*ReadREFromFile*/

/************************************************************************/
/*                                                                      */
/*      ReturnCode -- Allow other modules to write program return       */
/*                                                                      */
/*      Other modules may find faults during operation but be able      */
/*      keep going.  These modules must be able to report the faults    */
/*      to the main program so that the program return code can         */
/*      be updated correctly.                                           */
/*                                                                      */
/*      The use of module Main here is very dubious... sigh.            */
/*      This function might be more appropriate as part of Platform.    */
/*                                                                      */
/************************************************************************/
public_scope void
Main_ReturnCode(UINT Code)
{
        /*Is this code more severe than the current return code?*/
        if (Code > gGgrep.ReturnCode) {
                /*Yes, remember this code as program result*/
                gGgrep.ReturnCode = Code;

        }

} /*ReturnCode*/

/************************************************************************/
/*                                                                      */
/*      Usage -- Tell the user what we expect                           */
/*                                                                      */
/************************************************************************/
module_scope void
ggrep_Usage(void)
{
        fprintf(stdout,
              "Usage: ggrep [-bcdDFGhHilLnvVwx -McOxxx] ([-e] regexp | -f file) [files]");
        exit(MAIN_RETURN_FAULT);

} /*Usage*/

/************************************************************************/
/*                                                                      */
/*      Version -- Report program version and other info                */
/*                                                                      */
/************************************************************************/
module_scope void
ggrep_Version(void)
{
        /*Display program name and version*/
        fprintf(stdout, "ggrep (Grouse grep) v1.01  1-Sep-1997\n");
        fprintf(stdout,
                "Copyright (c) Grouse Software 1995-7.  All rights reserved.\n");
        fprintf(stdout,
                "Written for Grouse by behoffski (Brenton Hoff).\n");

        exit(0);

} /*Version*/

/************************************************************************/
/*                                                                      */
/*      OptiOpts -- Parse options to enable/disable optimisations       */
/*                                                                      */
/*      Yet another function inspired by the ludicrous lengths          */
/*      to which behoffski is going to build a (hopefully)              */
/*      worthwhile test rig.  The match reported by -M isn't            */
/*      strictly correct in some cases as optimisations can trim        */
/*      the ends away.  Being able to turn off optimisations at         */
/*      will allows this case to report correctly, as well as           */
/*      providing a way of checking that each layer of pattern          */
/*      interpretation is correct, not merely the final,                */
/*      optimised result.                                               */
/*                                                                      */
/************************************************************************/
module_scope void
Ggrep_OptiOpts(CHAR *pOpt)
{
        for (;;) {
                if (*pOpt == NUL) {
                        break;
                }
                switch (*pOpt++) {
                case 'A':
                        gGgrep.ScanOptions |=  SCANFILE_OPT_APPROXIMATE;
                        break;
                case 'a':
                        gGgrep.ScanOptions &= ~SCANFILE_OPT_APPROXIMATE;
                        break;
                case 'B':
                        gGgrep.ScanOptions |=  SCANFILE_OPT_BUFFER;
                        break;
                case 'b':
                        gGgrep.ScanOptions &= ~SCANFILE_OPT_BUFFER;
                        break;
                case 'E':
                        gGgrep.ScanOptions |=  SCANFILE_OPT_EASIEST_FIRST;
                        break;
                case 'e':
                        gGgrep.ScanOptions &= ~SCANFILE_OPT_EASIEST_FIRST;
                        break;
                case 'S':
                        gGgrep.ScanOptions |=  SCANFILE_OPT_SKIP;
                        break;
                case 's':
                        gGgrep.ScanOptions &= ~SCANFILE_OPT_SKIP;
                        break;

                /*default: ignored*/
                }
        }

} /*OptiOpts*/

/************************************************************************/
/*                                                                      */
/*      main -- Co-ordinate execution of program                        */
/*                                                                      */
/************************************************************************/
public_scope INT
ggrep_main(UINT argc, CHAR **argv)
{
        REGEXP_SPEC *pCompactRE = NIL;
        CHAR *pRegexpText = NULL;
        UINT NrFiles;
        CHAR *pArg;
        CHAR *pOpt;
        CHAR *pREFilename;
        CHAR Diagnosis[DIAGNOSIS_LENGTH_MAX];
        BOOL SuppressFilename = FALSE;

        /*If no arguments, give Usage message to guide user*/
        if (argc == 1) {
                ggrep_Usage();
        }
        argv++;

        /*Prepare default options*/
        gGgrep.Opt.SearchStyle = SEARCH_AS_GREP;
        gGgrep.Opt.FullLine = FALSE;
        gGgrep.ReportingOptions = DETAILS_LINE;
        gGgrep.ScanOptions = SCANFILE_OPT_APPROXIMATE | \
                             SCANFILE_OPT_EASIEST_FIRST | \
                             SCANFILE_OPT_BUFFER | \
                             SCANFILE_OPT_SKIP;

        /*Loop through arguments, processing options etc*/
        for ( ; ; argv++) {
                /*Have we run out of arguments?*/
                if (--argc == 0) {
                        /*Yes, was a RE specified by an option?*/
                        if (pRegexpText == NULL) {
                                /*No, a search without RE is meaningless*/
                                ggrep_Usage();
                        }

                        /*User wants to use stdin as input*/
                        break;

                }

                /*Refer to next argument via convenient pointer*/
                pArg = *argv;

                /*Is the next argument an option specifier?*/
                if (pArg[0] != '-') {
                        /*No, user has finished giving options*/
                        break;
                }

                /*Yes, decode option and record selection*/
                pOpt = &pArg[1];

                /*Did we find a "-" by itself?*/
                if (*pOpt == NUL) {
                        /*Yes, this is the RE to search for*/
                        break;
                }

DecodeOption:
                /*Act on next option byte*/
                switch (*pOpt++) {
                case 'b':
                        /*Include byte offset in display*/
                        gGgrep.ReportingOptions |=
                                        DETAILS_BYTEOFFSET;
                        break;

                case 'c':
                        /*Count selected lines and don't display them*/
                        gGgrep.ReportingOptions |= DETAILS_LINECOUNT;
                        gGgrep.ReportingOptions &= ~DETAILS_LINE;
                        break;

                case 'd':
                        /*Search subdirectories as well*/
                        Glob_Configure(0, GLOB_INCLUDE_SUBDIRECTORIES |
                                        GLOB_SUPPRESS_ERRORS);
                        break;

                case 'D':
                        /*Display state tables for test/debug*/
                        gGgrep.ScanOptions |= SCANFILE_DEBUG_DISPLAY;
                        break;

                case 'e':
                        /*Have we exhausted the current option word?*/
                        if (*pOpt != NUL) {
                                /*No, use remainder of word as RE*/
                                pRegexpText = pOpt;

                        } else {
                                /*Yes, next arg is RE*/
                                if (--argc == 0) {
                                        ggrep_Usage();
                                }
                                pRegexpText = *++argv;
                        }

                        /*Finished handling current option word(s)*/
                        continue;
                        break;

                case 'E':
                        /*Select extended RE syntax*/
                        gGgrep.Opt.SearchStyle = SEARCH_AS_EGREP;
                        break;

                case 'f':
                        /*Have we exhausted the current option word?*/
                        if (*pOpt != NUL) {
                                /*No, use remainder of word as file name*/
                                pREFilename = pOpt;

                        } else {
                                /*Yes, next word contains RE file name*/
                                if (--argc == 0) {
                                        ggrep_Usage();
                                }
                                pREFilename = *++argv;
                        }
                        pRegexpText = ggrep_ReadREFromFile(pREFilename);
                        if (pRegexpText == NULL) {
                                /*Failed to get RE from file*/
                                return MAIN_RETURN_FAULT;
                        }

                        /*Finished handling current option word(s)*/
                        continue;
                        break;

                case 'F':
                        /*Select fixed-string RE syntax and operation*/
                        gGgrep.Opt.SearchStyle = SEARCH_AS_FGREP;
                        break;

                case 'G':
                        /*Select basic RE syntax*/
                        /*??*/
                        break;

                case 'h':
                        /*Suppress file name prefix if displaying lines*/
                        SuppressFilename = TRUE;
                        break;

                case 'H':
                        /*Highlight match as part of display*/
                        gGgrep.ReportingOptions |= DETAILS_HIGHLIGHT;
                        break;

                case 'i':
                        /*RE becomes case-insensitive*/
                        gGgrep.ScanOptions |= SCANFILE_IGNORE_CASE;
                        break;

                case 'l':
                        /*Only show filename for matching files*/
                        gGgrep.ReportingOptions &= ~DETAILS_LINE;
                        gGgrep.ReportingOptions |= DETAILS_FILENAME;
                        break;

                case 'L':
                        /*Only show filename for nonmatch files*/
                        gGgrep.ReportingOptions |= DETAILS_NONMATCH_FILES;
                        break;

                case 'M':
                        /*Marker character to delimit match/nonmatch in line*/
                        gGgrep.ReportingOptions |= DETAILS_MARKER(*pOpt);
                        if (*pOpt++ == NUL) pOpt--;
                        break;

                case 'n':
                        /*Include file line number in display*/
                        gGgrep.ScanOptions |= SCANFILE_NUMBER_LINES;
                        gGgrep.ReportingOptions |= DETAILS_LINENUMBER;
                        break;

                case 'O':
                        /*Configure optimisations (mainly for test rig)*/
                        Ggrep_OptiOpts(pOpt);
                        continue;
                        break;

                case 'v':
                        /*Invert match sense*/
                        gGgrep.ScanOptions |= SCANFILE_INVERT_MATCH_SENSE;
                        break;

                case 'V':
                        /*Display program version info*/
                        ggrep_Version();
                        break;

                case 'w':
                        /*Only match if RE matches a word*/
                        gGgrep.ScanOptions |= SCANFILE_MATCH_WORDS;
                        break;

                case 'x':
                        /*Only match full lines*/
                        gGgrep.Opt.FullLine = TRUE;
                        break;

                default:
                        fprintf(stdout, "ggrep: Unknown option: %c\n",
                                        pOpt[-1]);
                        ggrep_Usage();
                        break;

                }

                /*Have we reached the end of the word?*/
                if (*pOpt != NUL) {
                        /*No, handle next option within this word*/
                        goto DecodeOption;
                }

                /*Loop will consider next word*/

        }


        /*Did we get a RE spec while handling the option list?*/
        if (pRegexpText == NULL) {
                /*No, use next parameter as RE text*/
                pRegexpText = *argv++;
                argc--;

        }

        /*Translate escape codes*/
        if (! Bakslash_DecodeLine(pRegexpText)) {
                fprintf(stdout, "ggrep: Trailing backslash\n");
                return MAIN_RETURN_FAULT;

        }

        /*Did the client want only matches that were full lines?*/
        if (gGgrep.Opt.FullLine) {
                /*Yes, get RegExp to edit pattern accordingly*/
                RegExp_Configure(0, REGEXP_MATCH_FULL_LINE_ONLY);
        }

        /*Convert ASCII RE description into compact form*/
        switch (gGgrep.Opt.SearchStyle) {
        case SEARCH_AS_EGREP:
                /*Sorry, can only handle basic and fixed-string behaviours*/
                fprintf(stdout,
                        "ggrep: Can't cope with extended REs (yet!)\n");
                return MAIN_RETURN_FAULT;
                break;

        case SEARCH_AS_FGREP:
                /*Treat pattern as a simple string without special chars*/
                if (! RegExp_CompileString(pRegexpText, &pCompactRE)) {
                        /*Found some fault during conversion*/
                        RegExp_Diagnose(Diagnosis);
                        fprintf(stdout, "ggrep: %s\n", Diagnosis);
                        return MAIN_RETURN_FAULT;
                }
                break;

        case SEARCH_AS_GREP:
                /*Compile RE syntax and interpret basic RE specifiers*/
                if (! RegExp_Compile(pRegexpText, &pCompactRE)) {
                        /*Found some fault during conversion*/
                        RegExp_Diagnose(Diagnosis);
                        fprintf(stdout, "ggrep: %s\n", Diagnosis);
                        return MAIN_RETURN_FAULT;
                }
                break;

        }

        /*Are we displaying any lines?*/
        if (!(gGgrep.ReportingOptions & (DETAILS_LINE | DETAILS_HIGHLIGHT))) {
                /*No, don't count lines (in case we were asked to)*/
                gGgrep.ScanOptions &= ~SCANFILE_NUMBER_LINES;
        }

        /*Does the client want to see the exact matching text?*/
        if (gGgrep.ReportingOptions & DETAILS_HIGHLIGHT) {
                /*Yes, scanfile can't use approximations at start and end*/
                gGgrep.ScanOptions &= ~SCANFILE_OPT_APPROXIMATE;
        }

        /*Give ScanFile regular expression pattern plus search modifications*/
        if (! ScanFile_Pattern(pCompactRE, gGgrep.ScanOptions)) {
                /*ScanFile was unable to accept pattern for search*/
                fprintf(stdout, "ggrep: Unable to set up search?!\n");
                return MAIN_RETURN_FAULT;
        }

        /*Discard compact pattern since it isn't used again*/
        /*RegExp_Discard(pCompactRE);*/

        /*Remaining arguments (if any) are files to search*/

        /*Was the only file name "-"?*/
        if ((argc == 1) && (strcmp(*argv, "-") == 0)) {
                /*Yes, delete it so that stdin is selected*/
                argc--;
                argv++;
        }

        /*Were any files given?*/
        if (argc == 0) {
                /*No, immediately configure file scan for search*/
                ScanFile_Configure(gGgrep.ReportingOptions);

                /*Scan stdin using line-by-line search*/
                ScanFile_Search(NULL);

        } else {
                /*Pass all file specs to Glob for consideration*/
                Glob_Clear();
                do {
                        /*Add specification to list maintained by Glob*/
                        if (! Glob_AddSpec(*argv++)) {
                                /*Glob was unable to cope?!*/
                                fprintf(stdout,
                                        "ggrep: Unable to glob the files!\n");
                                return MAIN_RETURN_FAULT;
                        }

                } while (--argc != 0);

                /*Expand file specifications to complete file list*/
                NrFiles = Glob_Expand();
                switch (NrFiles) {
                case 0:
                        /*No files found: ?? report error*/
                        fprintf(stdout, "ggrep: No files found\n");
                        return MAIN_RETURN_FAULT;
                        break;

                case 1:
                        /*Single file: filename not required*/
                        break;

                default:
                        /*Multiple files: have filenames been suppressed?*/
                        if (! SuppressFilename) {
                                /*No, display filename as part of match*/
                                gGgrep.ReportingOptions |= DETAILS_FILENAME;
                        }
                        break;

                }

                /*Configure file scan for search*/
                ScanFile_Configure(gGgrep.ReportingOptions);

                /*Get Glob to loop through each file and call us back*/
                Glob_EnumerateFiles(ScanFile_Search);

        }

        /*Did no lines match (and no other errors detected)?*/
        if ((! ScanFile_MatchedAny()) && (gGgrep.ReturnCode == 0)) {
                /*Yes, report failure to environment*/
                gGgrep.ReturnCode = 1;
        }

        /*Report success or failure*/
        return gGgrep.ReturnCode;

} /*main*/
