/************************************************************************/
/*                                                                      */
/*      Glob -- Perform file globbing, with Unix/DOS flavours           */
/*                                                                      */
/*      This module helps non-Unix programs handle files in a           */
/*      similar fashion to programs executed under a Unix shell.        */
/*      Wildcard specifications are collected by repeated calls to      */
/*      AddSpec.  Configuation options such as including                */
/*      subdirectories may be specified before or during the            */
/*      collection.                                                     */
/*                                                                      */
/*      Once all specs have been described, call Expand which           */
/*      finds matching files and sorts them into alphabetical           */
/*      order.  The number of files found by Expand is reported.        */
/*      However, if subdirectories are included, this count is          */
/*      reported as 9999.                                               */
/*                                                                      */
/*      After expanding, call EnumerateFiles to step through            */
/*      the files found.  EnumerateFiles will also descend into         */
/*      subdirectories if this option was requested, but will not       */
/*      try to perform any file name sorting.                           */
/*                                                                      */
/*      This module is a tad ugly, because it tries to blend three      */
/*      conflicting design criteria.  The first is that explicit        */
/*      file specifications should be globbed as if globbed by a        */
/*      Unix shell.  The second criteria is that recursing down         */
/*      subdirectories should be supported under DOS because            */
/*      DOS is so primitive and this utility tries to be helpful.       */
/*      The third criteria is that each directory is only examined      */
/*      once to minimise activity where the search is being             */
/*      conducted on a network drive.                                   */
/*                                                                      */
/*      NOTE: This file is _not_ intended to be portable -- it is       */
/*      focussed on bringing Unix-like globbing to DOS.  The correct    */
/*      way of porting the functions offered by this module             */
/*      (specified in glob.h) is to replace this file (glob.c)          */
/*      with a different file that operates correctly for the           */
/*      target system.  For example, globbing in Unix would simply      */
/*      become recording the filename pointers given via AddSpec        */
/*      and replaying those filespecs when EnumerateFiles is called.    */
/*      Bits of this module might still be useful if subdirectory       */
/*      recursion is desired even in the face of wonderful facilities   */
/*      such as find.                                                   */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include "ascii.h"
#include <compdef.h>
#include <ctype.h>
#include <dir.h>
#include <dos.h>
#include "glob.h"
#include <io.h>
#include "main.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define FILENAME_SIZE_MAX               256
#define MATCH_STACK_SIZE                64
#define GLOB_PATH_SEPARATOR             '/'
#define GLOB_PATH_SEPARATOR_STR         "/"

#ifndef _FARFUNC
#define _FARFUNC
#endif /*FARFUNC*/

/*Structure to allow specs to be handled as path/filename pairs*/

typedef struct {
        /*Pointers to specification path and name components*/
        CHAR *pPath;
        CHAR *pName;

        /*?? Counter of specification matches for error reporting*/

} GLOB_FILE_PATH_PAIR;

typedef int (qsort_compare_func)(const void *_e1, const void *_e2);

typedef struct {
        /*Module configuration options*/
        UINT32 ConfigOpts;

        /*List of file specifiers from client*/
        UINT NrSpecs;
        UINT NrSpecsMax;
        GLOB_FILE_PATH_PAIR *pSpecs;

        /*Buffer to store expanded file names*/
        UINT ExpandBytesMax;
        CHAR *pExpandBuf;
        CHAR *pExpandCurr;

        /*Number of expanded files found*/
        UINT NrFiles;

        /*Pointers to filenames for sorting*/
        CHAR **ppSort;
        CHAR **ppSortCurr;

        /*Program name for error reporting*/
        CHAR *pProgName;

} GLOB_MODULE_CONTEXT;

module_scope GLOB_MODULE_CONTEXT gGlob;

/************************************************************************/
/*                                                                      */
/*      LowerStr -- Convert string to lower-case                        */
/*                                                                      */
/************************************************************************/
module_scope void
Glob_LowerStr(CHAR *pStr)
{
        for (; *pStr != NUL; pStr++) {
                *pStr = tolower(*pStr);
        }

} /*LowerStr*/

/************************************************************************/
/*                                                                      */
/*      Configure -- Set how the module reacts to important events      */
/*                                                                      */
/*      Some behaviours of the module may change from run to run,       */
/*      or may be set differently from one application to another.      */
/*      This function allows these behaviours to be specified by        */
/*      the client.  This helps isolate the module from other parts     */
/*      of the application so that it is easier to reuse.  The          */
/*      default configuration has all options off (all bits 0).         */
/*                                                                      */
/************************************************************************/
public_scope void
Glob_Configure(UINT32 OffConfigs, UINT32 OnConfigs)
{
        /*Record configuration in internal variable*/
        gGlob.ConfigOpts = (gGlob.ConfigOpts & ~OffConfigs) | OnConfigs;

} /*Configure*/

/************************************************************************/
/*                                                                      */
/*      Clear -- Erase spec and expanded lists and start again          */
/*                                                                      */
/*      Although calling Glob multiple times seems unlikely, it         */
/*      may be the most effective way of handling file searches         */
/*      that recurse into subdirectories.                               */
/*                                                                      */
/************************************************************************/
public_scope void
Glob_Clear(void)
{
        /*Destroy list of specifiers*/
        gGlob.NrSpecs = 0;

        /*Reset expanded file name work area*/
        *gGlob.ppSort = NULL;

} /*Clear*/

/************************************************************************/
/*                                                                      */
/*      AddSpec -- Collect pointer to file specification for globbing   */
/*                                                                      */
/*      The client should call this function for each file specific-    */
/*      ation to be globbed.  The function returns FALSE if it was      */
/*      unable to store the specification.                              */
/*                                                                      */
/*      The memory containing the string text must be preserved         */
/*      by the client, as Glob merely remembers the pointer.            */
/*                                                                      */
/*      The insertion sort here is appropriate for DOS where the        */
/*      wildcards are presented intact.  It is not intended for use     */
/*      by Unix systems, for the entire implementation (glob.c) should  */
/*      replaced by a version enormously simpler because all the        */
/*      hard work has already been done by the command shell.           */
/*                                                                      */
/************************************************************************/
public_scope BOOL
Glob_AddSpec(CHAR *pFileSpec)
{
        GLOB_FILE_PATH_PAIR *pSpecs;

        /*Do we have space for another specification?*/
        if (gGlob.NrSpecs >= gGlob.NrSpecsMax) {
                /*No, cannot store*/
                return FALSE;
        }

        /*Add specification into ordered list using insertion sort*/
        pSpecs = &gGlob.pSpecs[gGlob.NrSpecs];
        for (;;) {
                /*Have we reached the top of the list?*/
                if (pSpecs == &gGlob.pSpecs[0]) {
                        /*Yes, insert at top of list*/
                        break;

                }

                /*Does the specification come after this list entry?*/
                if (strcmp(pSpecs[-1].pPath, pFileSpec) < 0) {
                        /*Yes, insert after this specification*/
                        break;
                }

                /*Move entry down to make space and try prev entry in list*/
                memcpy(&pSpecs[0], &pSpecs[-1], sizeof(*pSpecs));
                pSpecs--;

        }

        /*Add specification to list*/
        gGlob.NrSpecs++;
        pSpecs[0].pPath = pFileSpec;

        /*Report success to client*/
        return TRUE;

} /*AddSpec*/

/************************************************************************/
/*                                                                      */
/*      ConvertSpec -- Split and reformat wildcard spec into path/file  */
/*                                                                      */
/************************************************************************/
module_scope void
Glob_ConvertSpec(CHAR **ppDOSPath, CHAR **ppFile)
{
        CHAR *pPath = *ppDOSPath;
        CHAR Drive[MAXDRIVE+1];
        CHAR Dir[MAXDIR+1];
        CHAR fName[MAXFILE+1];
        CHAR Ext[MAXEXT+1];
        CHAR *pName;

        /*Extract path details from spec for full name*/
        memset(Drive, NUL, sizeof(Drive));
        memset(Dir, NUL, sizeof(Dir));
        memset(fName, NUL, sizeof(fName));
        memset(Ext, NUL, sizeof(Ext));
        (void) fnsplit(pPath, Drive, Dir, fName, Ext);

        /*Use initial specification as path string*/
        pName = &pPath[strlen(Dir) + strlen(Drive)];
        *pName = NUL;

        /*Place filename in expansion buffer*/
        *ppFile = gGlob.pExpandCurr;
        strcpy(gGlob.pExpandCurr, fName);
        if (strlen(Ext) == 0) {
                /*Use extension of "." if not specified (sigh)*/
                strcpy(Ext, ".");
        }
        strcat(gGlob.pExpandCurr, Ext);
        gGlob.pExpandCurr += strlen(gGlob.pExpandCurr) + 1;

        /*Convert path to lower case; also change '\' to '/'*/
        for (pName = pPath; *pName != NUL; pName++) {
                /*Is this character a DOS directory path separator?*/
                if (*pName == '\\') {
                        /*Yes, convert to Unix format*/
                        *pName = GLOB_PATH_SEPARATOR;
                } else {
                        /*No, ensure letters are lower case*/
                        *pName = tolower(*pName);

                }

        }

        /*Convert filename letters to lower case*/
        Glob_LowerStr(*ppFile);

} /*ConvertSpec*/

/************************************************************************/
/*                                                                      */
/*      AddName -- Convert DOS filename to globbed format, add to list  */
/*                                                                      */
/************************************************************************/
module_scope BOOL
Glob_AddName(CHAR *pDOSName, CHAR *pPathName, CHAR *pExpandCurr)
{
        /*Do we have space to add this to the expansion?*/
        if ((strlen(pDOSName) + 2 + (pExpandCurr -
                        gGlob.pExpandBuf)) >
                        gGlob.ExpandBytesMax) {
                /*No, cannot complete expansion*/
                return FALSE;
        }

        /*Add path as a prefix to file*/
        strcpy(pExpandCurr, pPathName);

        /*Add this file to expansion*/
        strcat(pExpandCurr, pDOSName);

        /*Convert file name to lower case*/
        Glob_LowerStr(pExpandCurr);

        /*Expansion succeeded*/
        return TRUE;

} /*AddName*/

/************************************************************************/
/*                                                                      */
/*      MatchFilename -- Report if filename matches pattern             */
/*                                                                      */
/*      This function only knows about very primitive DOS-style         */
/*      wildcard matches.  In the future more intelligence will be      */
/*      added.  Note that the pattern match performed here is           */
/*      case-sensitive: if you want case insensitivity, fold the        */
/*      case of the pattern and the filename before calling             */
/*      this routine.                                                   */
/*                                                                      */
/*      behoffski really should use the RE search engine here, but      */
/*      is too tired to have a go just now.                             */
/*                                                                      */
/************************************************************************/
module_scope BOOL
Glob_MatchFilename(CHAR *pPattern, CHAR *pFilename)
{
        CHAR FileCh;
        CHAR PatCh;
        CHAR *Stack[MATCH_STACK_SIZE];
        CHAR **pStack = &Stack[MATCH_STACK_SIZE - 1];

        /*Conform to DOS's expectations: "*.*" really means "everything"*/
        if (strcmp(pPattern, "*.*") == 0) {
                return TRUE;
        }

        /*Does the file name have an extension?*/
        if (strchr(pFilename, '.') == NULL) {
                /*No, add a period so we conform to DOS's rules*/
                strcat(pFilename, ".");
        }

        *--pStack = NULL;
        *--pStack = NULL;

        /*Work through match, driven by filename*/
        PatCh = *pPattern++;

        /*Did we start with a '*' which includes a zero trip?*/
        if (PatCh == '*') {
                /*Yes, add zero trip to backtracking stack*/
                *--pStack = pPattern;
                *--pStack = pFilename;
        }

        for (;;) {
                /*Get next char of filename*/
                FileCh = *pFilename++;
                if (FileCh == PatCh) {
                        /*Pattern/file matched: both reached the end?*/
                        if (FileCh == NUL) {
                                /*Yes, match succeeded*/
                                return TRUE;
                        }

                        /*Note: assumes filename char is not '*' or '?'*/
                        PatCh = *pPattern++;

                        /*Have we advanced to a '*'?*/
                        if (PatCh == '*') {
                                /*Yes, push zero-trip match*/
                                *--pStack = pPattern;
                                *--pStack = pFilename;
                        }

                        continue;
                }

                /*Does the pattern match any character?*/
                if (PatCh == '*') {
                        /*Yes, match if filename not NUL*/
                        if (FileCh != NUL) {
                                /*Is backtracking worthwhile?*/
                                if (*pFilename == *pPattern) {
                                        /*Yes, record backtrack possibility*/
                                        *--pStack = pPattern;
                                        *--pStack = pFilename;
                                }

                                /*Don't match if filename char was '.'*/
                                if (FileCh != '.') {
                                        continue;
                                }

                        /*Have we reached the end of a trailing * match?*/
                        } else if (*pPattern == NUL) {
                               /*Yes, match succeeded*/
                               return TRUE;

                        }

                }


                if (FileCh == '.') {
                        /*Component did not match... try backtracking*/
                        goto Backtrack;
                }

                /*Does the pattern match any single char?*/
                if (PatCh == '?') {
                        /*Yes, match if filename not NUL*/
                        if (FileCh != NUL) {
                                PatCh = *pPattern++;
                                continue;
                        }
                }

                /*No match, backtrack*/
Backtrack:
                pFilename = *pStack++;
                pPattern = *pStack++;
                if (pPattern == NULL) {
                        /*Did not find a match*/
                        return FALSE;
                }

                PatCh = *pPattern++;

                /*Is the next element '*' which includes a zero trip?*/
                if (PatCh == '*') {
                        /*Yes, add zero trip to backtracking stack*/
                        *--pStack = pPattern;
                        *--pStack = pFilename;
                }

        }

        /*Should not reach here... but merely fail match for now*/
        return FALSE;

} /*MatchFilename*/

/************************************************************************/
/*                                                                      */
/*      FilenameMatchInList -- Report if filename matches any of list   */
/*                                                                      */
/*      This function takes the current specification path/name         */
/*      entry plus any following entries with the same path, and        */
/*      reports whether the filename matches any of the filename        */
/*      specifications.                                                 */
/*                                                                      */
/************************************************************************/
module_scope BOOL
Glob_FilenameMatchInList(GLOB_FILE_PATH_PAIR *pSpecs, CHAR *pFilename)
{
        /*Loop through all specs containing same path*/
        for (;;) {
                /*Does the filename match this specifier?*/
                if (Glob_MatchFilename(pSpecs->pName, pFilename))  {
                        /*Yes, report success*/
                        return TRUE;
                }

                /*Is the next specifier for the same path?*/
                if (pSpecs[1].pPath != pSpecs[0].pPath) {
                        /*No, match failed*/
                        return FALSE;
                }

                /*Move to the next specifier and try again*/
                pSpecs++;

        }

} /*FilenameMatchInList*/

/************************************************************************/
/*                                                                      */
/*      SeaprateSpecs -- Break file specifications into path/name       */
/*                                                                      */
/************************************************************************/
module_scope void
Glob_SeparateSpecs(void)
{
        GLOB_FILE_PATH_PAIR *pSpecs;
        GLOB_FILE_PATH_PAIR *pSpecEnd = gGlob.pSpecs + gGlob.NrSpecs;
        CHAR *pPrevPath = NULL;

        /*Loop through each specification*/
        for (pSpecs = gGlob.pSpecs; pSpecs != pSpecEnd; pSpecs++) {
                /*Break spec into (reformatted) path and file components*/
                Glob_ConvertSpec(&pSpecs->pPath, &pSpecs->pName);

                /*Did we perform a split previously?*/
                if (pPrevPath != NULL) {
                        /*Yes, is this path the same as the previous one?*/
                        if (strcmp(pPrevPath, pSpecs->pPath) == 0) {
                                /*Yes, use same pointer to optimise search*/
                                pSpecs->pPath = pPrevPath;
                                continue;
                        }

                }

                /*Path is different (or first time through): remember it*/
                pPrevPath = pSpecs->pPath;

        }

        /*Write NULL for final specifier to serve as list terminator*/
        pSpecs->pPath = NULL;
        pSpecs->pName = NULL;

} /*SeparateSpecs*/

/************************************************************************/
/*                                                                      */
/*      SearchDirectory -- Perform globbing on specified directory      */
/*                                                                      */
/*      This function is the backbone of globbing both with and         */
/*      without subdirectory search.  It inserts the subdirectory       */
/*      specifier into the filespec and uses the limited                */
/*      filename globbing facilities to match file names.  Each         */
/*      file found is added to the list of file found by the            */
/*      expansion.  The function returns FALSE if it was unable to      */
/*      complete the expansion for some reason.                         */
/*                                                                      */
/************************************************************************/
module_scope BOOL
Glob_SearchDirectory(void)
{
        struct ffblk DosFB;
        CHAR *pExpansion = gGlob.pExpandCurr;
        GLOB_FILE_PATH_PAIR *pSpecs;
        GLOB_FILE_PATH_PAIR *pSpecEnd = gGlob.pSpecs + gGlob.NrSpecs;
        CHAR Path[FILENAME_SIZE_MAX];
        CHAR **ppSort = gGlob.ppSortCurr;
        CHAR NameLower[13];

        /*Loop through each specification*/
        for (pSpecs = gGlob.pSpecs; pSpecs != pSpecEnd; pSpecs++) {
                /*Have we considered any specs previously?*/
                if (pSpecs != gGlob.pSpecs) {
                         /*Yes, is this path same as the previous one?*/
                         if (pSpecs->pPath == pSpecs[-1].pPath) {
                                 /*Yes, skip spec: already handled earlier*/
                                 continue;
                         }
                }

                /*Scan directory named by this path*/
                strcpy(Path, pSpecs->pPath);
                strcat(Path, "*.*");

                /*Search files and directories in specified path*/
                if (findfirst(Path, &DosFB, FA_DIREC) == -1) {
                        /*Unable to find any... do we complain?*/
                        if (! (gGlob.ConfigOpts & GLOB_SUPPRESS_ERRORS)) {
                                /*Yes, generate message*/
                                fprintf(stderr,
                                        "%s: %s: No such file or directory\n",
                                        gGlob.pProgName, Path);

                                /*Tell program of fault*/
                                Main_ReturnCode(MAIN_RETURN_FAULT);

                        }

                        /*Should we give up?*/
                        if (gGlob.ConfigOpts & GLOB_HALT_ON_ERROR) {
                                /*Yes, abandon expansion*/
                                return FALSE;

                        }

                        /*Spec not found, but keep going*/
                        continue;

                }

                /*Write each file matching specification*/
                for (;;) {
                        /*Is this spec "." or ".."?*/
                        if ((DosFB.ff_name[0] == '.') &&
                            (DosFB.ff_name[strlen(DosFB.ff_name)-1] == '.')) {
                                /*Yes, ignore it*/
                                goto SkipFile;
                        }

                        /*Convert filename to lower case*/
                        memcpy(&NameLower[0], &DosFB.ff_name[0],
                                        sizeof(NameLower));
                        Glob_LowerStr(NameLower);

                        /*Is this file a subdirectory?*/
                        if (DosFB.ff_attrib & FA_DIREC) {
                                /*Yes, are we recursing into subdirectories?*/
                                if (gGlob.ConfigOpts &
                                               GLOB_INCLUDE_SUBDIRECTORIES) {
                                       /*Yes, add directory to expansion*/
                                       if (! Glob_AddName(NameLower,
                                                        pSpecs->pPath,
                                                        pExpansion)) {
                                                /*Sorry, unable to add*/
                                                return FALSE;
                                       }

                                       *ppSort++ = pExpansion;
                                       strcat(pExpansion, "/");
                                       pExpansion += strlen(pExpansion) + 1;
                                       gGlob.NrFiles++;

                                }

                                /*Don't seek name match*/
                                goto SkipFile;
                        }

                        /*Does file match any of the specs for this path?*/
                        if (! Glob_FilenameMatchInList(pSpecs, NameLower)) {
                                /*No, skip to next file, if any*/
                                goto SkipFile;
                        }

                        /*Add this file to expansion*/
                        if (! Glob_AddName(NameLower,
                                        pSpecs->pPath, pExpansion)) {
                                /*Sorry, unable to add (not enough storage?)*/
                                return FALSE;
                        }

                        gGlob.NrFiles++;

                        /*Remember file name for sorting*/
                        *ppSort++ = pExpansion;

                        /*Prepare for next expansion, if any*/
                        pExpansion += strlen(pExpansion) + 1;

                        /*Find next file that matches spec, if any*/
SkipFile:
                        if (findnext(&DosFB) == -1) {
                                /*No more files found*/
                                break;

                        }

                }

        }

        /*Finished search correctly... terminate list and report success*/
        *ppSort = NULL;
        gGlob.ppSortCurr = ppSort;
        return TRUE;

} /*SearchDirectory*/

/************************************************************************/
/*                                                                      */
/*      Compare -- Filename order comparison for qsort                  */
/*                                                                      */
/************************************************************************/
module_scope int _FARFUNC
Glob_Compare(const void *_e1, const void *_e2)
{
        return strcmp(*(CHAR **)_e1, *(CHAR **)_e2);

} /*Compare*/

/************************************************************************/
/*                                                                      */
/*      Expand -- Compile list of files from specification(s)           */
/*                                                                      */
/*      Once the list of file specifications is complete, the client    */
/*      must call expand to expand the list into file names.  The       */
/*      function returns the number of files found.                     */
/*                                                                      */
/************************************************************************/
public_scope UINT
Glob_Expand(void)
{
        /*Reset current expanded file to start of buffer*/
        gGlob.pExpandCurr = gGlob.pExpandBuf;
        *gGlob.pExpandCurr = NUL;
        *gGlob.ppSort = NULL;
        gGlob.ppSortCurr = gGlob.ppSort;

        /*Break specifications into path/filename parts (uses expansion buf)*/
        Glob_SeparateSpecs();

        /*Count how many files were found during expansion*/
        gGlob.NrFiles = 0;

        /*Expand the files starting with the default directory*/
        Glob_SearchDirectory();

        /*Sort files into alphabetical order*/
        qsort(gGlob.ppSort, gGlob.NrFiles, sizeof(CHAR *), Glob_Compare);

        /*Did the user request that we search subdirectories?*/
        if (gGlob.ConfigOpts & GLOB_INCLUDE_SUBDIRECTORIES) {
                /*Yes, lie now since we didn't count subdirectory files*/
                return 9999;

        }

        /*Report number of files found to caller*/
        return gGlob.NrFiles;

} /*Expand*/

/************************************************************************/
/*                                                                      */
/*      ExpandSubdir -- Recursive subdirectory expansion                */
/*                                                                      */
/************************************************************************/
module_scope BOOL
Glob_ExpandSubdir(CHAR *pSubdir, GLOB_ENUM_FUNCTION *pClientFunc,
                GLOB_FILE_PATH_PAIR *pSpec)
{
        CHAR Path[FILENAME_SIZE_MAX];
        struct ffblk DosFB;
        CHAR *pFile;

        /*Form the full path from spec path plus subdirectory*/
        strcpy(Path, pSpec->pPath);
        Glob_LowerStr(pSubdir);
        strcat(Path, pSubdir);
        pFile = &Path[strlen(Path)] + 1;
        strcat(Path, "/*.*"); /*?? GLOB_PATH_SEPARATOR_STR + "*.*"*/

        /*Ask DOS to find all files (and directories) of this path*/
        if (findfirst(Path, &DosFB, FA_DIREC) == -1) {
                /*Unable to find any... do we complain?*/
                return TRUE;
        }

        /*Loop through each name found by search*/
        for (;;) {
                /*Is this spec "." or ".."?*/
                if ((DosFB.ff_name[0] == '.') &&
                        (DosFB.ff_name[strlen(DosFB.ff_name)-1] == '.')) {
                        /*Yes, ignore it*/
                        goto SkipFile;
                }

                /*Force DOS file name to lower case*/
                Glob_LowerStr(DosFB.ff_name);

                /*Is this file a subdirectory?*/
                if (DosFB.ff_attrib & FA_DIREC) {
                        /*Yes, recurse into it now*/
                        strcpy(Path, pSubdir);
                        strcat(Path, GLOB_PATH_SEPARATOR_STR);
                        strcat(Path, DosFB.ff_name);
                        Glob_ExpandSubdir(Path, pClientFunc, pSpec);

                        /*Restore path to previous for name generation*/
                        strcpy(Path, pSpec->pPath);
                        strcat(Path, pSubdir);
                        strcat(Path, GLOB_PATH_SEPARATOR_STR);

                /*Otherwise, does name match specifications?*/
                } else if (Glob_FilenameMatchInList(pSpec,
                                DosFB.ff_name)) {
                        /*Yes, get client to process it now*/
                        strcpy(pFile, DosFB.ff_name);
                        if (! pClientFunc(Path)) {
                                /*Client has asked us to abandon*/
                                return FALSE;
                        }

                }

                /*Find next file in directory, if any*/
SkipFile:
                if (findnext(&DosFB) == -1) {
                        /*No more files found*/
                        break;
                }

        }

        /*Finished recursion successfully*/
        return TRUE;

} /*ExpandSubdir*/

/************************************************************************/
/*                                                                      */
/*      RecurseSubdir -- Top-level function to search subdirectories    */
/*                                                                      */
/*      This function finds the list of specifications given for        */
/*      the specified path and invokes the recursive search.            */
/*                                                                      */
/*      The function returns FALSE if the search is to be abandoned.    */
/*                                                                      */
/************************************************************************/
module_scope BOOL
Glob_RecurseSubdir(CHAR *pSubdir, GLOB_ENUM_FUNCTION *pClientFunc)
{
        CHAR *pSub;
        GLOB_FILE_PATH_PAIR *pSpec;
        CHAR SubFirstChar;

        /*Remove the last path component to reveal the original spec*/
        pSub = strrchr(pSubdir, GLOB_PATH_SEPARATOR);
        if (pSub == NULL) {
                /*Sigh... DOS path might be present: ends with ':'*/
                pSub = strrchr(pSubdir, ':');
        }

        if (pSub == NULL) {
                /*No path above this subdirectory found*/
                pSub = pSubdir;
                pSubdir = "";
                SubFirstChar = *pSub;

        } else {
                /*Break into parent path and last component*/
                SubFirstChar = *++pSub;
                *pSub = NUL;

        }

        /*Find this spec in original list*/
        pSpec = gGlob.pSpecs;
        for (;;) {
                if (strcmp(pSubdir, pSpec->pPath) == 0) {
                        /*Found the first original specifier*/
                        break;
                }

                /*Move to next spec*/
                pSpec++;

                /*Have we hit the bottom of the list?*/
                if (pSpec->pPath == NULL) {
                        /*Yes, serious error: this should never happen*/
                        fprintf(stderr, "?? Glob: recurse subdir %s?",
                                        pSubdir);
                        exit(99);
                }

        }

        /*Perform recursion now correct set of specs found*/
        *pSub = SubFirstChar;
        if (! Glob_ExpandSubdir(pSub, pClientFunc, pSpec)) {
                /*Abandon expansion (probably by client request)*/
                return FALSE;
        }

        return TRUE;

} /*RecurseSubdir*/

/************************************************************************/
/*                                                                      */
/*      EnumerateFiles -- Call client's function for each file found    */
/*                                                                      */
/*      This function simplifies the client's job in acting on          */
/*      each expanded file.  It also allows the implementation of       */
/*      globbing to be changed more easily on different systems.        */
/*                                                                      */
/*      During the enumeration, the client function specified here      */
/*      should return TRUE if the enumeration is to continue or         */
/*      FALSE if the enumeration is to be abandoned.                    */
/*                                                                      */
/************************************************************************/
public_scope void
Glob_EnumerateFiles(GLOB_ENUM_FUNCTION *pClientFunc)
{
        CHAR **ppFilesSorted;
        CHAR *pFile;
        CHAR *pLastChar;

        /*Loop through each file found*/
        for (ppFilesSorted = gGlob.ppSort; *ppFilesSorted != NULL;
                        ppFilesSorted++) {
                /*Convert file list reference into more convenient handle*/
                pFile = *ppFilesSorted;

                /*Were we collecting directories during enumeration?*/
                if (gGlob.ConfigOpts & GLOB_INCLUDE_SUBDIRECTORIES) {
                        /*Yes, is this a directory?*/
                        pLastChar = &pFile[strlen(pFile) - 1];
                        if (*pLastChar == GLOB_PATH_SEPARATOR) {
                                /*Yes, perform directory recursion now*/
                                *pLastChar = NUL;
                                Glob_RecurseSubdir(pFile, pClientFunc);
                                continue;

                        }

                }

                /*Call client so it can handle this file*/
                if (! pClientFunc(pFile)) {
                        /*Client has requested that enumeration abort*/
                        break;
                }

        }

} /*EnumerateFiles*/

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/*      pProgName allows Glob to produce messages prefixed by the       */
/*      program name.  SpecsMax and ExpandBytesMax allow the client     */
/*      to have some control over memory consumption, but allow         */
/*      the implementation to choose very simple memory management      */
/*      strategies.                                                     */
/*                                                                      */
/*      The function returns FALSE if the requested resources were      */
/*      not available.                                                  */
/*                                                                      */
/************************************************************************/
public_scope BOOL
Glob_Start(CHAR *pProgName, UINT SpecsMax, UINT ExpandBytesMax)
{
        /*Allocate space for pointers to specifications*/
        gGlob.pSpecs = malloc(sizeof(GLOB_FILE_PATH_PAIR) * SpecsMax);
        if (gGlob.pSpecs == NULL) {
                /*Sorry, unable to allocate buffer*/
                return FALSE;
        }

        /*Allocate space for expanded file names*/
        gGlob.pExpandBuf = malloc(ExpandBytesMax);
        if (gGlob.pExpandBuf == NULL) {
                /*Sorry, unable to allocate expansion buffer*/
                free(gGlob.pSpecs);
                return FALSE;
        }

        /*Allocate space for name pointers for sorting*/
        gGlob.ppSort = (CHAR **) malloc((ExpandBytesMax / 8) *
                        sizeof(CHAR *));
        if (gGlob.ppSort == NULL) {
                /*Sorry, unable to allocate storage for sort pointers*/
                free(gGlob.pExpandBuf);
                free(gGlob.pSpecs);
                return FALSE;

        }

        /*Record parameters provided by caller*/
        gGlob.pProgName = pProgName;
        gGlob.NrSpecsMax = SpecsMax;
        gGlob.ExpandBytesMax = ExpandBytesMax;

        /*Prepare space for operation*/
        Glob_Clear();

        /*Started successfully*/
        return TRUE;

} /*Start*/

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
Glob_Init(void)
{
        /*No buffers for specs or expansion provided initially*/
        gGlob.pSpecs = (void *) NIL;
        gGlob.pExpandBuf = NIL;
        gGlob.pProgName = NIL;

        /*All configuration options default to OFF*/
        gGlob.ConfigOpts = 0;

} /*Init*/
