/************************************************************************/
/*                                                                      */
/*      LineFile -- Provide low-cost line-based file access             */
/*                                                                      */
/*      LineFile provides a service by which a file can be              */
/*      retrieved line-by-line or buffer-by-buffer.  The module         */
/*      provides space for the client to add bytes to the start         */
/*      and end of the buffer.  The module is faster than               */
/*      the typical stdio line handing functions, for both              */
/*      line-based access but especially for buffer-based access.       */
/*                                                                      */
/*      The module contains some provision for double buffering         */
/*      the file as this improves performance, but requires             */
/*      a worthwhile operating system to provide the necessary          */
/*      file handling primitives.                                       */
/*                                                                      */
/*      The treatment of end-of-file cases is rather dubious in         */
/*      some cases, including where the last line is terminated         */
/*      by a CTRL/Z.  Definitive descriptions on the correct way        */
/*      to behave, and perhaps even corrected source code, would        */
/*      be welcomed.                                                    */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include "ascii.h"
#include <compdef.h>
#include "linefile.h"
#include <io.h>
#include <memlchr.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef TRACE
#define TRACE(s)                        /*printf(s)*/
#define TRACE1(s, s1)                   /*printf((s), (s1))*/
#endif /*TRACE*/

#define BLOCK_SIZE_MAX                  65520l

typedef struct {
        /*Pointers into file buffer*/
        CHAR *pBufNext;
        CHAR *pBufCurr;

        /*Buffer pointer*/
        CHAR *pInBuf;

        /*Options specified by client*/
        UINT BufSize;

        /*File handle for buffering*/
        INT Handle;
        FILE *f;
        UINT32 BufPos;

        /*Operating system function to read bytes from a file*/
        LineFile_ReadFn *pReadFn;

} LINEFILE_CONTEXT;

module_scope LINEFILE_CONTEXT gLineFile;

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/*      This function allows the client to specify the size of the      */
/*      file buffer, plus the size of the blocks preceding and          */
/*      following the buffer.  The function reports whether the         */
/*      module was able to prepare the requested buffer successfully.   */
/*                                                                      */
/************************************************************************/
public_scope BOOL
LineFile_Start(UINT PreBytes, UINT BufBytes, UINT PostBytes)
{
        TRACE("\nLineFile_Start()");

        /*Can the environment cope with the total block size requested?*/
        if ((((UINT32) BufBytes) + ((UINT32) PreBytes) + ((UINT32) PostBytes))
                        > BLOCK_SIZE_MAX) {
                /*No, cannot serve request with available resources*/
                return FALSE;
        }

        /*Remember buffer size requested by client*/
        gLineFile.BufSize = BufBytes;

        /*Acquire space for file buffer*/
        gLineFile.pInBuf = PreBytes + ((CHAR *) malloc(PreBytes +
                        BufBytes + PostBytes));

        /*Did we acquire the buffer memory?*/
        if (gLineFile.pInBuf == NULL) {
                /*No, unable to start*/
                return FALSE;
        }

        /*Started successfully*/
        return TRUE;

} /*Start*/

/************************************************************************/
/*                                                                      */
/*      Open -- Provide optimised access using specified file           */
/*                                                                      */
/*      This function closes any previous file that was in use and      */
/*      attempts to open the specified file for optimised access.       */
/*      If the filename parameter is a NULL pointer, then standard      */
/*      input is used.                                                  */
/*                                                                      */
/*      Returns FALSE if the file was not able to be opened.            */
/*                                                                      */
/************************************************************************/
public_scope BOOL
LineFile_Open(CHAR *pFilename)
{
        TRACE1("\nLineFile_Open(%s)", pFilename);

        /*Does the client wish to use stdin?*/
        if (pFilename == NULL) {
                /*Yes, use stdin handle instead of opening another*/
                gLineFile.f = stdin;

        } else {

                /*No, was a file open before?*/
                if (gLineFile.f != NULL) {
                        /*Yes, close it now*/
                        fclose(gLineFile.f);
                        gLineFile.f = NULL;
                }

                /*Attempt to open file specified by client*/
                gLineFile.f = fopen(pFilename, "rb");
                if (gLineFile.f == NULL) {
                        /*Open failed*/
                        return FALSE;
                }

        }


        gLineFile.Handle = fileno(gLineFile.f);

        /*Set up buffer to show no data present*/
        gLineFile.pBufCurr = gLineFile.pInBuf;
        gLineFile.pBufNext = gLineFile.pInBuf;
        *gLineFile.pBufCurr = NUL;

        /*Report success to caller*/
        return TRUE;

} /*Open*/

/************************************************************************/
/*                                                                      */
/*      RemovePartialLastLine -- Simplify buffer/line relationship      */
/*                                                                      */
/*      Part of the search optimisation is to simplify the file into    */
/*      lines that sit in contiguous memory.  Where a file is read      */
/*      into a buffer, it is likely that the end of the buffer will     */
/*      occur in the middle of a line.  This case is extremely messy    */
/*      to handle.                                                      */
/*                                                                      */
/*      This function gets rid of the messy case by finding the         */
/*      LF that terminates the last complete line of the buffer,        */
/*      and deleting the partial line that was present.  The file       */
/*      is also repositioned so that the discarded partial line is      */
/*      intact and processed as the first line of the next buffer.      */
/*                                                                      */
/*      If the buffer does not contain a LF at all, this function       */
/*      appends one and returns -1 to extend the buffer to include      */
/*      the appended character.  Although this is somewhat dubious      */
/*      behaviour, in practice it is a good engineering compromise      */
/*      as long as the buffer is larger than the longest expected       */
/*      line.                                                           */
/*                                                                      */
/************************************************************************/
module_scope INT
LineFile_RemovePartialLastLine(CHAR *pLine)
{
        CHAR *pNextEOL = pLine + 1;

        TRACE("\nLineFile_RemovePartialLastLine()");

        /*Use memlchr to search for last LF in buffer*/
        pNextEOL = (CHAR *) memlchr(pLine, LF, gLineFile.BufSize);

        /*Did we find any line break characters?*/
        if (pNextEOL == NULL) {
                /*No, append LF and report -1 to keep buffer consistent*/
                *pLine = LF;
                return -1;
        }

        /*Rewind file to start with partial line*/
        fseek(gLineFile.f, -(pLine - pNextEOL), SEEK_CUR);

        /*Probably should use ANSI fgetpos/fsetpos instead of fseek*/
        /*Also, using fseek might be risky on some systems (pipes?)*/

        /*Report how many characters were lopped off the end*/
        return (INT) (pLine - pNextEOL);

} /*RemovePartialLastLine*/

/************************************************************************/
/*                                                                      */
/*      GetLine -- Find line in buffer without lotsa overhead           */
/*                                                                      */
/*      Getline returns a pointer to the next line of the file, or      */
/*      NULL if the file is finished.  The line is NUL terminated.      */
/*                                                                      */
/************************************************************************/
public_scope CHAR *
LineFile_GetLine(void)
{
        CHAR *pLine = gLineFile.pBufNext;
        CHAR *pNextEOL;
        UINT NrChars;

        TRACE("\nLineFile_GetLine()");

        /*Is the buffer empty?*/
        if (*pLine == NUL) {
                /*Yes, acquire some more data from the operating system*/
                goto FillBuffer;
        }

NextLine:
        /*Find end of next line in buffer and remove CR if present*/
        pNextEOL = memchr(pLine, LF, 0xffff);
        if (pNextEOL == NULL) {
                /*Buffer is almost certainly not a text file: abandon*/
                printf("?? File does not have enough LFs to search?!");
                return NULL;
        }
        if (pNextEOL[-1] == CR) {
                pNextEOL[-1] = NUL;
        }

        /*Replace LF with NUL*/
        *pNextEOL = NUL;

        /*Prefix line with LF in case EasiestFirst searches for it*/
        pLine[-1] = LF;

        /*Remember where next line starts*/
        gLineFile.pBufNext = pNextEOL + 1;

        /*And report next line to client*/
        gLineFile.pBufCurr = pLine;
        return pLine;

FillBuffer:
        /*Remember file position of start of buffer*/
        gLineFile.BufPos = tell(gLineFile.Handle);

        /*Attempt to fill buffer from file*/
        NrChars = gLineFile.pReadFn(gLineFile.Handle,
                        gLineFile.pInBuf, gLineFile.BufSize);

        /*Did we get any more from the file?*/
        if (NrChars == 0) {
                /*No, finished file data*/
                return NULL;

        }

        /*Was the file exhausted during that read?*/
        if (NrChars < gLineFile.BufSize) {
                /*Yes, did the last line end with CTRL/Z (ASCII SUB)?*/
                pLine = gLineFile.pInBuf;
                if (pLine[NrChars - 1] == ASCII_SUB) {
                        /*Yes, remove it*/
                        NrChars--;
                }

                /*Is there any text at the end without trailing LF?*/
                if (pLine[NrChars - 1] != LF) {
                        /*Yes, append LF to keep match simple (a bit dodgy)*/
                        pLine[NrChars++] = LF;
                }

        } else {
                NrChars -= LineFile_RemovePartialLastLine(
                                &gLineFile.pInBuf[gLineFile.BufSize - 1]);

                pLine = gLineFile.pInBuf;

        }

        /*Set up buffer for operation and continue operation*/
        pLine[-1] = LF;
        pLine[NrChars] = NUL;

        goto NextLine;

} /*GetLine*/

/************************************************************************/
/*                                                                      */
/*      GetBuffer -- Fetch next buffer of file (if any)                 */
/*                                                                      */
/*      This function retrieves a slab of file data and returns it      */
/*      in a contiguous memory space.  The number of bytes read is      */
/*      returned, or 0 if the file is exhausted.                        */
/*                                                                      */
/*      Some of the more realistic operating systems make it easy       */
/*      to overlap reading the file during program operation            */
/*      (e.g. using DMA).  This overlap can be exploited by using       */
/*      multiple buffers in turn.  To cater for this, GetBuffer         */
/*      reports the buffer pointer on each call, even though this       */
/*      does not change on the original DOS implementation.             */
/*                                                                      */
/************************************************************************/
public_scope UINT
LineFile_GetBuffer(CHAR **ppBuf)
{
        UINT NrChars;

        TRACE("\nLineFile_GetBuffer()");

        /*Read in next part of file into buffer*/
        gLineFile.BufPos = tell(gLineFile.Handle);
        NrChars = gLineFile.pReadFn(gLineFile.Handle, gLineFile.pInBuf,
                        gLineFile.BufSize);

        /*Did we run out of file to read?*/
        if (NrChars == 0) {
                /*Yes, file finished*/
                *ppBuf = NIL;
                return 0;
        }

        /*Was the file exhausted during that read?*/
        if (NrChars < gLineFile.BufSize) {
                /*DOS (well, CP/M actually) hack: was the last char CTRL/Z?*/
                if (gLineFile.pInBuf[NrChars - 1] == ASCII_SUB) {
                        /*Yes, remove it*/
                        NrChars--;
                }

                /*Is there any text at the end without trailing LF?*/
                if (gLineFile.pInBuf[NrChars - 1] != LF) {
                        /*Yes, append LF to keep match simple (a bit dodgy)*/
                        gLineFile.pInBuf[NrChars++] = LF;
                }


        } else {
                /*No, remove partial line from end of buffer*/
                NrChars -= LineFile_RemovePartialLastLine(
                                &gLineFile.pInBuf[gLineFile.BufSize - 1]);

        }

        /*Report fetched characters to caller*/
        *ppBuf = gLineFile.pInBuf;
        return NrChars;

} /*GetBuffer*/

/************************************************************************/
/*                                                                      */
/*      Position -- Report file offset for specified buffer pointer     */
/*                                                                      */
/*      The search uses buffers for speed, but must relate the          */
/*      position in the buffer back to the file offset for reporting.   */
/*      This function performs the required mapping.                    */
/*                                                                      */
/************************************************************************/
public_scope UINT32
LineFile_Position(CHAR *pBufferPos)
{
        return gLineFile.BufPos + (pBufferPos - gLineFile.pInBuf);

} /*Position*/

/************************************************************************/
/*                                                                      */
/*      ReadFunction -- Receive function to read bytes from a file      */
/*                                                                      */
/*      LineFile assumes that the ANSI read function is sufficient      */
/*      and correct to fetch file lines.  If this assumption is         */
/*      incorrect for some reason, the client may supply an             */
/*      alternative function via this interface.                        */
/*                                                                      */
/*      This interface was included when behoffski found that           */
/*      the operation of read for pipes was incorrect,                  */
/*      whereas _read worked correctly under TopSpeed C v3.10.          */
/*                                                                      */
/************************************************************************/
public_scope void
LineFile_ReadFunction(LineFile_ReadFn *pReadFn)
{
        /*Remember function in private variable*/
        gLineFile.pReadFn = pReadFn;

} /*ReadFunction*/

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
LineFile_Init(void)
{
        TRACE("\nLineFile_Init()");

        /*Default to no buffer provided initially*/
        gLineFile.pInBuf = NIL;

        /*Initially use stdin as the file (??)*/
        gLineFile.f = stdin;
        gLineFile.Handle = fileno(gLineFile.f);

        /*Use ANSI-compliant read function*/
        gLineFile.pReadFn = read;

} /*Init*/
