/************************************************************************/
/*                                                                      */
/*      LineFile -- Provide low-cost line-based file access             */
/*                                                                      */
/*      LineFile provides a service by which a file can be              */
/*      retrieved line-by-line or buffer-by-buffer.  The module         */
/*      provides space for the client to add bytes to the start         */
/*      and end of the buffer.  The module is faster than               */
/*      the typical stdio line handing functions, for both              */
/*      line-based access but especially for buffer-based access.       */
/*                                                                      */
/*      The module contains some provision for double buffering         */
/*      the file as this improves performance, but requires             */
/*      a worthwhile operating system to provide the necessary          */
/*      file handling primitives.                                       */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#ifndef LINEFILE_H
#define LINEFILE_H

#include <compdef.h>

#ifdef __cplusplus
extern "C" {
#endif /*__cplusplus*/

/*Prototype for "read" function (ANSI-compatible)*/
typedef int (LineFile_ReadFn)(int Handle, void *pBuf, unsigned NrBytes);

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
void
LineFile_Init(void);

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/*      This function allows the client to specify the size of the      */
/*      file buffer, plus the size of the blocks preceding and          */
/*      following the buffer.  The function reports whether the         */
/*      module was able to prepare the requested buffer successfully.   */
/*                                                                      */
/************************************************************************/
BOOL
LineFile_Start(UINT PreBytes, UINT BufBytes, UINT PostBytes);

/************************************************************************/
/*                                                                      */
/*      ReadFunction -- Receive function to read bytes from a file      */
/*                                                                      */
/*      LineFile assumes that the ANSI read function is sufficient      */
/*      and correct to fetch file lines.  If this assumption is         */
/*      incorrect for some reason, the client may supply an             */
/*      alternative function via this interface.                        */
/*                                                                      */
/*      This interface was included when behoffski found that           */
/*      the operation of read for pipes was incorrect,                  */
/*      whereas _read worked correctly under TopSpeed C v3.10.          */
/*                                                                      */
/************************************************************************/
void
LineFile_ReadFunction(LineFile_ReadFn *pReadFn);

/************************************************************************/
/*                                                                      */
/*      Open -- Provide optimised access using specified file           */
/*                                                                      */
/*      This function closes any previous file that was in use and      */
/*      attempts to open the specified file for optimised access.       */
/*      If the filename parameter is a NULL pointer, then standard      */
/*      input is used.                                                  */
/*                                                                      */
/*      Returns FALSE if the file was not able to be opened.            */
/*                                                                      */
/************************************************************************/
BOOL
LineFile_Open(CHAR *pFilename);

/************************************************************************/
/*                                                                      */
/*      GetLine -- Find line in buffer without lotsa overhead           */
/*                                                                      */
/*      Getline returns a pointer to the next line of the file, or      */
/*      NULL if the file is finished.  The line is NUL terminated.      */
/*                                                                      */
/************************************************************************/
CHAR *
LineFile_GetLine(void);

/************************************************************************/
/*                                                                      */
/*      GetBuffer -- Fetch next buffer of file (if any)                 */
/*                                                                      */
/*      This function retrieves a slab of file data and returns it      */
/*      in a contiguous memory space.  The number of bytes read is      */
/*      returned, or 0 if the file is exhausted.                        */
/*                                                                      */
/*      Some of the more realistic operating systems make it easy       */
/*      to overlap reading the file during program operation            */
/*      (e.g. using DMA).  This overlap can be exploited by using       */
/*      multiple buffers in turn.  To cater for this, GetBuffer         */
/*      reports the buffer pointer on each call.                        */
/*                                                                      */
/************************************************************************/
UINT
LineFile_GetBuffer(CHAR **ppBuf);

/************************************************************************/
/*                                                                      */
/*      Position -- Report file offset for specified buffer pointer     */
/*                                                                      */
/*      The search uses buffers for speed, but must relate the          */
/*      position in the buffer back to the file offset for reporting.   */
/*      This function performs the required mapping.                    */
/*                                                                      */
/************************************************************************/
UINT32
LineFile_Position(CHAR *pBufferPos);

#ifdef __cplusplus
}
#endif /*__cplusplus*/

#endif /*LINEFILE_H*/
