/************************************************************************/
/*                                                                      */
/*      MatchEng -- Match engine interface and supporting definitions   */
/*                                                                      */
/*      This interface file describes the language used to specify      */
/*      how to perform the pattern match and to report match results.   */
/*      The language is closely related to the assembly-language        */
/*      match engine since the search outperforms conventional          */
/*      coding techniques only where the assembly is used.              */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#ifndef MATCHENG_H
#define MATCHENG_H

#include <compdef.h>
#include <limits.h>

#define MATCHENG_BOYER_MOORE_LENGTH     18

/*Actions to use while searching*/

/*NOTE: Assembly version of search determines values below*/

#define SKIP_MAX                        0x00  /*Skip max bytes*/
#define SKIP17                          0x07
#define SKIP16                          0x08
#define SKIP15                          0x09
#define SKIP14                          0x0a
#define SKIP13                          0x0b
#define SKIP12                          0x13
#define SKIP11                          0x14
#define SKIP10                          0x15
#define SKIP9                           0x1d
#define SKIP8                           0x1e
#define SKIP7                           0x26
#define SKIP6                           0x27
#define SKIP5                           0x2f
#define SKIP4                           0x30
#define SKIP3                           0x38
#define SKIP2                           0x39
#define AGAIN                           0x3a /*Next char, same table*/
#define ADVANCE                         0x40 /*Next char, next table*/
#define START_MATCH                     0x48 /*Start, prev ch, next tab*/
#define START_MATCH_PUSH                0x52 /*Save backtrack, start*/
#define START_MATCH_PUSH_ADVANCE        0x5e /*Optimise first match*/
#define AGAIN_PUSH_ADVANCE              0x6c /*Save backtrack, repeat table*/
#define BACK_AND_ADVANCE                0x7a /*Prev char but next table*/
#define SKIP_NEXT_TABLE                 0x82 /*Prev char, but table + 2*/
#define ADVANCE_PUSH_ZERO               0x8c /*Matched, but push 0 iter case*/
#define NO_MATCH                        0x98 /*No match: pop previous state*/
#define MARK_LINE_START                 0xa0 /*LF found in buffer*/
#define MARK_LINE_START_PUSH_ADVANCE    0xaa /*LF in buf: start anchor match*/
#define START_OFFSET_MATCH              0xbc /*Save opt srch, try RE match*/
#define BYTE_SEARCH                     0xcc /*Search for literal first byte*/
#define ABANDON                         0xf6 /*No match possible*/
#define COMPLETE_IF_WORD_START          0xf8 /*Success only if started word*/
#define SPECIAL                         0xfa /*Reserved for future expansion*/
#define COMPLETED                       0xfc /*Match completed*/

#define MATCHENG_ACTIONS_MAX            40 /*Actually 35 as of June 1997*/

/*Flags describing important table properties*/

#define ITERATION_FLAG                  BIT7
#define ZERO_TRIP_FLAG                  BIT6
#define TABLE_TYPE_MASK                 0x0f
#define TABLE_TYPE(Flag)                ((Flag) & TABLE_TYPE_MASK)
        #define FAIL_TYPE               0x00
        #define START_TYPE              0x01
        #define LITERAL_TYPE            0x02
        #define CLASS_TYPE              0x03
        #define ALL_BUT_TYPE            0x04
        #define MATCH_ANY_TYPE          0x05
        #define SUCCESS_TYPE            0x06
        #define THIS_IS_NOT_A_TYPE      0x07

#define TABLE_SIZE                      (UCHAR_MAX + 1)

typedef BYTET TABLE[TABLE_SIZE];

/*Layout of expanded state tables optimised for speed*/
/*WARNING: Layout of this structure intimately known by assembly version!*/

struct MatchEng_Spec {
        /*Note: Flags summarising table entries work backwards from here*/

        /*Offset for Boyer-Moore search optimisation*/
        BYTET SearchSkipOffset;

        /*Array of tables to store actions per byte for each state*/
        TABLE Tables[99];

        /*Note: Storage is determined dynamically (99 is for compiler only)*/

};
typedef struct MatchEng_Spec MATCHENG_SPEC;

/*Flags to select reporting options*/

#define DETAILS_FILENAME                BIT0
#define DETAILS_LINENUMBER              BIT1
#define DETAILS_BYTEOFFSET              BIT2
#define DETAILS_LINECOUNT               BIT3
#define DETAILS_HIGHLIGHT               BIT4
#define DETAILS_SUPPRESS_FILENAME       BIT5
#define DETAILS_NONMATCH_FILES          BIT6
#define DETAILS_LINE                    BIT7
#define DETAILS_MARKER_FLAG             BIT8
#define DETAILS_MARKER(Ch)              (BIT8 | (((LWORD) (Ch)) << 24))
#define DETAILS_MARKER_UNPACK(Flags)    ((CHAR) ((Flags) >> 24))

/*Structure for match engine to report details of successful match*/

typedef struct {
        /*Note: Layout of first part of structure hard-coded into assembly!*/

        /*Match details*/
        CHART *pMatchStart;
        CHART *pMatchEnd;

        /*Line details (written by match if buffer search used)*/
        UINT16T BufLineNr;
        CHART *pLineStart;

        /*End of buffer pointer to simplify CPU string search instruction*/
        CHART *pAfterEndOfBuffer;

        /*Warning: Variables above this line referenced by assembly!*/

        /*Other information about match used for reporting*/
        LWORD ReportingOptions;
        CHAR *pFilename;
        UINT32 LineNr;

        /*Match counter: only valid if selection function supports it*/
        UINT32 LineMatchCount;

        /*Character for delimited output*/
        CHAR MarkerChar;

} MATCHENG_DETAILS;

/************************************************************************/
/*                                                                      */
/*      Prototype for select action by search engine                    */
/*                                                                      */
/*      In order to keep the code simple and portable (and also to      */
/*      improve the speed of the code in some cases), the function      */
/*      that acts on selected lines is configured via a variable.       */
/*      This prototype describes the format of the function call.       */
/*      The function returns TRUE if search is to continue or FALSE     */
/*      if the search can be abandoned.                                 */
/*                                                                      */
/************************************************************************/
typedef BOOL
(SELECTION_FUNCTION)(MATCHENG_DETAILS *pDetails);

/************************************************************************/
/*                                                                      */
/*      Prototype for match action by search engine                     */
/*                                                                      */
/*      In order to improve the portability of the code (and also to    */
/*      allow some trickery to optimise match operation), the           */
/*      match function is configured via a variable.  This prototype    */
/*      describes the calling sequence for the function.                */
/*                                                                      */
/*      You might be wondering why this function type is declared       */
/*      as a pointer while SELECTION_FUNCTION isn't.  The reason        */
/*      is that behoffski doesn't like formalising "pointer to..."      */
/*      as an explicit type, and this includes function definitions.    */
/*      However, the PAGE_ALIGNED attribute vital to the assembly       */
/*      version needs to be attached to the type: declaring the         */
/*      function type as a pointer here is the only reliable way        */
/*      to do it.                                                       */
/*                                                                      */
/************************************************************************/
typedef BOOL
(PAGE_ALIGNED *MATCH_FUNCTION)(MATCHENG_SPEC *pTable,
                CHAR *pTextStart,
                MATCHENG_DETAILS *pDetails);

#endif /*MATCHENG_H*/
