/************************************************************************/
/*                                                                      */
/*      RegExp -- Create and manipulate regular expressions             */
/*                                                                      */
/*      This module compresses the ASCII text form of regular           */
/*      expressions into a much more compact and manageable form.       */
/*      The compact form can be used to drive a match engine,           */
/*      or can be expanded where table-driven searching is used.        */
/*                                                                      */
/*      The encoding of elements could be made more compact and         */
/*      convenient by encoding the RE modifiers in bits 15 and 14       */
/*      of the code word as follows:                                    */
/*         BIT15/BIT14  Code                                            */
/*              00      Normal code, not optional                       */
/*              01      "?" modifier -- 0 or 1 occurrences              */
/*              10      "+" modifier -- 1 or more times                 */
/*              11      "*" modifier -- 0 or more times                 */
/*                                                                      */
/*      Note that some information about the search, such as            */
/*      case insensitivity, match word and match entire line,           */
/*      should be specified via this module but have instead            */
/*      been deferred to RETable as they are easier to                  */
/*      handle in the table-driven form.                                */
/*                                                                      */
/*      The code could also be improved by separating the 16-word       */
/*      bitmasks accompanying each class specifier into a separate      */
/*      array.  This would allow the RE encoding to be traversed        */
/*      backwards as well as forwards.  Requires more memory            */
/*      management than existing sequential scheme.                     */
/*                                                                      */
/*      Some portions of this code were derived after reading           */
/*      GNU Sed v1.x by Eric S. Raymond (modified by Hern Chen).        */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include "ascii.h"
#include "bakslash.h"
#include <compdef.h>
#include <memory.h>
#include "regexp.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/*Codes summarising why operations failed*/
typedef enum {
        FAIL_NO_FAILURE,
        FAIL_REGEXP_TOO_LARGE,
        FAIL_UNMATCHED_CLASS_OPEN,
        FAIL_NOT_ENOUGH_MEMORY,
        FAIL_RE_IS_OPTIMAL

} FAILURE_CODE;

/*Rough character frequency string to hopefully improve EasiestFirst*/

module_scope const char RegExp_Likely[] =
                " e*tirnoas-dlc/pfuhm._T0)SE(CRgA;I,DO1bNLyv=PM2wkBxFU:34#HGW>95\"6|8{}V7+z\\Y&K'X<[]qjQ!JZ%?~@$^`";

/*Fairly rough byte frequencies for estimating search cost*/

module_scope UINT RegExp_ByteStats[256] = {
         2358,   252,   196,   345,   159,   164,   232,   143, /*  0-  7*/
          225,  1468,  7426,    93,   119,  6618,   123,   242, /*  8- 15*/
          110,    61,   106,    65,    84,    76,    90,    52, /* 16- 23*/
           73,    52,    97,   130,    66,    49,    89,    72, /* 24- 31*/
       65535u,   248,   994,   374,   278,   366,   485,   651, /* 32- 39*/
         1380,  1362,  4321,   370,  2892,  3537,  4792,  2544, /* 40- 47*/
         8016,  4254,  3861,  2130,  2600,  1925,  1866,  1569, /* 48- 55*/
         2066,  2143,  1295,  1019,   759,  1038,   837,   198, /* 56- 63*/
          247,  2432,  1170,  2261,  1831,  2773,  1926,   927, /* 64- 71*/
          846,  2201,   269,   411,  1373,  1404,  1468,  1712, /* 72- 79*/
         1950,   226,  2075,  2469,  2169,   801,   641,   701, /* 80- 87*/
          414,   412,   627,   525,   775,   449,   284,  1557, /* 88- 95*/
           70,  9743,  2214,  4936,  5133, 15939,  3254,  2665, /* 96-103*/
         4340,  9307,   387,  1206,  5902,  3765,  8797,  9854, /*104-111*/
         3935,   260,  9408,  8591, 11138,  4156,  1738,  1756, /*112-119*/
         1168,  2075,   518,   327,   346,   334,   255,    75, /*120-127*/
          134,    71,    99,   243,    88,    59,   113,    66, /*128-135*/
           88,   243,    92,   368,   109,    91,   121,    52, /*136-143*/
           86,    49,    60,    47,    56,    44,    63,    53, /*144-151*/
           62,    66,   219,    70,    59,    45,    65,    55, /*152-159*/
           60,    57,    56,    55,    51,    49,    50,    45, /*160-167*/
           59,    48,    70,    52,    56,    53,    59,    58, /*168-175*/
           53,    48,    55,    51,    56,    40,    66,    52, /*176-183*/
          127,    60,    70,    55,    62,    44,    61,    52, /*184-191*/
           76,    68,    56,    58,   326,    51,    69,   131, /*192-199*/
           91,    68,    70,    67,    63,   100,    58,    48, /*200-207*/
           60,    61,    82,    49,    49,    44,    51,    44, /*208-215*/
          109,    52,    68,    61,    66,    54,    62,    80, /*216-223*/
           75,    45,    68,    48,    63,    47,    71,    49, /*224-231*/
           88,   260,    65,    66,    95,    48,    78,    52, /*232-239*/
           90,    56,    78,    45,    88,    42,    89,    62, /*240-247*/
          100,    67,   104,   202,   109,    71,   110,   660  /*248-255*/
};

typedef struct {
        /*Configure how module should generate compact RE*/
        LWORD Config;

        /*Reason for last failure*/
        FAILURE_CODE FailCode;

} REGEXP_MODULE_CONTEXT;

module_scope REGEXP_MODULE_CONTEXT gRegExp;

/*Array to optimise converting classes to bitsets*/
module_scope const WORD RegExp_bits[] = {BIT0, BIT1, BIT2, BIT3, BIT4, BIT5, BIT6, BIT7,
                BIT8, BIT9, BIT10, BIT11, BIT12, BIT13, BIT14, BIT15};

/************************************************************************/
/*                                                                      */
/*      ShowCodes -- Disassemble RE coding (for debugging)              */
/*                                                                      */
/************************************************************************/
public_scope void
RegExp_ShowCodes(REGEXP_SPEC *pRE)
{
        RE_ELEMENT *pElem = pRE->CodeList;
        UINT CodeIx;

        /*Display information carried around about RE*/
        printf("NrStates:%u ", pRE->NrStates);

        /*Loop through each code element*/
        for (;;) {
                /*Display element*/
                switch (*pElem++) {
                case REGEXP_MATCH_ANY:
                        printf(" _ANY ");
                        break;

                case REGEXP_ZERO_OR_MORE:
                        printf(" _0ORMORE ");
                        break;

                case REGEXP_ONE_OR_MORE:
                        printf(" _1ORMORE ");
                        break;

                case REGEXP_ANCHOR_END:
                        printf(" _EOL ");
                        break;

                case REGEXP_MATCH_CLASS:
                        printf(" _CLASS[");
                        for (CodeIx = 0; CodeIx < REGEXP_BITMASK_SIZE;
                                        CodeIx++) {
                                printf("%04x", (WORD) *pElem++);
                        }
                        printf("] ");
                        break;

                case REGEXP_MATCH_CLASS_ALL_BUT:
                        printf(" _ALLBUT[");
                        for (CodeIx = 0; CodeIx < REGEXP_BITMASK_SIZE;
                                        CodeIx++) {
                                printf("%04x", (WORD) *pElem++);
                        }
                        printf("] ");
                        break;

                case REGEXP_ANCHOR_START:
                        printf(" _ANCHOR ");
                        break;

                case REGEXP_END_OF_EXPR:
                        printf(" _END ");
                        return;
                        break;

                case REGEXP_ZERO_OR_ONE:
                        printf(" 0OR1 ");
                        break;

                default:
                        /*Is the code a literal?*/
                        if (pElem[-1] > 255) {
                                /*No, show code and confusion*/
                                printf("??%04x ", pElem[-1]);
                                break;

                        }

                        /*Presumably literal: display it directly (?)*/
                        printf("%c", (CHAR) pElem[-1]);
                        break;

                }
        }

} /*ShowCodes*/

/************************************************************************/
/*                                                                      */
/*      Compact -- Render text expression in compact form               */
/*                                                                      */
/*      This function is the first phase of handling a text regular     */
/*      expression.  It does a fairly simple and fast parse of the      */
/*      text to produce a set of codes that describe each element       */
/*      of the expression in a single item.  In particular, this        */
/*      means that multi-character specifiers such as "[A-Za-z]"        */
/*      are packaged into a convenient code.                            */
/*                                                                      */
/************************************************************************/
module_scope BOOL
RegExp_Compact(CHAR *pText, REGEXP_SPEC *pRE)
{
        BYTE ch;
        BYTE PrevCh;
        RE_ELEMENT *pCode = pRE->CodeList;
        BOOL Range = FALSE;
        BOOL MatchAnyImplied = TRUE;
        UINT NrStates = 0;
        UINT NrClassMembers;
        BOOL FirstLiteral;

        /*Is the RE anchored to the start of the space?*/
        if (*pText == '^') {
                /*Yes, insert anchor element and skip over char*/
                *pCode++ = REGEXP_ANCHOR_START;
                pText++;

        /*No anchor: was one selected by an option anyway?*/
        } else if (gRegExp.Config & REGEXP_MATCH_FULL_LINE_ONLY) {
                /*Yes, insert specifier so RE matches correctly*/
                *pCode++ = REGEXP_ANCHOR_START;

        }

        /*Work through each part of expression*/
ConversionLoop:

        /*Do we have space to store the next RE element?*/
        if ((pCode - pRE->CodeList) >= (REGEXP_CODES_MAX -
                        REGEXP_BITMASK_SIZE)) {
                /*No, give up in disgust*/
                gRegExp.FailCode = FAIL_REGEXP_TOO_LARGE;
                return FALSE;
        }

        /*Presume that specifier creates a match state*/
        NrStates++;

        ch = (BYTE) *pText++;
        switch ((CHAR) ch) {
        case '.':
                /*Match anything except EOL*/
                *pCode++ = REGEXP_MATCH_ANY;
                MatchAnyImplied = FALSE;
                break;

        case '*':
                /*Repeat 0 or more times*/
                if (MatchAnyImplied) {
                        *pCode++ = REGEXP_MATCH_ANY;
                }
                *pCode++ = REGEXP_ZERO_OR_MORE;
                MatchAnyImplied = TRUE;
                NrStates--;
                break;

        case '+':
                /*Repeat one or more times*/
                if (MatchAnyImplied) {
                        *pCode++ = REGEXP_MATCH_ANY;
                }
                *pCode++ = REGEXP_ONE_OR_MORE;
                MatchAnyImplied = TRUE;
                break;

        case '?':
                /*Repeat zero or one times*/
                if (MatchAnyImplied) {
                        *pCode++ = REGEXP_MATCH_ANY;
                }
                *pCode++ = REGEXP_ZERO_OR_ONE;
                MatchAnyImplied = TRUE;
                NrStates--;
                break;

        case '$':
                /*Match end of line*/
                *pCode++ = REGEXP_ANCHOR_END;
                MatchAnyImplied = TRUE;
                break;

        case '\0':
                /*Were we asked to limit matches to full lines?*/
                if (gRegExp.Config & REGEXP_MATCH_FULL_LINE_ONLY) {
                        /*Yes, was an end-of-line anchor given?*/
                        if (pCode[-1] != REGEXP_ANCHOR_END) {
                                /*No, add one now*/
                                *pCode++ = REGEXP_ANCHOR_END;
                        }

                }

                /*End of expression... write result to caller and exit*/
                *pCode = REGEXP_END_OF_EXPR;
                pRE->NrStates = NrStates;
                return TRUE;
                break;

        case '\\':
                /*Escape code*/
                *pCode++ = (RE_ELEMENT) Bakslash_Decode(&pText);
                MatchAnyImplied = FALSE;
                break;

        case '[':
                /*Start of class of characters: is first char "^"?*/
                ch = (BYTE) *pText++;
                if (ch == ((BYTE) '^')) {
                        /*Yes, specifier is all but named characters*/
                        *pCode++ = REGEXP_MATCH_CLASS_ALL_BUT;
                        ch = (BYTE) *pText++;

                } else {
                        /*Normal class*/
                        *pCode++ = REGEXP_MATCH_CLASS;

                }

                /*Collect characters specified in a bit set*/
                memset(pCode, 0, REGEXP_BITMASK_SIZE * sizeof(RE_ELEMENT));
                NrClassMembers = 0;

                /*Is the first char ']'?*/
                if (ch == ((BYTE) ']')) {
                        /*Yes, don't treat it as the class terminator*/
                        ((WORD *) pCode)[ch >> 4] |= RegExp_bits[ch & 0x0f];
                        ch = (BYTE) *pText++;
                        NrClassMembers++;
                }

                /*Prepare range expansion variables*/
                PrevCh = ch;
                Range = FALSE;
                FirstLiteral = TRUE;

                /*Is the first character a '-'?*/
                if (ch == ((BYTE) '-')) {
                        /*Yes, it's a literal, not a range*/
                        ((WORD *) pCode)[ch >> 4] |= RegExp_bits[ch & 0x0f];
                        ch = (BYTE) *pText++;
                        NrClassMembers++;
                }

                /*Loop until end-of-set specifier found*/
                while (ch != ((BYTE) ']')) {
                        /*Is this the middle of a range specifier?*/
                        if ((ch == ((BYTE) '-')) && (*pText != ']')) {
                                /*Yes, remember that range was given*/
                                Range = TRUE;
                                ch = (BYTE) *pText++;
                        }

                        /*Is this an escape code?*/
                        if (ch == ((BYTE) '\\')) {
                                /*Yes, convert next character*/
                                ch = (BYTE) *pText++;
                                if (ch == ((BYTE) 'n')) {
                                        ch = (BYTE) LF;
                                } else if (ch == ((BYTE) 't')) {
                                        ch = (BYTE) TAB;
                                }
                        }

                        /*Has the range terminated abnormally?*/
                        if (ch == ((BYTE) NUL)) {
                                /*Yes, bad RE syntax*/
                                gRegExp.FailCode = FAIL_UNMATCHED_CLASS_OPEN;
                                return FALSE;
                        }

                        /*Was a range specified?*/
                        if (Range) {
                                /*Yes, expand it now (PrevCh, ch unsigned)*/
                                Range = FALSE;
                                for (; PrevCh < ch; PrevCh++) {
                                        ((WORD *) pCode)[PrevCh >> 4] |=
                                                RegExp_bits[PrevCh & 0x0f];
                                        NrClassMembers++;
                                }
                                NrClassMembers++;
                                FirstLiteral = FALSE;

                        }

                        /*Add next character to the class*/
                        ((WORD *) pCode)[ch >> 4] |= RegExp_bits[ch & 0x0f];
                        if ((PrevCh != ch) || FirstLiteral) {
                                NrClassMembers++;
                        }

                        /*Move to the next char of the RE*/
                        PrevCh = ch;
                        ch = (BYTE) *pText++;

                }

                /*Finished collecting class specifier*/
                MatchAnyImplied = FALSE;

                /*Was this an all-but class?*/
                if (pCode[-1] == REGEXP_MATCH_CLASS_ALL_BUT) {
                        /*Yes, NUL mustn't match: include in negatives*/
                        ((WORD *) pCode)[0] |= 0x0001;

                } else {
                        /*No, did the class only contain one member?*/
                        if (NrClassMembers == 1) {
                                /*Yes, convert from class entry to literal*/
                                pCode[-1] = PrevCh;
                                goto ConversionLoop;
                        }

                        /*Ensure NUL is not present*/
                        ((WORD *) pCode)[0] &= 0xfffe;

                }

                /*WARNING: The code above assumes the ASCII character set*/

                /*Skip over class elements in RE*/
                pCode += REGEXP_BITMASK_SIZE;

                break;

        default:
                /*Presumably a literal: it is its own code (unsigned!)*/
                *pCode++ = (RE_ELEMENT) ((BYTE) ch);
                MatchAnyImplied = FALSE;
                break;

        }

        goto ConversionLoop;

} /*Compact*/

/************************************************************************/
/*                                                                      */
/*      Compile -- Check and convert text expression to compact form    */
/*                                                                      */
/*      The function returns FALSE if the text form was badly           */
/*      formatted.                                                      */
/*                                                                      */
/************************************************************************/
public_scope BOOL
RegExp_Compile(CHAR *pText, REGEXP_SPEC **ppSpec)
{
        REGEXP_SPEC *pRE = NIL;

        /*Destroy return value to reduce chances of being misunderstood*/
        *ppSpec = NIL;

        /*Allocate a specifier/full structure block*/
        pRE = (REGEXP_SPEC *) malloc(sizeof(REGEXP_SPEC));
        if (pRE == NULL) {
                /*Unable to obtain memory for compilation*/
                gRegExp.FailCode = FAIL_NOT_ENOUGH_MEMORY;
                return FALSE;
        }

        /*Work through string, converting each portion to compact form*/
        if (! RegExp_Compact(pText, pRE)) {
                /*Errors in RE syntax... discard buffer and report failure*/
                free(pRE);
                return FALSE;
        }

        /*Write handle to caller and report success*/
        *ppSpec = pRE;
        return TRUE;

} /*Compile*/

/************************************************************************/
/*                                                                      */
/*      CompileString -- Convert string without looking for RE syntax   */
/*                                                                      */
/*      The function returns FALSE if the conversion failed.            */
/*                                                                      */
/************************************************************************/
public_scope BOOL
RegExp_CompileString(CHAR *pText, REGEXP_SPEC **ppSpec)
{
        REGEXP_SPEC *pRE = NIL;
        RE_ELEMENT *pCode;

        /*Destroy return value to reduce chances of being misunderstood*/
        *ppSpec = NIL;

        /*Allocate a specifier/full structure block*/
        pRE = (REGEXP_SPEC *) malloc(sizeof(REGEXP_SPEC));
        if (pRE == NULL) {
                /*Unable to obtain memory for compilation*/
                gRegExp.FailCode = FAIL_NOT_ENOUGH_MEMORY;
                return FALSE;
        }
        pCode = pRE->CodeList;

        /*Convert each character to code*/
        for (; *pText != NUL; pText++) {
                /*Do we have space for this character?*/
                if ((pCode - pRE->CodeList) >= REGEXP_CODES_MAX) {
                        /*No, cannot complete compilation satisfactorily*/
                        gRegExp.FailCode = FAIL_REGEXP_TOO_LARGE;
                        free(pRE);
                        return FALSE;
                }

                /*Convert this character*/
                *pCode++ = (RE_ELEMENT) (BYTE) *pText;

        }

        /*Add end-of-list descriptor*/
        *pCode = REGEXP_END_OF_EXPR;
        pRE->NrStates = (UINT) (pCode - pRE->CodeList) + 1;

        /*Write handle to caller and report success*/
        *ppSpec = (REGEXP_SPEC *) pRE;
        return TRUE;

} /*CompileString*/

/************************************************************************/
/*                                                                      */
/*      CountStates -- Work out how many states are in RE               */
/*                                                                      */
/*      The number of states in a regular expression is a very          */
/*      useful piece of information.  Most importantly, it allows       */
/*      the table-driven expansion of the RE to use the least memory    */
/*      possible.  However, some RE operations such as the split        */
/*      in EasiestFirst modify this value.  Rather than include         */
/*      counting states in each operation, this function surveys        */
/*      the RE and builds the correct count of states.                  */
/*                                                                      */
/************************************************************************/
module_scope void
RegExp_CountStates(REGEXP_SPEC *pRE)
{
        RE_ELEMENT *pCode = pRE->CodeList;
        UINT NrStates = 0;

        /*Loop through spec and interpret each code*/
        for (;;) {
                switch (*pCode++) {
                case REGEXP_ZERO_OR_ONE:
                case REGEXP_ZERO_OR_MORE:
                case REGEXP_ANCHOR_START:
                        /*These don't get states in their own right*/
                        break;

                case REGEXP_END_OF_EXPR:
                        /*Code terminating RE found: write states and exit*/
                        pRE->NrStates = NrStates + 1;
                        return;
                        break;

                case REGEXP_MATCH_CLASS:
                case REGEXP_MATCH_CLASS_ALL_BUT:
                        /*Next REGEXP_BITMASK_SIZE codes are class members*/
                        pCode += REGEXP_BITMASK_SIZE;
                        /*FALLTHROUGH*/

                default:
                        /*Most codes are counted as states*/
                        NrStates++;
                        break;

                }

        }

} /*CountStates*/

/************************************************************************/
/*                                                                      */
/*      Score -- Estimate likelihood of finding string                  */
/*                                                                      */
/*      This function inspects the supplied string of pattern           */
/*      elements and estimates how often the pattern would appear       */
/*      in "typical" files.  While this estimate is amazingly crude     */
/*      it is in fact worth while, since patterns like "e.*q"           */
/*      benefit from searching for "q" first instead of "e".            */
/*                                                                      */
/*      This function assumes that the pattern elements presented       */
/*      are already simple, such as literals and classes.               */
/*                                                                      */
/************************************************************************/
module_scope UINT
RegExp_Score(RE_ELEMENT *pCode, UINT NrCodes)
{
        UINT Score = 0;
        CHAR SearchStr[2];
        const CHAR *pLikely;
        UINT ClassScore;
        UINT c;

        /*Prepare to find characters in frequency string*/
        SearchStr[1] = 0;

        /*Loop through each code, estimating how often the code appears*/
        while (NrCodes-- != 0) {
                switch (*pCode++) {
                case REGEXP_MATCH_CLASS:
                        /*Examine class members to estimate likelihood*/
                        ClassScore = 0;
                        for (c = 0; c < 255; c++) {
                            /*Is this character present?*/
                            if (! (pCode[c >> 3] & (0x01 << (c & 0x07)))) {
                                /*No, does not reduce efficiency of match*/
                                continue;
                            }
                            ClassScore++;
                        }

                        /*Estimate likelihood merely by counting members*/
                        if (ClassScore > 16) {
                            Score += 1;
                        } else if (ClassScore > 8) {
                            Score += 2;
                        } else if (ClassScore > 4) {
                            Score += 5;
                        } else {
                            Score += 8;
                        }

                        /*Skip past bitset describing class members*/
                        pCode += REGEXP_BITMASK_SIZE;

                        /*NOTE: Class scoring should use member freqs*/
                        /*      e.g. [qj] should score higher than [ea]*/

                        break;

                default:
                        /*Code represents literal*/
                        SearchStr[0] = (CHAR) pCode[-1];
                        pLikely = strstr(RegExp_Likely, SearchStr);
                        if (pLikely == NULL) {
                                /*Character isn't likely... did not appear!*/
                                Score += strlen(RegExp_Likely);
                        } else {
                                /*Position relates to likelihood*/
                                Score += (UINT) (pLikely - RegExp_Likely);
                        }
                        break;

                }

        }

        return Score;

} /*Score*/

/************************************************************************/
/*                                                                      */
/*      EasiestFirst -- Improve search by deferring complex starts      */
/*                                                                      */
/*      An expression like "[^a]*.*[^b]*.*c.*Quasimodo's Dream.*g"      */
/*      is very expensive to search if you simply use the expression    */
/*      as given, as the first elements involve lots of backtracking.   */
/*      However, most regular expressions like this have a              */
/*      relatively easy middle bit (e.g. "Quasimodo's Dream") that      */
/*      is required for the match and can be searched first instead.    */
/*      This function determines whether this optimisation is worth     */
/*      while.  If not, the function returns FALSE.                     */
/*                                                                      */
/*      If the search can be optimised by the split, EasiestFirst       */
/*      splits the original RE so that it matches the complicated       */
/*      starting bit, and creates a new RE spec which begins with       */
/*      the easiest bit.                                                */
/*                                                                      */
/*      A rather crude scoring system is used as a tie-breaker          */
/*      where multiple longest strings are the same length.             */
/*      This tie-breaker attempts to select the string containing       */
/*      characters less likely to occur in "typical" text files.        */
/*                                                                      */
/************************************************************************/
public_scope BOOL
RegExp_EasiestFirst(REGEXP_SPEC *pRE, REGEXP_SPEC **ppEasierBit)
{
        RE_ELEMENT *pCode;
        RE_ELEMENT *pEasyCode = pRE->CodeList;
        UINT NrEasyCodes = 0;
        RE_ELEMENT *pEasiestCode = pRE->CodeList;
        RE_ELEMENT *pFirstIter = NULL;
        UINT Longest = 0;
        BOOL Tricky;
        BOOL Scanning = TRUE;
        BOOL AnchorStart = FALSE;
        BOOL AnchorEnd = FALSE;
        REGEXP_SPEC *pRE2;

        /*Destroy return value to reduce chances of being misinterpreted*/
        *ppEasierBit = NIL;

        /*Step through RE, looking for literals*/
        pCode = pRE->CodeList;
        do {
                /*Decode the next RE element*/
                switch (*pCode++) {
                case REGEXP_MATCH_CLASS_ALL_BUT:
                        /*Next REGEXP_BITMASK_SIZE codes are class members*/
                        pCode += REGEXP_BITMASK_SIZE;
                        /*FALLTHROUGH*/

                case REGEXP_ONE_OR_MORE:
                        /*Have we seen iterative elements before?*/
                        if (pFirstIter == NULL) {
                                /*No, remember this as the first*/
                                pFirstIter = pCode - 1;
                        }
                        /*FALLTHROUGH*/

                case REGEXP_MATCH_ANY:
                        /*These are not easy*/
                        Tricky = TRUE;
                        break;

                case REGEXP_ANCHOR_START:
                        /*Start anchor can perform poorly sometimes*/
                        AnchorStart = TRUE;
                        Tricky = TRUE;
                        break;

                case REGEXP_ANCHOR_END:
                        /*Skip forward to the end of the expression*/
                        AnchorEnd = TRUE;
                        while (*pCode++ != REGEXP_END_OF_EXPR) {
                        }
                        /*FALLTHROUGH*/

                case REGEXP_END_OF_EXPR:
                        /*Terminate any easy bit and prepare to exit*/
                        Tricky = TRUE;
                        Scanning = FALSE;
                        break;

                case REGEXP_ZERO_OR_MORE:
                        /*Have we seen iterative elements before?*/
                        if (pFirstIter == NULL) {
                                /*No, remember this as the first*/
                                pFirstIter = pCode - 1;
                        }
                        /*FALLTHROUGH*/

                case REGEXP_ZERO_OR_ONE:
                        /*These are not easy AND they modify the prev code*/
                        Tricky = TRUE;
                        if (NrEasyCodes > 0) {
                                /*Last easy element has complications*/
                                NrEasyCodes--;
                        }
                        break;

                default:
                        /*Literal or (hopefully easy) class*/
                        Tricky = FALSE;
                        if (NrEasyCodes++ == 0) {
                                /*Remember start in case it's the longest*/
                                pEasyCode = pCode - 1;
                        }

                        if (pCode[-1] == REGEXP_MATCH_CLASS) {
                                /*Next REGEXP_BITMASK_SIZE codes are class members*/
                                pCode += REGEXP_BITMASK_SIZE;
                        }
                        break;

                }

                /*Did we see an anchor at the end of the RE?*/
                if (AnchorEnd) {
                        /*Yes, use last easy bit to reduce backtracking*/
                        pEasiestCode = pEasyCode;
                        break;
                }

                /*Did we find the end of some easy codes?*/
                if ((! Tricky) || (NrEasyCodes == 0)) {
                        /*No, continue search*/
                        continue;
                }

                /*Yes, is the run longer than the best previous?*/
                if (NrEasyCodes < Longest) {
                        /*No, ignore it*/
                        NrEasyCodes = 0;
                        continue;

                }

                /*Is this string the same length as the best so far?*/
                if ((Longest != 0) && (NrEasyCodes == Longest)) {
                        /*Yes, does new string seem less likely to appear?*/
                        if (RegExp_Score(pEasyCode, NrEasyCodes) <
                                        RegExp_Score(pEasiestCode, Longest)) {
                                /*No, not worth changing selection*/
                                NrEasyCodes = 0;
                                continue;
                        }

                }


                /*String just found seems to be the easiest to search*/
                pEasiestCode = pEasyCode;
                Longest = NrEasyCodes;
                NrEasyCodes = 0;

        } while (Scanning);

        /*Code exploits knowledge that pCode points to end of RE (+ 1)*/

        /*Did we find a good spot to start in the middle?*/
        if (pEasiestCode == pRE->CodeList) {
                /*No, don't muck around with RE*/
                gRegExp.FailCode = FAIL_RE_IS_OPTIMAL;
                return FALSE;
        }

        /*Is the search anchored and doesn't have a long string?*/
        if (AnchorStart && (Longest < 3)) {
                /*Yes, state tables do a good job of this already*/
                gRegExp.FailCode = FAIL_RE_IS_OPTIMAL;
                return FALSE;
        }

        /*Did we see any iteration before easiest bit?*/
        if ((pFirstIter == NULL) || (pFirstIter > pEasiestCode)) {
                /*No, splitting the RE costs more than it saves*/
                gRegExp.FailCode = FAIL_RE_IS_OPTIMAL;
                return FALSE;
        }

        /*Note: The definition of "good" used above isn't optimal*/

        /*Get buffer to store easier part of RE*/
        pRE2 = (REGEXP_SPEC *) malloc(sizeof(REGEXP_SPEC));
        if (pRE2 == NULL) {
                /*Unable to obtain memory to perform split*/
                gRegExp.FailCode = FAIL_NOT_ENOUGH_MEMORY;
                return FALSE;
        }

        /*Break RE into two by copying easy bit into new RE spec*/
        memcpy(pRE2->CodeList, pEasiestCode,
                        sizeof(RE_ELEMENT) * (UINT) (pCode - pEasiestCode));
        RegExp_CountStates(pRE2);

        /*Write easier spec back to caller*/
        *ppEasierBit = pRE2;

        /*Edit original RE to be anchored at end of the line*/
        *pEasiestCode++ = REGEXP_ANCHOR_END;
        *pEasiestCode = REGEXP_END_OF_EXPR;
        RegExp_CountStates(pRE);

        /*Tell caller that optimisation has been performed*/
        return TRUE;

} /*EasiestFirst*/

#if 0 /*Not implemented yet*/
/************************************************************************/
/*                                                                      */
/*      AnchorToLiteral -- Change starting anchor to LF literal         */
/*                                                                      */
/*      Treating the anchor at the start of a RE as a literal           */
/*      allows various search optimisations to work more effectively.   */
/*      This function looks for a start anchor, and converts it         */
/*      to a LF literal if found.                                       */
/*                                                                      */
/************************************************************************/
public_scope void
RegExp_AnchorToLiteral(REGEXP_SPEC *pRE)
{
        /*Do we have an anchor to the start of the RE?*/
        if (pRE->CodeList[0] == REGEXP_ANCHOR_START)
        {
                /*Yes, rewrite it as LF literal*/
                pRE->CodeList[0] = (RE_ELEMENT) LF;
        }

} /*AnchorToLiteral*/
#endif /*0*/

/*Flags summarising state of search for optional elements*/

#define REAL_ELEM                       BIT0
#define OPTIONAL                        BIT1
#define END_OPT                         BIT2
#define STOP_SCAN                       BIT3

/************************************************************************/
/*                                                                      */
/*      FindOptional -- Find sequence of optional elements in RE        */
/*                                                                      */
/*      This function assists SlashEnds by locating a string of         */
/*      optional elements, if present, in the RE.  The function         */
/*      starts from a nominated position in the RE, and reports         */
/*      the start and code after the optional bit, so that the          */
/*      function can work through all the optional components of        */
/*      the RE in order.                                                */
/*                                                                      */
/*      ppFirstCode is an in/out parameter: it names where to start     */
/*      and reports the start of the optional bit, if found.            */
/*      When starting searching, set ppFirstCode to pRE->CodeList.      */
/*      pNextCode points to the RE code after the optional bit.         */
/*      Note that pNextCode[-1] contains the last RE modifier,          */
/*      which might be the not-entirely-optional ONE_OR_MORE.           */
/*      The function returns TRUE if an optional bit was found.         */
/*      First/last pointers are preserved if no optional bit found.     */
/*                                                                      */
/*      Groups of elements including zero-trip options are              */
/*      reported as a unit.  Each ONE_OR_MORE element is also           */
/*      reported separately.  The different cases can be determined     */
/*      by inspecting pNextCode[-1].  See SlashEnds below for           */
/*      more details on why the optional elements are reported          */
/*      this way.                                                       */
/*                                                                      */
/************************************************************************/
module_scope BOOL
RegExp_FindOptional(REGEXP_SPEC *pRE,
                    RE_ELEMENT **ppFirstCode, RE_ELEMENT **ppNextCode)
{
        RE_ELEMENT *pCode = *ppFirstCode;
        RE_ELEMENT *pPrev = pCode;
        RE_ELEMENT *pStartOpt;

        /*Find next *, ? or + modifier (if any)*/
        for (;;) {
                /*Remember prev code for when we find code/modifier pair*/
                pStartOpt = pPrev;
                pPrev = pCode;

                /*Analyse next component of RE*/
                switch (*pCode++) {
                case REGEXP_ZERO_OR_MORE:
                case REGEXP_ZERO_OR_ONE:
                        /*Found fully-optional element*/
                        break;

                case REGEXP_ONE_OR_MORE:
                        /*Found partially-optional element: report to client*/
                        *ppNextCode = pCode;
                        *ppFirstCode = pStartOpt;
                        return TRUE;
                        /*break;*/

                case REGEXP_END_OF_EXPR:
                        /*Didn't find any optional elements*/
                        return FALSE;
                        /*break;*/

                case REGEXP_MATCH_CLASS:
                case REGEXP_MATCH_CLASS_ALL_BUT:
                        /*Skip over codes describing set membership*/
                        pCode += REGEXP_BITMASK_SIZE;
                        continue;

                default:
                        /*Match-self code*/
                        continue;

                }

                /*Break out of the loop if control reaches here*/
                break;

        }

        /*Extend string of fully-optional elements as far as possible*/
        *ppFirstCode = pStartOpt;
        for (;;) {
                /*Decode next element of RE*/
                pPrev = pCode;
                switch (*pCode++) {
                case REGEXP_END_OF_EXPR:
                        /*String of codes finishes with previous element*/
                        *ppNextCode = pCode - 1;
                        return TRUE;
                        /*break;*/

                case REGEXP_ZERO_OR_MORE:
                case REGEXP_ONE_OR_MORE:
                case REGEXP_ZERO_OR_ONE:
                        /*These codes should not appear in this state*/
                        /*??*/
                        return FALSE;
                        /*break;*/

                case REGEXP_MATCH_CLASS:
                case REGEXP_MATCH_CLASS_ALL_BUT:
                        /*Skip over codes descibing set*/
                        pCode += REGEXP_BITMASK_SIZE;
                        break;

                default:
                        /*Code matches itself*/
                        break;

                }

                /*Look for modifier to continue fully-optional string*/
                switch (*pCode++) {
                case REGEXP_ZERO_OR_MORE:
                case REGEXP_ZERO_OR_ONE:
                        /*Optional element... continue extending string*/
                        continue;

                case REGEXP_MATCH_CLASS:
                case REGEXP_MATCH_CLASS_ALL_BUT:
                        /*Skip over elements describing members*/
                        pCode += REGEXP_BITMASK_SIZE;
                        /*FALLTHROUGH*/

                case REGEXP_ONE_OR_MORE:
                case REGEXP_END_OF_EXPR:
                default:
                        /*Finished fully-optional elements*/
                        *ppNextCode = pPrev;
                        return TRUE;
                        /*break;*/

                }

        }

        /*Control should not reach here*/
        return FALSE;

} /*FindOptional*/

/************************************************************************/
/*                                                                      */
/*      FindEnd -- Find end of RE                                       */
/*                                                                      */
/*      This function finds and reports the position of the end         */
/*      of the regular expression.                                      */
/*                                                                      */
/************************************************************************/
module_scope void
RegExp_FindEnd(REGEXP_SPEC *pRE, RE_ELEMENT **ppEnd)
{
        RE_ELEMENT *pCode = pRE->CodeList;

        /*Step through codes until end found*/
        for (;;) {
                /*Decode code*/
                switch (*pCode++) {
                case REGEXP_END_OF_EXPR:
                        /*Found the end*/
                        *ppEnd = pCode - 1;
                        return;
                        /*break*/

                case REGEXP_MATCH_CLASS:
                case REGEXP_MATCH_CLASS_ALL_BUT:
                        pCode += REGEXP_BITMASK_SIZE;
                        break;

                }

        }

} /*FindEnd*/

/************************************************************************/
/*                                                                      */
/*      RemoveCodes -- Remove optional codes from start or end          */
/*                                                                      */
/*      This function removes optional components at the start or       */
/*      the end of the RE.                                              */
/*                                                                      */
/************************************************************************/
module_scope void
RegExp_RemoveCodes(REGEXP_SPEC *pRE, RE_ELEMENT *pStart, RE_ELEMENT *pNext)
{
        RE_ELEMENT *pEnd;

        /*Find end of RE*/
        RegExp_FindEnd(pRE, &pEnd);

        /*Move elements up*/
        memmove(pStart, pNext, (pEnd-pNext + 1) * sizeof(RE_ELEMENT));

        /*Count how many states are in reduced RE*/
        RegExp_CountStates(pRE);

} /*RemoveCodes*/

/************************************************************************/
/*                                                                      */
/*      SlashEnds -- Remove start/end elements for inexact matches      */
/*                                                                      */
/*      Where the client does not care to know the exact string         */
/*      that matched a regular expression, iterative or optional        */
/*      elements at the start or end of a regular exression can         */
/*      be eliminated, reducing the backtracking required to            */
/*      perform the search.  For example:                               */
/*              a*.*b*.+Quasimodo's Dream[^a]+                          */
/*      can be reduced to:                                              */
/*              .Quasimodo's Dream[^a]                                  */
/*                                                                      */
/*      This function finds cases like this and edits the RE to         */
/*      remove unnecessary elements.  The function returns TRUE         */
/*      if optimisations were found.  Remember that the resulting       */
/*      match correctly selects lines but is otherwise inexact.         */
/*                                                                      */
/*      The optimisation consists of editing the RE for each of         */
/*      four cases, in turn:                                            */
/*              1. Remove optional elements at the front.               */
/*                 e.g. "a*b?c+defgh+i*j?" becomes "c+defgh+i*j?"       */
/*              2. Edit any leading ONE_OR_MORE element to              */
/*                 remove the iteration.                                */
/*                 e.g. "c+defgh+i*j*" becomes "cdefgh+i*j?"            */
/*              3. Remove optional elements from the back.              */
/*                 e.g. "cdefgh+i*j?" becomes "cdefgh+"                 */
/*              4. Edit any trailing ONE_OR_MORE element to             */
/*                 remove the iteration.                                */
/*                 e.g. "cdefgh+" becomes "cdefgh"                      */
/*                                                                      */
/*      I chose "slash" because I used to play lacrosse!                */
/*                                                                      */
/************************************************************************/
public_scope BOOL
RegExp_SlashEnds(REGEXP_SPEC *pRE)
{
        RE_ELEMENT *pStart = pRE->CodeList;
        RE_ELEMENT *pEnd = NULL;
        RE_ELEMENT *pSearch;
        RE_ELEMENT *pLast1OrMore;
        BOOL Trimmed = FALSE;

        /*Find first optional section of RE (if any)*/
        if (! RegExp_FindOptional(pRE, &pStart, &pEnd)) {
                /*Did not find any optional components*/
                return FALSE;
        }

        /*Did we find a fully-optional string at the start of the RE?*/
        if ((pStart == pRE->CodeList) && (pEnd[-1] != REGEXP_ONE_OR_MORE)) {
                /*Yes, was entire RE optional?*/
                if (*pEnd == REGEXP_END_OF_EXPR) {
                        /*Yes, merely match all lines using "^" pattern*/
                        pStart[0] = REGEXP_ANCHOR_START;
                        pStart[1] = REGEXP_END_OF_EXPR;
                        RegExp_CountStates(pRE);
                        return TRUE;
                }

                /*Remove redundant codes at start*/
                RegExp_RemoveCodes(pRE, pStart, pEnd);
                Trimmed = TRUE;

                /*Look for ONE_OR_MORE iteration at start*/
                if (! RegExp_FindOptional(pRE, &pStart, &pEnd)) {
                        /*No more optional components found*/
                        return TRUE;
                }

        }

        /*Did we find a ONE_OR_MORE element at the start?*/
        if ((pStart == pRE->CodeList) && (pEnd[-1] == REGEXP_ONE_OR_MORE)) {
                /*Yes, edit first element to remove iteration*/
                RegExp_RemoveCodes(pRE, pEnd - 1, pEnd);
                pStart = pEnd;
        }

        /*Look for the last string of optional elements of RE*/
        pSearch = pStart;
        pEnd = pRE->CodeList;
        for (;;) {
                /*Does this optional section contain ONE_OR_MORE iteration?*/
                if (pEnd[-1] == REGEXP_ONE_OR_MORE) {
                        /*Yes, remember last ONE_OR_MORE item seen*/
                        pLast1OrMore = pEnd;
                }

                /*Find next optional section, if any*/
                if (! RegExp_FindOptional(pRE, &pSearch, &pEnd)) {
                        /*No more optional components found*/
                        break;
                }

                /*Keep searching until last string found*/
                pStart = pSearch;
                pSearch = pEnd;

        }

        /*Did we find a string of options at the end of the RE?*/
        if (*pEnd != REGEXP_END_OF_EXPR) {
                /*No, cannot trim any options off of the end*/
                return Trimmed;
        }

        /*Is this section fully optional and with ONE_OR_MORE before?*/
        if ((pEnd[-1] != REGEXP_ONE_OR_MORE) && (pStart == pLast1OrMore)) {
                /*Yes, increase code deletion to remove 1ormore modifier*/
                pStart--;
        }

        /*Remove optional elements from the end*/
        RegExp_RemoveCodes(pRE, pStart, pEnd);
        Trimmed = TRUE;

        /*Expression has been optimised*/
        return TRUE;

} /*SlashEnds*/

/************************************************************************/
/*                                                                      */
/*      Configure -- Specify variations in module's operation           */
/*                                                                      */
/************************************************************************/
public_scope void
RegExp_Configure(LWORD OffBits, LWORD OnBits)
{
        /*Edit configuration according to caller's specifications*/
        gRegExp.Config = (gRegExp.Config & ~OffBits) | OnBits;

} /*Configure*/

/************************************************************************/
/*                                                                      */
/*      Diagnose -- Report reason for last failure                      */
/*                                                                      */
/************************************************************************/
public_scope void
RegExp_Diagnose(CHAR *pDiagnosis)
{

        /*Convert internal code into explanation understood by others*/
        switch (gRegExp.FailCode) {
        case FAIL_NO_FAILURE:
               strcpy(pDiagnosis, "(no failure)");
               break;

        case FAIL_REGEXP_TOO_LARGE:
               strcpy(pDiagnosis, "RegExp: RE too big for buffer");
               break;

        case FAIL_UNMATCHED_CLASS_OPEN:
               strcpy(pDiagnosis, "RegExp: Unmatched [ or [^");
               break;

        case FAIL_NOT_ENOUGH_MEMORY:
               strcpy(pDiagnosis, "RegExp: Not enough memory");
               break;

        case FAIL_RE_IS_OPTIMAL:
               strcpy(pDiagnosis, "RegExp: RE doesn't need splitting");
               break;

        default:
               strcpy(pDiagnosis, "RegExp: ??");
               break;

        }

} /*Diagnose*/

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
RegExp_Init(void)
{
        /*All configuration options default to being turned off*/
        gRegExp.Config = 0;

        /*No failure known initially*/
        gRegExp.FailCode = FAIL_NO_FAILURE;

} /*Init*/
