/************************************************************************/
/*                                                                      */
/*      RegExp -- Functions to assist manipulating regular expressions  */
/*                                                                      */
/*      This module provides assistance to programs that offer RE       */
/*      facilities to the Real World.  It provides functions to         */
/*      describe regular expressions in a compact and convenient        */
/*      fashion.                                                        */
/*                                                                      */
/*      "All but" classes are given a code of their own separate to     */
/*      "any of" classes so that case insensitivity can be deferred     */
/*      until the class is used.                                        */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse be behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#ifndef REGEXP_H
#define REGEXP_H

#include <compdef.h>
#include <limits.h>

#ifdef __cplusplus
extern "C" {
#endif /*__cplusplus*/

/*Module configuration options*/

#define REGEXP_MATCH_FULL_LINE_ONLY     BIT0

/*Limit on number of codes allowed in a compiled format (sigh)*/
#define REGEXP_CODES_MAX                1024

/*Codes for specifying regular expression elements*/

/*Codes 0..UCHAR_MAX represent themselves (literals)*/
typedef WORD RE_ELEMENT;

/*Each class specifier is followed by a bitmask of members*/
#define REGEXP_BITMASK_SIZE             16

/*Codes after UCHAR_MAX represent more sophisticated pattern elements*/
#define REGEXP_MATCH_ANY                ((RE_ELEMENT) (UCHAR_MAX +  1))
#define REGEXP_MATCH_CLASS              ((RE_ELEMENT) (UCHAR_MAX +  2))
#define REGEXP_MATCH_CLASS_ALL_BUT      ((RE_ELEMENT) (UCHAR_MAX +  3))
#define REGEXP_ANCHOR_START             ((RE_ELEMENT) (UCHAR_MAX +  4))
#define REGEXP_ANCHOR_END               ((RE_ELEMENT) (UCHAR_MAX +  5))
#define REGEXP_ZERO_OR_ONE              ((RE_ELEMENT) (UCHAR_MAX +  6))
#define REGEXP_ZERO_OR_MORE             ((RE_ELEMENT) (UCHAR_MAX +  7))
#define REGEXP_ONE_OR_MORE              ((RE_ELEMENT) (UCHAR_MAX +  8))
#define REGEXP_END_OF_EXPR              ((RE_ELEMENT) (UCHAR_MAX +  9))

/*Specifier which both identifies expression and provides return information*/
typedef struct {
        /*Number of states within RE*/
        UINT NrStates;

        /*Compiled form of expression*/
        RE_ELEMENT CodeList[REGEXP_CODES_MAX];

} REGEXP_SPEC;

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
void
RegExp_Init(void);

/************************************************************************/
/*                                                                      */
/*      Configure -- Specify variations in module's operation           */
/*                                                                      */
/************************************************************************/
void
RegExp_Configure(LWORD OffBits, LWORD OnBits);

/************************************************************************/
/*                                                                      */
/*      Compile -- Check and convert text expression to compact form    */
/*                                                                      */
/*      The function returns FALSE if the text form was badly           */
/*      formatted.                                                      */
/*                                                                      */
/************************************************************************/
BOOL
RegExp_Compile(CHAR *pText, REGEXP_SPEC **ppSpec);

/************************************************************************/
/*                                                                      */
/*      CompileString -- Convert string without looking for RE syntax   */
/*                                                                      */
/*      The function returns FALSE if the conversion failed.            */
/*                                                                      */
/************************************************************************/
BOOL
RegExp_CompileString(CHAR *pText, REGEXP_SPEC **ppSpec);

/************************************************************************/
/*                                                                      */
/*      EasiestFirst -- Improve search by deferring complex starts      */
/*                                                                      */
/*      An expression like "[^a]*.*[^b]*.*c.*Quasimodo's Dream.*g"      */
/*      is very expensive to search if you simply use the expression    */
/*      as given, as the first elements involve lots of backtracking.   */
/*      However, most regular expressions like this have a              */
/*      relatively easy middle bit (e.g. "Quasimodo's Dream") that      */
/*      is required for the match and can be searched first instead.    */
/*      This function determines whether this optimisation is worth     */
/*      while.  If not, the function returns FALSE.                     */
/*                                                                      */
/*      If the search can be optimised by the split, EasiestFirst       */
/*      splits the original RE so that it matches the complicated       */
/*      starting bit, and creates a new RE spec which begins with       */
/*      the easiest bit.                                                */
/*                                                                      */
/************************************************************************/
BOOL
RegExp_EasiestFirst(REGEXP_SPEC *pSpec, REGEXP_SPEC **ppEasierBit);

/************************************************************************/
/*                                                                      */
/*      SlashEnds -- Remove start/end elements for inexact matches      */
/*                                                                      */
/*      Where the client does not care to know the exact string         */
/*      that matched a regular expression, iterative or optional        */
/*      elements at the start or end of a regular exression can         */
/*      be eliminated, reducing the backtracking required to            */
/*      perform the search.  For example:                               */
/*              a*.*b*.*Quasimodo's Dream[^a]+                          */
/*      can be reduced to:                                              */
/*              Quasimodo's Dream[^a]                                   */
/*                                                                      */
/*      This function finds cases like this and edits the RE to         */
/*      remove unnecessary elements.  The function returns TRUE         */
/*      if optimisations were found.  Remember that the resulting       */
/*      match correctly selects lines but is otherwise inexact.         */
/*                                                                      */
/************************************************************************/
BOOL
RegExp_SlashEnds(REGEXP_SPEC *pRE);

/************************************************************************/
/*                                                                      */
/*      Diagnose -- Report reason for last failure                      */
/*                                                                      */
/************************************************************************/
void
RegExp_Diagnose(CHAR *pDiagnosis);

#ifdef __cplusplus
}
#endif /*__cplusplus*/

#endif /*REGEXP_H*/
