/************************************************************************/
/*                                                                      */
/*      RETable -- Efficient language for implementing regexp searches  */
/*                                                                      */
/*      Introducing the Grouse finite state machine architecture!       */
/*                                                                      */
/*      This module converts a regular expression description           */
/*      from a compact form (as created by RegExp) to a                 */
/*      high-performance table-driven form (as described by             */
/*      MatchEng).  Several optimisations are available that            */
/*      can significantly improve search performance.  These            */
/*      optimisations fall into two categories: some try to             */
/*      minimise backtracking initiated by iterative states             */
/*      by analysing the state table following the iterative            */
/*      state; others use string or byte searches to improve            */
/*      the speed of scanning the text for possible matches.            */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include "ascii.h"
#include "bakslash.h"
#include <compdef.h>
#include <ctype.h>
#include "matcheng.h"
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "retable.h"

#define RE_STATES_MAX                   200

/*Module variables*/

typedef struct {
        /*Configuration options*/
        UINT32 Config;

        /*Debugging options*/
        LWORD Debug;

} RETABLE_MODULE_CONTEXT;

module_scope RETABLE_MODULE_CONTEXT gRETable;

#if DEBUGGING
#define DEBUG(Item, Msg)           {if (gRETable.Debug & Item) {printf(Msg);}}
#define DEBUG1(Item, Msg, p1)      {if (gRETable.Debug & Item) {printf(Msg, p1);}}
#define DEBUG2(Item, Msg, p1, p2)  {if (gRETable.Debug & Item) {printf(Msg, p1, p2);}}
#else
#define DEBUG(Item, Msg)
#define DEBUG1(Item, Msg, p1)
#define DEBUG2(Item, Msg, p1, p2)
#endif /*DEBUGGING*/

/************************************************************************/
/*                                                                      */
/*      Configure -- Modify search behaviour according to selections    */
/*                                                                      */
/*      Although strictly speaking the configuration items provided     */
/*      should be specified via RegExp, in practice it is much more     */
/*      convenient to deal with them in the table-driven format.        */
/*                                                                      */
/************************************************************************/
public_scope void
RETable_Configure(LWORD OffBits, LWORD OnBits)
{
        gRETable.Config = (gRETable.Config & ~OffBits) | OnBits;

} /*Configure*/

/************************************************************************/
/*                                                                      */
/*      AddMatchChars -- Write action for chars that match RE code      */
/*                                                                      */
/*      This function interprets what characters match the specified    */
/*      regular expression code, and changes the table entries for      */
/*      those characters to ADVANCE.  It returns a type code            */
/*      summarising table behaviour so that later users of the          */
/*      table don't need to reverse-engineer the table type.            */
/*                                                                      */
/************************************************************************/
module_scope UINT
RETable_AddMatchChars(RE_ELEMENT *pRE, BYTET *pTable)
{
        UINT GroupIx;
        UINT ByteIx;
        WORD Bitmask;
        UINT Lit;
        BYTET NULAction;
        BYTET LFAction;

        switch (*pRE++) {
        case REGEXP_MATCH_ANY:
                NULAction = pTable[NUL];
                LFAction = pTable[LF];
                memset(pTable, ADVANCE, TABLE_SIZE);
                pTable[NUL] = NULAction;
                pTable[LF] = LFAction;
                return MATCH_ANY_TYPE;
                break;

        case REGEXP_END_OF_EXPR:
                return SUCCESS_TYPE;
                break;

        case REGEXP_MATCH_CLASS:
                /*Match class: are we ignoring case?*/
                if (gRETable.Config & RETABLE_IGNORE_CASE) {
                        /*Yes, combine (or) uppercase and lowercase letters*/
                        pRE[4] |= (pRE[6] & 0xfffe);    /*A-O*/
                        pRE[5] |= (pRE[7] & 0x07ff);    /*P-Z*/
                        pRE[6] |= (pRE[4] & 0xfffe);    /*a-o*/
                        pRE[7] |= (pRE[5] & 0x07ff);    /*p-z*/

                        /*Warning: Code above presumes ASCII character set*/

                }

                /*Decode 16 word specifier into table entries*/
                for (GroupIx = 0; GroupIx < TABLE_SIZE; GroupIx += 16) {
                        Bitmask = BIT0;
                        for (ByteIx = 0; ByteIx < 16; ByteIx++) {
                                if (*pRE & Bitmask) {
                                        pTable[GroupIx + ByteIx] = ADVANCE;
                                }
                                Bitmask <<= 1;
                        }
                        pRE++;
                }
                return CLASS_TYPE;
                break;

        case REGEXP_MATCH_CLASS_ALL_BUT:
                /*Match all but class: are we ignoring case?*/
                if (gRETable.Config & RETABLE_IGNORE_CASE) {
                        /*Yes, combine (or) uppercase and lowercase letters*/
                        pRE[4] |= (pRE[6] & 0xfffe);    /*A-O*/
                        pRE[5] |= (pRE[7] & 0x07ff);    /*P-Z*/
                        pRE[6] |= (pRE[4] & 0xfffe);    /*a-o*/
                        pRE[7] |= (pRE[5] & 0x07ff);    /*p-z*/

                        /*Warning: Code above presumes ASCII character set*/

                }

                /*Decode 16 word specifier into table entries*/
                for (GroupIx = 0; GroupIx < TABLE_SIZE; GroupIx += 16) {
                        Bitmask = BIT0;
                        for (ByteIx = 0; ByteIx < 16; ByteIx++) {
                                if (! (*pRE & Bitmask)) {
                                        pTable[GroupIx + ByteIx] = ADVANCE;
                                }
                                Bitmask <<= 1;
                        }
                        pRE++;
                }
                return ALL_BUT_TYPE;
                break;

        default:
                /*Is the code a literal?*/
                Lit = (UINT) *(pRE-1);
                if (Lit >= TABLE_SIZE) {
                        /*No, cannot expand*/
                        printf("??AddMatch: Code %u not handled", Lit);
                        break;
                }

                /*Set up table for first character of literal*/
                pTable[Lit] = ADVANCE;

                /*Are we ignoring case and handling a letter?*/
                if ((gRETable.Config & RETABLE_IGNORE_CASE) &&
                                isalpha(Lit)) {
                        /*Yes, ensure both lower and upper case match*/
                        pTable[tolower(Lit)] = ADVANCE;
                        pTable[toupper(Lit)] = ADVANCE;
                }

                return LITERAL_TYPE;

        }

        /*We should never reach here... return a faulty type*/
        return THIS_IS_NOT_A_TYPE;

} /*AddMatchChars*/

/************************************************************************/
/*                                                                      */
/*      IterationOpt -- Use following table to optimise backtracking    */
/*                                                                      */
/*      Each time an iterative part of the RE is reached, it is         */
/*      matched as far as possible, and then matching the rest of       */
/*      the expression is attempted.  If the next match fails, the      */
/*      last iteration is undone and matching the remainder is          */
/*      attempted again.  This is powerful but expensive.               */
/*                                                                      */
/*      IterationOpt finds significant savings by reducing the number   */
/*      of backtracking pathways that need to be considered.            */
/*      This is done by observing where the following table does        */
/*      not match, and editing the actions in the iteration table       */
/*      to not bother pushing a backtracking path for those bytes.      */
/*      For example, "fred.*bloggs" can be improved since we can        */
/*      see that during the ".*" iteration, it is only worth            */
/*      attempting to backtrack if the character was "b".               */
/*                                                                      */
/*      The optimisation is very easy to implement: for each byte       */
/*      of the input, if the iteration table matches with               */
/*      AGAIN_PUSH_ADVANCE but the following table has NO_MATCH,        */
/*      the backtracking push is worthless, so change the entry         */
/*      in the iteration table to AGAIN.                                */
/*                                                                      */
/************************************************************************/
module_scope void
RETable_IterationOpt(BYTET *pIter, BYTET *pNext)
{
        UINT i;

        /*Check each iteration table entry against next table*/
        for (i = 0; i < TABLE_SIZE; i++, pIter++, pNext++) {
                /*Does the current table entry include backtracking?*/
                if (*pIter != AGAIN_PUSH_ADVANCE) {
                        /*No, ignore entry*/
                        continue;
                }

                /*Does this char in the next table represent a dead-end?*/
                if (*pNext != NO_MATCH) {
                        /*No, optimisation is not possible*/
                        continue;
                }

                /*Change iteration to not bother backtracking*/
                *pIter = AGAIN;

        }


} /*IterationOpt*/

/************************************************************************/
/*                                                                      */
/*      MakeIterative -- Add "star" iterations to table                 */
/*                                                                      */
/*      This function changes a table from simply being a match with    */
/*      advance if successful and backtrack if failure to push          */
/*      backtracking if successful and advance if failure.              */
/*                                                                      */
/************************************************************************/
module_scope void
RETable_MakeIterative(BYTET *pTab)
{
        UINT i;

        /*Loop through each element of table, changing action*/
        for (i = 0; i < TABLE_SIZE; i++) {
                /*Change this table action*/
                switch (*pTab) {
                case ADVANCE:
                        *pTab++ = AGAIN_PUSH_ADVANCE;
                        break;

                default:
                        *pTab++ = BACK_AND_ADVANCE;
                        break;

                }

        }

} /*MakeIterative*/

/************************************************************************/
/*                                                                      */
/*      WordsOnly -- Edit state table to only include word characters   */
/*                                                                      */
/*      Changes characters that are not part of a word -- where word    */
/*      characters are [0-9A-Za-z_] -- to be no match.                  */
/*                                                                      */
/************************************************************************/
module_scope void
RETable_WordsOnly(BYTET *pTab)
{
        UINT i;
        BYTET FailAction = pTab[0];

        /*Loop through each element of table, checking match cases*/
        for (i = 0; i < TABLE_SIZE; i++, pTab++) {
                /*Is this item currently a match?*/
                if (*pTab == FailAction) {
                        /*No, ignore it*/
                        continue;
                }

                /*Is this match case an word character?*/
                if (isalnum(i) || (((CHAR) i) == '_')) {
                        /*Yes, it is okay*/
                        continue;
                }

                /*Change action to be no match*/
                *pTab = FailAction;

        }

} /*WordsOnly*/

/************************************************************************/
/*                                                                      */
/*      EndOfWord -- Convert last state to require word end & start     */
/*                                                                      */
/*      Matching words is performed by restricting states within the    */
/*      match to only accept characters that are in a word, and to      */
/*      restrict the end and start of the match to match only nonword   */
/*      characters.  This function provides the start/end checks.       */
/*      The end check is provided by removing in-word characters        */
/*      from the final success state; the start check is provided       */
/*      by changing COMPLETED to COMPLETED_IF_WORD_STARTED actions      */
/*      so that the match only succeeds if the character before the     */
/*      match start was not from within a word.                         */
/*                                                                      */
/************************************************************************/
module_scope void
RETable_EndOfWord(BYTET *pTab)
{
        UINT i;

        /*Loop through success table, restricting match to nonword chars*/
        for (i = 0; i < TABLE_SIZE; i++, pTab++) {
                /*Is this entry not considered a match (e.g. end-anchor)?*/
                if (*pTab != COMPLETED) {
                        /*Yes, don't play with it here*/
                        continue;
                }

                /*Are we checking that match ends on a word boundary?*/
                if (gRETable.Config & RETABLE_MATCH_WORD_END) {
                        /*Yes, is this character part of a word?*/
                        if (isalnum(i) || (((CHAR) i) == '_')) {
                                /*Yes, it is no longer a match*/
                                *pTab = NO_MATCH;
                                continue;
                        }
                }

                /*Are we checking that match starts on word boundary?*/
                if (gRETable.Config & RETABLE_MATCH_WORD_START) {
                        /*Yes, change nonword matches to check start of word*/
                        *pTab = COMPLETE_IF_WORD_START;
                }

        }

} /*EndOfWord*/

/************************************************************************/
/*                                                                      */
/*      AllowZeroMatch -- Convert table to match 0 or 1 times           */
/*                                                                      */
/*      This function modifies non-match actions of specified table     */
/*      to instead mean retry character in next table.                  */
/*                                                                      */
/************************************************************************/
module_scope void
RETable_AllowZeroMatch(BYTET *pTab)
{
        UINT i;

        /*Loop through each element of table, changing action*/
        for (i = 0; i < TABLE_SIZE; i++, pTab++) {
                /*Change this table action*/
                switch (*pTab) {
                case NO_MATCH:
                        *pTab = BACK_AND_ADVANCE;
                        break;

                case ADVANCE:
                        /*Matched, but push skip case into backtrack stack*/
                        *pTab = ADVANCE_PUSH_ZERO;
                        break;

                }

        }

} /*AllowZeroMatch*/

/************************************************************************/
/*                                                                      */
/*      AllOptional -- Check RE specs after end-of-line spec            */
/*                                                                      */
/*      The pattern "$.*" matches the end of line; "$b" must not.       */
/*      This function examines all the RE codes presented before        */
/*      the end-of-expression marker, and returns TRUE if the           */
/*      null string matches the list (all the RE elements are           */
/*      optional).                                                      */
/*                                                                      */
/************************************************************************/
module_scope BOOL
RETable_AllOptional(RE_ELEMENT *pRE)
{
        UINT NrSpecs = 0;

        /*Loop through codes until end of RE reported*/
        for (;;) {
                switch (*pRE++) {
                case REGEXP_END_OF_EXPR:
                        /*Search completed: report if all specs optional*/
                        return NrSpecs == 0;
                        break;

                case REGEXP_ZERO_OR_MORE:
                case REGEXP_ZERO_OR_ONE:
                        /*Previous component was optional*/
                        NrSpecs--;
                        break;

                case REGEXP_ANCHOR_END:
                        /*Multiple end-of-line specifiers found: okay*/
                        break;

                default:
                        /*Anything else matches a character*/
                        NrSpecs++;
                        break;

                }

        }

} /*AllOptional*/

/************************************************************************/
/*                                                                      */
/*      PrepareStart -- Edit start table to trigger RE match attempts   */
/*                                                                      */
/*      First of all, don't ASK how much time behoffski has spent       */
/*      trying to get the start of RE searches correct.  This           */
/*      function promises (fingers crossed) to be the correct           */
/*      solution.  While it costs more in terms of false starts,        */
/*      having a separate "starting" state makes buffer optimisations   */
/*      for anchored searches and the Boyer-Moore algo much easier.     */
/*      It also doubles as a convenient way of capturing the start      */
/*      of the match which was starting to look rather ugly.            */
/*                                                                      */
/*      This function observes what characters are matched in the       */
/*      client's first RE table, and places actions into the search-    */
/*      for-start table to trigger a match attempt.  It has been        */
/*      optimised to include advancing to the second table where        */
/*      the first table merely contains a simple match & advance.       */
/*                                                                      */
/************************************************************************/
module_scope void
RETable_PrepareStart(MATCHENG_SPEC *pSpec, BOOL Anchored)
{
        BYTET *pTab = pSpec->Tables[1];
        BYTET *pFoll = pSpec->Tables[2];
        UINT i;
        BYTE StartAction;
        BYTE StartActionAdvance;

        /*Is the search anchored?*/
        if (Anchored) {
                /*Yes, don't push backtrack when first char matches*/
                StartAction = START_MATCH;
                StartActionAdvance = START_MATCH;

                /*Default of NO_MATCH in start state will fail match*/

        } else {
                /*No, sliding state achieved by push into backtrack stack*/
                StartAction = START_MATCH_PUSH;
                StartActionAdvance = START_MATCH_PUSH_ADVANCE;

                /*Advance if no pattern match (except for NUL)*/
                memset(pTab, AGAIN, TABLE_SIZE);

        }

        /*Write actions for RE start search*/
        for (i = 0; i < TABLE_SIZE; i++, pTab++, pFoll++) {
                /*Does the following table entry contain a non-match action?*/
                switch (*pFoll) {
                case NO_MATCH:
                case ABANDON:
                        /*Yes, don't trigger start of RE*/
                        break;

                case ADVANCE:
                        /*Extra optimisation: start RE match from 2nd state*/
                        *pTab = StartActionAdvance;
                        break;

                default:
                        /*Character matches start of RE*/
                        *pTab = StartAction;

                }

        }

        /*Handle special cases of zero-trip or completed first states*/
        switch (pSpec->Tables[2][NUL]) {
        case COMPLETED:
                /*RE is "$"... leave start-match action in place*/
                break;

        case BACK_AND_ADVANCE:
                /*Match might select empty string since first item does*/
                pSpec->Tables[1][NUL] = START_MATCH;
                break;

        default:
                /*NUL in starting state abandons search*/
                pSpec->Tables[1][NUL] = ABANDON;
        }

        /*Does LF match with zero trip in first state?*/
        if (pSpec->Tables[2][LF] == BACK_AND_ADVANCE) {
                /*Yes, it may start match*/
                pSpec->Tables[1][LF] = START_MATCH;
        }

} /*PrepareStart*/

/************************************************************************/
/*                                                                      */
/*      Expand -- Convert "compact" RE codes to optimised table form    */
/*                                                                      */
/*      This function is used to prepare for table-driven pattern       */
/*      match searching.  It creates a fairly verbose form of the       */
/*      regular expression which is highly optimised for match          */
/*      searches.  A typical pattern will occupy 8 to 20 kbytes of      */
/*      RAM in its expanded form.                                       */
/*                                                                      */
/*      The function returns TRUE and a handle for the expanded format  */
/*      or FALSE if unable to prepare an expanded version.              */
/*                                                                      */
/************************************************************************/
public_scope BOOL
RETable_Expand(REGEXP_SPEC *pRESpec, MATCHENG_SPEC **ppSpec)
{
        RE_ELEMENT *pRE = pRESpec->CodeList;
        MATCHENG_SPEC *pSpec;
        UINT SpecSize;
        BYTET *pTab;
        BYTET *pFlags;
        BOOL Anchored = FALSE;

        /*Destroy return pointer to reduce chances of being misinterpreted*/
        *ppSpec = NIL;

        /*Is the specification small enough to be handled?*/
        if (pRESpec->NrStates >= RE_STATES_MAX) {
                /*No, cannot expand*/
                return FALSE;

        }

        /*Buffer size is one table/state, plus flags/misc/fail/start tables*/
        SpecSize = sizeof(TABLE) * (pRESpec->NrStates + 4);
        pSpec = (MATCHENG_SPEC *) malloc(SpecSize);
        if (pSpec == NULL) {
                /*Unable to acquire memory for expanded format*/
                return FALSE;
        }

        /*Initialise all tables with most likely value (doesn't match)*/
        memset(pSpec, NO_MATCH, SpecSize);

        /*Position pointer so flags summarising each state work backwards*/
        pSpec = (MATCHENG_SPEC *) (((BYTET *) pSpec) + sizeof(TABLE));

        /*Initialise pointers into spec*/
        pTab = (BYTET *) &pSpec->Tables[0];
        pFlags = ((BYTET *) pSpec) - 1;

        /*Initialise miscellaneous variables within specification*/
        pSpec->SearchSkipOffset = 0;

        /*First table is always failure to simplify backtracking*/
        memset(pTab, ABANDON, TABLE_SIZE);
        pTab += TABLE_SIZE;
        *pFlags-- = FAIL_TYPE;

        /*Set up initial table non-match actions*/
        if (*pRE == REGEXP_ANCHOR_START) {
                /*Skip over anchor specifier*/
                pRE++;

                /*Remember specifier when conditioning first table*/
                Anchored = TRUE;

        }

        /*Skip start search table for the moment (but write flag code now)*/
        pTab += TABLE_SIZE;
        *pFlags-- = START_TYPE;

        /*Continue creating tables until end-of-RE specifier found*/
        for (;;) {
                /*Default to no iteration in previous table*/
                *pFlags = 0;

                /*All entries of each table begin as NO_MATCH*/

                /*Handle non-table specifiers, if any*/
MetaSpecifiers:
                switch (*pRE) {
                case REGEXP_ANCHOR_END:
                        /*Are all remaining RE specs (if any) optional?*/
                        if (RETable_AllOptional(pRE + 1)) {
                                /*Yes, line terminator means success*/
                                pTab[NUL] = COMPLETED;
                                pTab[LF] = COMPLETED;

                        } else {
                                /*Match can never occur: destroy start*/
                                memset(pSpec->Tables[1], ABANDON,
                                                TABLE_SIZE);
                                *pFlags = SUCCESS_TYPE;
                                *ppSpec = pSpec;
                                return TRUE;

                        }

                        /*Nothing else in the RE is worth expanding*/
                        goto FinishedExpanding;
                        break;

                case REGEXP_ONE_OR_MORE:
                case REGEXP_ZERO_OR_MORE:
                        /*Copy previous table if one or more*/
                        if (*pRE == REGEXP_ONE_OR_MORE) {
                                /*Copy previous table to current table*/
                                memcpy(pTab, pTab - TABLE_SIZE, TABLE_SIZE);
                                pTab += TABLE_SIZE;
                                memset(pTab, NO_MATCH, TABLE_SIZE);
                                *--pFlags = 0;
                        }

                        /*Modify match actions of prev table*/
                        RETable_MakeIterative(pTab - TABLE_SIZE);
                        pFlags[1] |= ITERATION_FLAG;
                        break;

                case REGEXP_ZERO_OR_ONE:
                        /*Convert no match in prev to be attempt advance*/
                        RETable_AllowZeroMatch(pTab - TABLE_SIZE);
                        pFlags[1] |= ZERO_TRIP_FLAG;
                        break;

                case REGEXP_END_OF_EXPR:
                        goto FinishedExpr;
                        break;

                default:
                        /*Specifier is presumably real, not meta*/
                        goto AddMatchChars;
                        break;

                }

                /*Skip over processed specifier*/
                pRE++;
                goto MetaSpecifiers;

AddMatchChars:
                /*Convert current RE element into table format*/
                *pFlags = RETable_AddMatchChars(pRE, pTab);
                switch (*pFlags) {
                case CLASS_TYPE:
                case ALL_BUT_TYPE:
                        /*RE specifier took 17 codes*/
                        pRE += 17;
                        break;

                default:
                        /*RE took 1 code*/
                        pRE += 1;
                        break;

                }

                /*Advance to next table*/
                pTab += TABLE_SIZE;
                pFlags--;

        }

FinishedExpr:
        /*Final table selects successful match for any char*/
        memset(pTab, COMPLETED, TABLE_SIZE);

        /*Ensure NUL is not a match in starting table (e.g. RE "^")*/
        pSpec->Tables[1][NUL] = ABANDON;

        /*?? Not sure if this is correct*/
        pSpec->Tables[1][LF] = ABANDON;

FinishedExpanding:
        /*Write flag for final match-completed state*/
        *pFlags = SUCCESS_TYPE;

        /*Are we restricted to matching words?*/
        if (gRETable.Config & RETABLE_MATCH_WORD) {
                /*Yes, change success state to check end and/or start of word*/
                RETable_EndOfWord(pTab);
        }

        /*Post-process tables for iteration optimisations and/or word search*/
        for (; TABLE_TYPE(*pFlags) != START_TYPE;
                        pFlags++, pTab -= TABLE_SIZE) {
                /*Is the match restricted to words?*/
                if (gRETable.Config & RETABLE_MATCH_WORD) {
                        /*Reclassify match-many table types (Boyer-Moore)*/
                        switch (TABLE_TYPE(*pFlags)) {
                        case SUCCESS_TYPE:
                                /*Do not convert this state for words*/
                                break;

                        case LITERAL_TYPE:
                                /*Leave table alone even if non-word char*/
                                break;

                        case ALL_BUT_TYPE:
                        case MATCH_ANY_TYPE:
                                /*These are now effectively classes*/
                                *pFlags = (*pFlags & ~TABLE_TYPE_MASK)
                                                | CLASS_TYPE;
                                /*FALLTHROUGH*/

                        default:
                                /*Convert table so only word chars match*/
                                RETable_WordsOnly(pTab);
                                break;

                        }

                }

                /*Did this table contain iteration?*/
                if (*pFlags & ITERATION_FLAG) {
                        /*Yes, optimise iteration backtracking*/
                        RETable_IterationOpt(pTab, pTab + TABLE_SIZE);
                }

                /*Did this table contain an optional (zero-trip) match?*/
                /*?? not implemented yet*/
                        /*Yes, optimise backtracking as per iteration*/
                        /*?? not implemented yet*/


        }

        /*Now that first state matches known, set search start conditions*/
        RETable_PrepareStart(pSpec, Anchored);

        /*Write expanded form to caller*/
        *ppSpec = pSpec;
        return TRUE;

} /*Expand*/

/************************************************************************/
/*                                                                      */
/*      UseBuffer -- Optimise search for multi-line buffer              */
/*                                                                      */
/*      Once the match loop itself has been optimised, the slowest      */
/*      part of the search becomes the process of breaking the          */
/*      file up into lines.  Since the search is almost always on       */
/*      files that already contain multiple lines separated by          */
/*      LF or CR/LF, there are huge benefits in having the search       */
/*      work on the entire buffer.                                      */
/*                                                                      */
/*      This function converts the expanded pattern to work with        */
/*      a buffer where each line is guaranteed to be preceded and       */
/*      terminated by a LF (with an option to include CR as a           */
/*      terminator).  The most important requirements are that a        */
/*      LF be the first char presented (e.g. added to start of          */
/*      buffer) and that the last line of the buffer be terminated      */
/*      by LF and NUL.  Searches anchored to the start of the line      */
/*      then merely require a LF as the first literal.                  */
/*                                                                      */
/*      Extra match engine actions are added so that the search for     */
/*      start of regular expression can count line numbers, plus        */
/*      report start of line on a match to the caller.                  */
/*                                                                      */
/*      Selecting CR as a line terminator is given as a parameter       */
/*      so this behaviour can be varied dynamically.                    */
/*                                                                      */
/*      A few patterns require the line-by-line search.  If             */
/*      UseBuffer is requested for such a pattern, the function         */
/*      returns FALSE.                                                  */
/*                                                                      */
/************************************************************************/
public_scope BOOL
RETable_UseBuffer(MATCHENG_SPEC *pSpec, BOOL CRIsTerminator)
{
        BYTET *pTab = pSpec->Tables[1];
        BOOL Anchored;
        BYTET *pFlags = ((BYTET *) pSpec) - 1;

        /*Is the RE very short ("$" or "^")?*/
        if (TABLE_TYPE(pFlags[-2]) == SUCCESS_TYPE) {
                /*Yes, cannot optimise (?)*/
                return FALSE;
        }

        /*Remember if the RE is anchored to the start of the line*/
        Anchored = (pTab[1] == NO_MATCH) || (pTab[1] == START_MATCH);

        /*Is this search anchored?*/
        if (Anchored) {
                /*Anchored: RE start search no longer triggers on next state*/
                memset(pTab, AGAIN, TABLE_SIZE);

                /*LF triggers start (and pushes backtracking marker)*/
                pSpec->Tables[1][LF] = MARK_LINE_START_PUSH_ADVANCE;

        } else {
                /*Unanchored search: does LF start a match?*/
                if (pSpec->Tables[2][LF] == NO_MATCH) {
                        /*No, merely mark line start*/
                        pSpec->Tables[1][LF] = MARK_LINE_START;
                } else {
                        /*Yes, mark line start and start match*/
                        pSpec->Tables[1][LF] = MARK_LINE_START_PUSH_ADVANCE;
                }

        }

        /*NUL in starting table always means abandon buffer*/
        pSpec->Tables[1][NUL] = ABANDON;

        /*Were we given LF in the first state as a literal?*/
        pTab = pSpec->Tables[2];
        pFlags = ((BYTET *) pSpec) - 2;

        /*Change LF action in RE search states to be same as NUL*/
        for (; TABLE_TYPE(*pFlags--) != SUCCESS_TYPE; pTab += TABLE_SIZE) {
                pTab[LF] = pTab[NUL];

        }

        /*Is a carriage return deemed to be a line terminator?*/
        if (! CRIsTerminator) {
                /*No, conversion finished*/
                return TRUE;
        }

        /*Change CR action in all tables except start search to follow LF*/
        for (pTab = pSpec->Tables[2], pFlags = ((BYTET *) pSpec) - 2;
                        TABLE_TYPE(*pFlags--) != SUCCESS_TYPE;
                        pTab += TABLE_SIZE) {
                pTab[CR] = pTab[LF];

        }

        /*CR is terminator: does it match in first state?*/
        if (pSpec->Tables[2][CR] == NO_MATCH) {
                /*No, merely skip over it during starting search*/
                pSpec->Tables[1][CR] = AGAIN;
        } else {
                /*Pattern might match empty lines*/
                pSpec->Tables[1][CR] = START_MATCH_PUSH_ADVANCE;
        }

        /*Conversion finished*/
        return TRUE;

} /*UseBuffer*/

/************************************************************************/
/*                                                                      */
/*      UseByteSearch -- Use byte search optimisation if feasible       */
/*                                                                      */
/*      It's a terrible thing, you know.  As soon as you try to be      */
/*      the fastest at something, you have to beat EVERYBODY who has    */
/*      come before,  which means including all sorts of algorithms     */
/*      in your code.  This function covers one case which was only     */
/*      included very late so that the search ran faster than           */
/*      another simple and highly efficient assembly-language grep      */
/*      which searched for the start of the match using the CPU's       */
/*      string (byte) search instruction.                               */
/*                                                                      */
/*      While the lodsb/xlat/jmp sequence only costs roughly            */
/*      20 cycles or so per byte on a 486, it still is considerably     */
/*      slower than the string search "rep; scasb" which only takes     */
/*      10 cycles per byte.  This function determines whether the       */
/*      first state contains only a single byte as the start-of-match   */
/*      condition, and, if so, changes the starting state to use the    */
/*      CPU instruction by placing the BYTE_SEARCH action on every      */
/*      bytes except the desired value itself.                          */
/*                                                                      */
/*      The function reports whether the tables were edited to use      */
/*      byte searching.  This allows the byte search to have            */
/*      priority over Boyer-Moore searching where both are possible.    */
/*                                                                      */
/************************************************************************/
module_scope BOOL
RETable_UseByteSearch(MATCHENG_SPEC *pSpec, CHAR *pTrailingLiteral)
{
        BYTET *pStart = pSpec->Tables[1];
        BYTET *pSearch;
        CHAR Literal;
        UINT i;

        /*Count number of ways first state table can match*/
        pSearch = &pSpec->Tables[2][1];
        Literal = NUL;
        for (i = 1; i < TABLE_SIZE; pSearch++, i++) {
                /*Does this entry start a match?*/
                if (*pSearch == NO_MATCH) {
                        /*No, ignore it*/
                        continue;
                }

                /*Have we seen a literal previously?*/
                if (Literal != NUL) {
                        /*Yes, cannot optimise unless only one match*/
                        return FALSE;
                }

                /*Remember literal for possible optimisation*/
                Literal = (CHAR) (pSearch - pSpec->Tables[2]);

        }

        /*Did we find an entry?*/
        if (Literal == NUL) {
                /*No, rather unusual RE: cannot optimise start*/
                return FALSE;
        }

        /*Found literal: change all start non-matches to byte search*/
        pStart++;
        for (i = 1; i < TABLE_SIZE; i++, pStart++) {
                /*Is this entry a non-match?*/
                if (*pStart == AGAIN) {
                        /*Yes, change it to switch to byte search*/
                        *pStart = BYTE_SEARCH;
                }
        }

        /*Search engine finds literal hacked into SearchSkipOffset*/
        pSpec->SearchSkipOffset = Literal;

        /*Tell client since optimisation requires buffer conditioning*/
        *pTrailingLiteral = Literal;

        /*Tell client that byte search was installed*/
        return TRUE;

} /*UseByteSearch*/

/************************************************************************/
/*                                                                      */
/*      StartSearchOpt -- Optimise where RE begins with literal(s)      */
/*                                                                      */
/*      Virtually the entire cost of the average regular expression     */
/*      search is the cost of finding a search starting point.          */
/*      If the client does not require special intelligence during      */
/*      the start search (e.g. no line counting), then many patterns    */
/*      can be optimised by either the Boyer-Moore algorithm or         */
/*      by the string instructions provided by the CPU.                 */
/*                                                                      */
/*      The Boyer-Moore search operates where the pattern starts        */
/*      with two or more literals or classes.  The algorithm works      */
/*      by checking for the last character of the string instead of     */
/*      the first.  This allows the search to skip forward many         */
/*      bytes if the character checked cannot match the string at       */
/*      all.  For example, if searching for "typewriter", a "Q"         */
/*      allows the search to skip forward 10 bytes; a "t" allows        */
/*      the search to skip forward two bytes.                           */
/*                                                                      */
/*      If the Boyer-Moore optimisation cannot be used, the start       */
/*      search can still be optimised if the first state is a           */
/*      simple literal (case sensitive if a letter).  In this           */
/*      case (on the 80x86) the processor's string search instruction   */
/*      costs about half the time per byte of the losb/xlat/jmp         */
/*      sequence at the heart of the Grouse architecture.  In this      */
/*      case, the buffer must include the literal after the final       */
/*      byte of the buffer.  The parameter pTrailingLiteral is          */
/*      provided so that the caller can prepare the buffer correctly.   */
/*                                                                      */
/************************************************************************/
public_scope void
RETable_StartSearchOpt(MATCHENG_SPEC *pSpec, CHAR *pTrailingLiteral)
{
        BYTET *pFlags = ((BYTET *) pSpec) - 3;
        UINT NrStates = 0;
        BYTET *pTab = pSpec->Tables[2];
        BYTET *pStart = pSpec->Tables[1];
        BYTET *pSearch;
        UINT i;
        UINT SkipSize;

        /*Default to no trailing literal reported*/
        *pTrailingLiteral = NUL;

        /*Is the search anchored to the start of the line?*/
        if (pStart[LF] == MARK_LINE_START_PUSH_ADVANCE) {
                /*Yes, not optimised*/
                return;

                /*?? We could insert match LF state...?*/

        }

        /*Use summary of search contained in state flags for analysis*/
        for (; ; pFlags--, NrStates++, pTab += TABLE_SIZE) {
                /*Have we seen the maximum number of states we can handle?*/
                if (NrStates == MATCHENG_BOYER_MOORE_LENGTH) {
                        /*Yes, stop analysing specification*/
                        break;
        
                }
        
                /*Does this table contain nonlinear effects?*/
                if (*pFlags & (ITERATION_FLAG | ZERO_TRIP_FLAG)) {
                        /*Sorry, state table needs > 1 second's thought*/
                        break;

                }

                /*Is this table worth optimising?*/
                switch (TABLE_TYPE(*pFlags)) {
                case LITERAL_TYPE:
                case CLASS_TYPE:
                        /*These are okay*/
                        continue;

                }

                /*Any table which gets us here is NOT okay: finish summary*/
                break;

        }

        /*Did we find any states that might be optimised?*/
        if (NrStates == 0) {
                /*No, give up in disgust*/
                return;
        }

        /*Should we consider byte (literal) search in preference to B-M?*/
        if (NrStates < 3) {
                /*Yes, check for and use byte search if available*/
                if (RETable_UseByteSearch(pSpec, pTrailingLiteral)) {
                        /*Using CPU string search instruction*/
                        return;
                }

                /*Did the search only have one state with a literal?*/
                if (NrStates == 1) {
                        /*Yes, Boyer-Moore does not help: abandon optimising*/
                        return;
                }

        }

        /*Okay, rewrite starting state to work from new initial state*/
        pTab -= TABLE_SIZE;
        pSpec->SearchSkipOffset = NrStates - 1;

        /*Don't process NUL (note loop counter starts at 1)*/
        pTab++;
        pStart++;

        /*Loop through each character of table, looking for improvements*/
        for (i = 1; i < TABLE_SIZE; i++, pStart++, pTab++) {
                /*Does this character match in the new initial state?*/
                if (*pTab != NO_MATCH) {
                        /*Yes, action is start of RE and try to match*/
                        *pStart = START_OFFSET_MATCH;
                        continue;
                }

                /*Find out how far non-match activity extends back*/
                pSearch = pTab;
                for (;;) {
                        /*Does this character match in this table?*/
                        if (*pSearch != NO_MATCH) {
                                /*No, finished looking*/
                                break;
                        }

                        /*Have we hit the front of the RE?*/
                        if (pSearch <= pStart) {
                                /*Yes, search complete*/
                                break;
                        }

                        /*Move to the next previous table*/
                        pSearch -= TABLE_SIZE;

                }

                /*Work out how many characters can be skipped*/
                SkipSize = (UINT) ((pTab - pSearch) / TABLE_SIZE);

                /*Is this skip the maximum possible?*/
                if (SkipSize == NrStates) {
                        /*Yes, use efficient action*/
                        *pStart = SKIP_MAX;
                        continue;
                }

                /*Write search action based on survey*/
                switch (SkipSize) {
                case  1: *pStart = AGAIN; break;
                case  2: *pStart = SKIP2;  break;
                case  3: *pStart = SKIP3;  break;
                case  4: *pStart = SKIP4;  break;
                case  5: *pStart = SKIP5;  break;
                case  6: *pStart = SKIP6;  break;
                case  7: *pStart = SKIP7;  break;
                case  8: *pStart = SKIP8;  break;
                case  9: *pStart = SKIP9;  break;
                case 10: *pStart = SKIP10; break;
                case 11: *pStart = SKIP11; break;
                case 12: *pStart = SKIP12; break;
                case 13: *pStart = SKIP13; break;
                case 14: *pStart = SKIP14; break;
                case 15: *pStart = SKIP15; break;
                case 16: *pStart = SKIP16; break;
                case 17: *pStart = SKIP17; break;
                default:
                       printf("RETable: Unknown StartSearchOpt action?!");
                       abort();
                       break;

                }

        }

} /*StartSearchOpt*/

/************************************************************************/
/*                                                                      */
/*      Debug -- Set debugging options                                  */
/*                                                                      */
/************************************************************************/
public_scope void
RETable_Debug(LWORD OffBits, LWORD OnBits)
{
        /*Edit debugging options according to request*/
        gRETable.Debug = (gRETable.Debug & ~ OffBits) | OnBits;

} /*Debug*/

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/************************************************************************/
public_scope BOOL
RETable_Start(void)
{
        /*Started successfully*/
        return TRUE;

} /*Start*/

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
RETable_Init(void)
{
        /*Initialise configuration parameters to default values*/
        gRETable.Config = 0;

        /*Don't display any debug information*/
        gRETable.Debug = 0;

} /*Init*/
