/************************************************************************/
/*                                                                      */
/*      RETable -- Efficient language for implementing regexp searches  */
/*                                                                      */
/*      This module converts a regular expression description           */
/*      from a compact form (as created by RegExp) to a                 */
/*      high-performance table-driven form (as described by             */
/*      MatchEng).  Several optimisations are available that            */
/*      can significantly improve search performance.  These            */
/*      optimisations fall into two categories: some try to             */
/*      minimise backtracking initiated by iterative states             */
/*      by analysing the state table following the iterative            */
/*      state; others use string or byte searches to improve            */
/*      the speed of scanning the text for possible matches.            */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#ifndef RETABLE_H
#define RETABLE_H

#include <compdef.h>
#include "matcheng.h"
#include "regexp.h"

#ifdef __cplusplus
extern "C" {
#endif /*__cplusplus*/

/*Configuration specifiers*/
#define RETABLE_IGNORE_CASE             BIT0
#define RETABLE_MATCH_WORD_START        BIT1
#define RETABLE_MATCH_WORD_END          BIT2
#define RETABLE_MATCH_WORD              (RETABLE_MATCH_WORD_START | \
                                         RETABLE_MATCH_WORD_END)

/*Debugging options*/
#define RETABLE_D_CONTROL               BIT0

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
void
RETable_Init(void);

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/************************************************************************/
BOOL
RETable_Start(void);

/************************************************************************/
/*                                                                      */
/*      Expand -- Convert "compact" RE codes to optimised table form    */
/*                                                                      */
/*      This function is used to prepare for table-driven pattern       */
/*      match searching.  It creates a fairly verbose form of the       */
/*      regular expression which is highly optimised for match          */
/*      searches.  A typical pattern will occupy 8 to 20 kbytes of      */
/*      RAM in its expanded form.                                       */
/*                                                                      */
/*      The function returns TRUE and a handle for the expanded format  */
/*      or FALSE if unable to prepare an expanded version.              */
/*                                                                      */
/************************************************************************/
BOOL
RETable_Expand(REGEXP_SPEC *pRE, MATCHENG_SPEC **ppSpec);

/************************************************************************/
/*                                                                      */
/*      Match -- Perform search and report success                      */
/*                                                                      */
/************************************************************************/
BOOL
RETable_Match(MATCHENG_SPEC *pSpec, CHAR *pTextStart,
                MATCHENG_DETAILS *pDetails);

/************************************************************************/
/*                                                                      */
/*      Configure -- Modify search behaviour according to selections    */
/*                                                                      */
/*      Although strictly speaking the configuration items provided     */
/*      should be specified via RegExp, in practice it is much more     */
/*      convenient to deal with them in the table-driven format.        */
/*                                                                      */
/************************************************************************/
void
RETable_Configure(LWORD OffBits, LWORD OnBits);

/************************************************************************/
/*                                                                      */
/*      UseBuffer -- Optimise search for multi-line buffer              */
/*                                                                      */
/*      Once the match loop itself has been optimised, the slowest      */
/*      part of the search becomes the process of breaking the          */
/*      file up into lines.  Since the search is almost always on       */
/*      files that already contain multiple lines separated by          */
/*      LF or CR/LF, there are huge benefits in having the search       */
/*      work on the entire buffer.                                      */
/*                                                                      */
/*      This function converts the expanded pattern to work with        */
/*      a buffer where each line is guaranteed to be preceded and       */
/*      terminated by a LF (with an option to include CR as a           */
/*      terminator).  The most important requirements are that a        */
/*      LF be the first char presented (e.g. added to start of          */
/*      buffer) and that the last line of the buffer be terminated      */
/*      by LF and NUL.  Searches anchored to the start of the line      */
/*      then merely require a LF as the first literal.                  */
/*                                                                      */
/*      Extra match engine actions are added so that the search for     */
/*      start of regular expression can count line numbers, plus        */
/*      report start of line on a match to the caller.                  */
/*                                                                      */
/*      Selecting CR as a line terminator is given as a parameter       */
/*      so this behaviour can be varied dynamically.                    */
/*                                                                      */
/*      A few patterns require the line-by-line search.  If             */
/*      UseBuffer is requested for such a pattern, the function         */
/*      returns FALSE.                                                  */
/*                                                                      */
/************************************************************************/
BOOL
RETable_UseBuffer(MATCHENG_SPEC *pSpec, BOOL CRIsTerminator);

/************************************************************************/
/*                                                                      */
/*      StartSearchOpt -- Optimise where RE begins with literal(s)      */
/*                                                                      */
/*      Virtually the entire cost of the average regular expression     */
/*      search is the cost of finding a search starting point.          */
/*      If the client does not require special intelligence during      */
/*      the start search (e.g. no line counting), then many patterns    */
/*      can be optimised by either the Boyer-Moore algorithm or         */
/*      by the string instructions provided by the CPU.                 */
/*                                                                      */
/*      The Boyer-Moore search operates where the pattern starts        */
/*      with two or more literals or classes.  The algorithm works      */
/*      by checking for the last character of the string instead of     */
/*      the first.  This allows the search to skip forward many         */
/*      bytes if the character checked cannot match the string at       */
/*      all.  For example, if searching for "typewriter", a "Q"         */
/*      allows the search to skip forward 10 bytes; a "t" allows        */
/*      the search to skip forward two bytes.                           */
/*                                                                      */
/*      If the Boyer-Moore optimisation cannot be used, the start       */
/*      search can still be optimised if the first state is a           */
/*      simple literal (case sensitive if a letter).  In this           */
/*      case (on the 80x86) the processor's string search instruction   */
/*      costs about half the time per byte of the losb/xlat/jmp         */
/*      sequence at the heart of the Grouse architecture.  In this      */
/*      case, the buffer must include the literal after the final       */
/*      byte of the buffer.  The parameter pTrailingLiteral is          */
/*      provided so that the caller can prepare the buffer correctly.   */
/*                                                                      */
/************************************************************************/
void
RETable_StartSearchOpt(MATCHENG_SPEC *pSpec, CHAR *pTrailingLiteral);

/************************************************************************/
/*                                                                      */
/*      Debug -- Set debug display options                              */
/*                                                                      */
/************************************************************************/
void
RETable_Debug(LWORD OffBits, LWORD OnBits);

#ifdef __cplusplus
}
#endif /*__cplusplus*/

#endif /*RETABLE_H*/
