/************************************************************************/
/*                                                                      */
/*      ScanFile -- Manage RE search for a single file                  */
/*                                                                      */
/*      This module accepts the regular expression to be used           */
/*      plus various search and display options, and performs           */
/*      the requested search on each file given.  It uses LineFile      */
/*      to bring the file in a suitably-edited memory buffer,           */
/*      and RETable to execute the search on each buffer.               */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include "ascii.h"
#include <compdef.h>
#include "linefile.h"
#include "main.h"
#include "matchc.h"
#include "matcheng.h"
#include <memory.h>
#include "retable.h"
#include "scanfile.h"
#include <stdio.h>
#include "tbldisp.h"

#define TRACE(p, s) printf("\n%s", s)

/*Parameters for buffering file into lines*/
#define FILE_BUFFER_SIZE                65024u
#define BYTES_BEFORE_BUFFER             8
#define BYTES_AFTER_BUFFER              (32 - BYTES_BEFORE_BUFFER)

/*Note: BYTES_AFTER_BUFFER must be >= BOYER_MOORE_LOOKAHEAD_MAX*/

/*Codes to specify how to condition the end of the buffer*/
typedef enum {
        ENDWITH_ANYTHING,
        ENDWITH_NUL_PADDING,
        ENDWITH_TRAILING_LITERAL
} END_CONDITION_ENUM;

typedef BOOL (FILE_SCANNER)(void);

typedef struct {
        /*Function to perform regular expression matching*/
        MATCH_FUNCTION pMatchFunc;

        /*Expanded RE specification for searching*/
        MATCHENG_SPEC *pExpandedRE;

        /*Search/match context shared between modules*/
        MATCHENG_DETAILS Details;

        /*Function to handle lines selected by search*/
        SELECTION_FUNCTION *pSelectFunc;

        /*Operation options*/
        BOOL MatchSense;

        /*Function to select file scanning style*/
        FILE_SCANNER *pFileScanner;

        /*RE match function provided by client*/
        MATCH_FUNCTION pExternMatchFunc;

        /*Harder portion of match if RE split into two*/
        MATCHENG_SPEC *pHarderRE;

        /*Flags recording if any lines matched overall and for current file*/
        BOOL MatchedAny;

        /*Variables to condition end of line for optimal search behaviour*/
        END_CONDITION_ENUM EndCondition;
        CHAR TrailingLiteral;

        /*Platform-specific functions to display matches*/
        SELECTION_FUNCTION *pNormalOut;
        SELECTION_FUNCTION *pHighlightOut;
        SELECTION_FUNCTION *pFilenameOut;

        /*Debugging options*/
        LWORD Debug;

} SCANFILE_MODULE_CONTEXT;

module_scope SCANFILE_MODULE_CONTEXT gScanFile;

#if DEBUGGING
#define DEBUG(Item, Msg)           {if (gScanFile.Debug & Item) {printf(Msg);}}
#define DEBUG1(Item, Msg, p1)      {if (gScanFile.Debug & Item) {printf(Msg, p1);}}
#define DEBUG2(Item, Msg, p1, p2)  {if (gScanFile.Debug & Item) {printf(Msg, p1, p2);}}
#else
#define DEBUG(Item, Msg)
#define DEBUG1(Item, Msg, p1)
#define DEBUG2(Item, Msg, p1, p2)
#endif /*DEBUGGING*/

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/************************************************************************/
public_scope BOOL
ScanFile_Start(void)
{
        /*Set up LineFile for optimised buffer and/or line file access*/
        if (! LineFile_Start(BYTES_BEFORE_BUFFER,
                        FILE_BUFFER_SIZE, BYTES_AFTER_BUFFER)) {
                /*Sorry, LineFile was unable to prepare for operation*/
                return FALSE;
        }

        return TRUE;

} /*Start*/

/************************************************************************/
/*                                                                      */
/*      CountMatches -- Function for scan engine when counting          */
/*                                                                      */
/*      This function merely increments the line match counter.         */
/*                                                                      */
/************************************************************************/
module_scope BOOL
ScanFile_CountMatches(MATCHENG_DETAILS *pDetails)
{
        pDetails->LineMatchCount++;

        /*Search should continue*/
        return TRUE;

} /*CountMatches*/

/************************************************************************/
/*                                                                      */
/*      MatchedAbandon -- Halt search if matching line found            */
/*                                                                      */
/*      This function is used for the -L search option.                 */
/*                                                                      */
/************************************************************************/
module_scope BOOL
ScanFile_MatchedAbandon(MATCHENG_DETAILS *pDetails)
{
        /*Search can abandon current file*/
        return FALSE;

} /*MatchedAbandon*/

/************************************************************************/
/*                                                                      */
/*      ScanFileLines -- Break file into lines and perform RE search    */
/*                                                                      */
/************************************************************************/
module_scope BOOL
ScanFile_ScanFileLines(void)
{
        CHAR *pLine;
        BOOL Matched = FALSE;

        /*Set up match details structure for reporting*/
        gScanFile.Details.LineNr = 0;

        /*Loop through each line of the file*/
        while ((pLine = LineFile_GetLine()) != NULL) {
                /*Count lines of file for reporting*/
                gScanFile.Details.LineNr++;

                /*Does this line match (or not) the specified RE?*/
                if (gScanFile.pMatchFunc(gScanFile.pExpandedRE, pLine,
                                &gScanFile.Details) != gScanFile.MatchSense) {
                        /*No, skip to next line*/
                        continue;
                }

                /*At least one match has been found*/
                Matched = TRUE;

                /*Write information for reporting*/
                gScanFile.Details.pLineStart = pLine;

                /*Handle selected line*/
                if (! gScanFile.pSelectFunc(&gScanFile.Details)) {
                        /*Don't continue with the search loop*/
                        return TRUE;

                }

        }

        /*Finished scanning file: report if any matches found*/
        return Matched;

} /*ScanFileLines*/

/************************************************************************/
/*                                                                      */
/*      ScanFileBuffer -- Read file in chunks and search each chunk     */
/*                                                                      */
/************************************************************************/
module_scope BOOL
ScanFile_ScanFileBuffer(void)
{
        UINT NrChars;
        CHAR *pLineEnd;
        CHAR *pBufCurr;
        CHAR *pInBuf;
        BOOL Matched = FALSE;
        BOOL Found;

        /*Set up match details structure for reporting*/
        gScanFile.Details.LineNr = 0;
        gScanFile.Details.BufLineNr = 0;

ReadFile:
        /*Read next buffer of file*/
        NrChars = LineFile_GetBuffer(&pInBuf);
        if (NrChars == 0) {
                /*Finished processing the file*/
                return Matched;
        }

        /*Start buffer from prepended LF to simplify search's job*/
        /*?? This may cause buggy behaviour if RE beins with LF*/
        pBufCurr = pInBuf - 1;
        *pBufCurr = LF;

        /*First character after buffer is always NUL*/
        pInBuf[NrChars++] = NUL;

        switch (gScanFile.EndCondition) {
        case ENDWITH_NUL_PADDING:
                /*Condition end of file for Boyer-Moore search*/
                memset(&pInBuf[NrChars], NUL,
                                MATCHENG_BOYER_MOORE_LENGTH);
                break;

        case ENDWITH_TRAILING_LITERAL:
                /*Add literal to simplify memory search specification*/
                gScanFile.Details.pAfterEndOfBuffer = &pInBuf[NrChars + 1];
                pInBuf[NrChars] = gScanFile.TrailingLiteral;
                break;

        default:
                break;

        }

        /*Loop through buffer, looking for RE matches*/
        for (;;) {
/*
                UINT j;
                for (j = 0; j < 20; j++) {
*/
                Found = gScanFile.pMatchFunc(gScanFile.pExpandedRE, pBufCurr,
                        &gScanFile.Details);
/*
                }
*/
                if (! Found) {
                        break;
                }

                /*Match found: update line count (not necessarily valid)*/
                gScanFile.Details.LineNr += gScanFile.Details.BufLineNr;
                gScanFile.Details.BufLineNr = 0;

                /*At least one match has been found*/
                Matched = TRUE;

                /*Find end of line a tad crudely (should match do this?)*/
                pLineEnd = memchr(gScanFile.Details.pMatchEnd,
                                LF, 0xffff);

                /*Replace LF (and CR) with NUL so line becomes a C string*/
                *pLineEnd = NUL;
                if (pLineEnd[-1] == CR) {
                        pLineEnd[-1] = NUL;
                }

                /*Handle selected line*/
                if (! gScanFile.pSelectFunc(&gScanFile.Details)) {
                        /*Don't continue with the search loop*/
                        return TRUE;

                }

                /*Is end-of-buffer marker after this LF?*/
                if (pLineEnd[1] == NUL) {
                        /*Yes, finished this buffer*/
                        break;
                }

                /*Update search to (LF before) next line*/
                *pLineEnd = LF;
                pBufCurr = pLineEnd;


        }

        /*Accumulate lines counted during search*/
        gScanFile.Details.LineNr += gScanFile.Details.BufLineNr;

        /*Handle next buffer of file, if any*/
        goto ReadFile;

} /*ScanFileBuffer*/

/************************************************************************/
/*                                                                      */
/*      Search -- Perform specified search on a file                    */
/*                                                                      */
/*      This function searches the specified file (or stdin if          */
/*      pFilename is NULL) using the search options specified           */
/*      by Configure.  The function is intended to be called via        */
/*      Glob_EnumerateFiles, and returns FALSE if multi-file            */
/*      searches are to be skipped.                                     */
/*                                                                      */
/************************************************************************/
public_scope BOOL
ScanFile_Search(CHAR *pFilename)
{
        BOOL Matched;

        /*Perform search on this file*/
        if (! LineFile_Open(pFilename)) {
                /*?? Unable to open file*/
                fprintf(stderr, "ggrep: Unable to open: %s\n",
                                pFilename);

                /*Record problem for exit value reporting*/
                Main_ReturnCode(MAIN_RETURN_FAULT);

                /*Skip to next file to process*/
                return TRUE;

        }

        /*Is the filename anything other than stdin?*/
        if (pFilename != NULL) {
                /*Yes, record filename for reporting*/
                gScanFile.Details.pFilename = pFilename;

        } else {
                /*Standard input -- plug in name ourselves*/
                gScanFile.Details.pFilename = "(standard input)";

        }

        /*Search using file scanner appropriate to search optimisations*/
        Matched = gScanFile.pFileScanner();

        /*Were we asked to report if no lines matched within file?*/
        if ((gScanFile.Details.ReportingOptions & DETAILS_NONMATCH_FILES) &&
                        ! Matched) {
                /*Yes, report filename now and proceed to next file*/
                gScanFile.pFilenameOut(&gScanFile.Details);
                gScanFile.MatchedAny = TRUE;
                return TRUE;
        }

        /*Accumulate status of match across all files*/
        if (Matched) {
                gScanFile.MatchedAny = TRUE;
        }

        /*Were we asked to count lines?*/
        if (gScanFile.Details.ReportingOptions &
                        DETAILS_LINECOUNT) {
                /*Yes, report count now (include filename if selected)*/
                gScanFile.pFilenameOut(&gScanFile.Details);
                gScanFile.Details.LineMatchCount = 0;
        }

        /*Request that enumeration continue*/
        return TRUE;

} /*Search*/

/************************************************************************/
/*                                                                      */
/*      MatchDual -- Search for easy pattern and then complex pattern   */
/*                                                                      */
/*      This function performs a pattern search in an efficient         */
/*      fashion.  The easy search will usually be able to span          */
/*      an entire buffer quickly.  Where a matching line is found,      */
/*      the buffer is edited to impose a line break at the start        */
/*      of the matching text, and a second search looks for a match     */
/*      from the start of the line using the more difficult leading     */
/*      part of the RE.                                                 */
/*                                                                      */
/************************************************************************/
public_scope BOOL PAGE_ALIGNED
ScanFile_MatchDual(MATCHENG_SPEC *pTables,
                   CHAR *pTextStart,
                   MATCHENG_DETAILS *pDetails)
{
        CHAR Ch;
        MATCHENG_DETAILS Details2;
        UINT BufLineNr = 0;

        pDetails->pLineStart = NULL;

        /*Look for the easy part of the RE*/
FindEasy:
        if (! gScanFile.pExternMatchFunc(pTables, pTextStart, pDetails)) {
                /*Easy bit of RE was not found*/
                pDetails->BufLineNr += BufLineNr;
                return FALSE;
        }

        /*Check for weakness in Boyer-Moore search*/
        if (pDetails->pMatchStart < pTextStart) {
                /*RETable's Boyer-Moore found spurious match at start*/
                pTextStart++;
                goto FindEasy;
        }

        /*Work out start of line by working back from match start*/
        pDetails->pLineStart = pDetails->pMatchStart;
        while (*--pDetails->pLineStart != LF) {
        }
        pDetails->pLineStart++;
        BufLineNr += pDetails->BufLineNr;

        /*Split line at match start and then attempt to complete match*/
        Ch = *pDetails->pMatchStart;
        *pDetails->pMatchStart = NUL;
        if (! gScanFile.pExternMatchFunc(gScanFile.pHarderRE,
                        pDetails->pLineStart,
                        &Details2)) {
                /*Harder part of RE did not match: undo edit and continue*/
                *pDetails->pMatchStart = Ch;
                pTextStart = pDetails->pMatchStart + 1;

                goto FindEasy;

        }

        /*Both parts of match were found: undo edit and report to caller*/
        *pDetails->pMatchStart = Ch;
        pDetails->pMatchStart = Details2.pMatchStart;
        pDetails->BufLineNr = BufLineNr;
        return TRUE;

} /*MatchDual*/

/************************************************************************/
/*                                                                      */
/*      Pattern -- Specify RE to be searched                            */
/*                                                                      */
/*      pPattern is the "compiled" version created by RegExp.           */
/*      ScanOptions allows modifications to the pattern such as         */
/*      case insensitivity, word match and inverted match sense         */
/*      to be specified.                                                */
/*                                                                      */
/*      Pattern expands the RE into a version optimised for speed,      */
/*      and return FALSE if it is unable to handle the RE               */
/*      (for example, if it runs out of RAM).                           */
/*                                                                      */
/************************************************************************/
public_scope BOOL
ScanFile_Pattern(REGEXP_SPEC *pPattern, LWORD ScanOptions)
{
        REGEXP_SPEC *pEasyBit = NULL;
        REGEXP_SPEC *pHarderBit = NULL;

        /*Does the main search reference a valid RE table specification?*/
        if (pPattern == NULL) {
                /*No, fault in configuration specification*/
                return FALSE;
        }

        /*Alloable optimisations given in ScanOptions*/

        /*Default to no "easy" search before match search*/
        gScanFile.pHarderRE = NULL;

        /*Have we been asked to ignore case?*/
        if (ScanOptions & SCANFILE_IGNORE_CASE) {
                /*Yes, tell RETable*/
                RETable_Configure(0, RETABLE_IGNORE_CASE);
        }

        /*Does the client want us to count lines?*/
        if (ScanOptions & SCANFILE_NUMBER_LINES) {
                /*Yes, can't use search optimisations that skip bytes*/
                ScanOptions &= ~SCANFILE_OPT_SKIP;

        }

        /*Does the client want to report non-matching lines?*/
        gScanFile.MatchSense = TRUE;
        if (ScanOptions & SCANFILE_INVERT_MATCH_SENSE) {
                /*Yes, select inverted match sense*/
                gScanFile.MatchSense = FALSE;

                /*Disable most search optimisations, including buffer search*/
                ScanOptions &= ~(SCANFILE_OPT_BUFFER | \
                                 SCANFILE_OPT_SKIP |
                                 SCANFILE_OPT_EASIEST_FIRST);

        }

        /*Is the client happy with only an approximate match?*/
        if (ScanOptions & SCANFILE_OPT_APPROXIMATE) {
                /*Yes, remove optional first/last elements that slow us down*/
                RegExp_SlashEnds(pPattern);
        }

        /*Is there an easier bit to search in the middle of the RE?*/
        gScanFile.pMatchFunc = gScanFile.pExternMatchFunc;
        if ((ScanOptions & SCANFILE_OPT_EASIEST_FIRST) &&
                        RegExp_EasiestFirst(pPattern, &pEasyBit)) {
                /*Yes, modify search to match that part first*/
                pHarderBit = pPattern;
                pPattern = pEasyBit;

                /*Use function to check for first match after second found*/
                gScanFile.pMatchFunc = ScanFile_MatchDual;

                /*Is the match restricted to words?*/
                if (ScanOptions & SCANFILE_MATCH_WORDS) {
                        /*Yes, ensure harder bit matches word start*/
                        RETable_Configure(0, RETABLE_MATCH_WORD_START);
                }

                /*Expand hard bit to line-based search*/
                if (! RETable_Expand(pHarderBit, &gScanFile.pHarderRE)) {
                        /*Expansion failed for some reason*/
                        return FALSE;
                }

        }

        /*Have we been asked to only match words?*/
        if (ScanOptions & SCANFILE_MATCH_WORDS) {
                /*Yes, are we doing the search in a single step?*/
                if (pHarderBit == NULL) {
                        /*Yes, get RETable to enforce match at start & end*/
                        RETable_Configure(0, RETABLE_MATCH_WORD);

                } else {
                        /*Pattern is second half of match: word end only*/
                        RETable_Configure(RETABLE_MATCH_WORD_START,
                                          RETABLE_MATCH_WORD_END);

                }

        }

        /*Default to line-based search and then try to optimise*/
        gScanFile.EndCondition = ENDWITH_ANYTHING;
        gScanFile.TrailingLiteral = NUL;
        gScanFile.pFileScanner = ScanFile_ScanFileLines;

        /*Expand compact RE spec into table-driven version to scan one line*/
        if (! RETable_Expand(pPattern, &gScanFile.pExpandedRE)) {
                /*Expansion failed for some reason*/
                return FALSE;
        }

        /*Improve search specification based on allowable optimisations*/
        gScanFile.pFileScanner = ScanFile_ScanFileLines;
        if (! (ScanOptions & SCANFILE_OPT_BUFFER)) {
                /*No buffer-oriented optimisations are possible*/
                goto Complete;
        }

        /*Edit search spec to process lines in a buffer*/
        if (! RETable_UseBuffer(gScanFile.pExpandedRE, TRUE)) {
                /*RETable was unable to create buffer-oriented version*/
                goto Complete;
        }

        /*Scan buffers without breaking into lines*/
        gScanFile.pFileScanner = ScanFile_ScanFileBuffer;

        /*Buffer needs NUL terminator: use Boyer-Moore NUL padding*/
        gScanFile.EndCondition = ENDWITH_NUL_PADDING;

        /*Are we allowed to attempt optimisations that skip bytes?*/
        if (ScanOptions & SCANFILE_OPT_SKIP) {
                /*Yes, try to optimise search start*/
                RETable_StartSearchOpt(gScanFile.pExpandedRE,
                                &gScanFile.TrailingLiteral);

                /*Did the optimisation use Boyer-Moore?*/
                if (gScanFile.TrailingLiteral != NUL) {
                        /*No, string search: buf ends with literal*/
                        gScanFile.EndCondition =
                                        ENDWITH_TRAILING_LITERAL;
                }

        }

        /*Completed pattern preparation*/
Complete:

        /*Display tables if requested*/
        if (ScanOptions & SCANFILE_DEBUG_DISPLAY) {
                TblDisp_Describe(gScanFile.pExpandedRE, "Scan");

                /*Is there a second (harder) part of the RE to search?*/
                if (gScanFile.pHarderRE != NULL) {
                        /*Yes, display it as well*/
                        TblDisp_Describe(gScanFile.pHarderRE, "Match");

                }

        }

        /*Report success to caller*/
        return TRUE;

} /*Pattern*/

/************************************************************************/
/*                                                                      */
/*      Configure -- Define how the module searches and reports matches */
/*                                                                      */
/************************************************************************/
public_scope BOOL
ScanFile_Configure(LWORD ReportingOptions)
{
        /*Record reporting options and select functions accordingly*/
        gScanFile.Details.ReportingOptions = ReportingOptions;

        /*Default to normal reporting of lines*/
        gScanFile.pSelectFunc = gScanFile.pNormalOut;

        /*Does the client want to see lines?*/
        if (! (ReportingOptions & DETAILS_LINE)) {
                /*No, just show filename on match and finish file*/
                gScanFile.pSelectFunc = gScanFile.pFilenameOut;
        }

        /*Does the client want line counting?*/
        if (ReportingOptions & DETAILS_LINECOUNT) {
                /*Yes, use optimised function for simple case*/
                gScanFile.pSelectFunc = ScanFile_CountMatches;
        }

        /*Does the client want highlighted matches?*/
        if (ReportingOptions & DETAILS_HIGHLIGHT) {
                /*Yes, use platform-specific function*/
                gScanFile.pSelectFunc = gScanFile.pHighlightOut;
        }

        /*Has the client requested that matches be reported with delimiters?*/
        if (ReportingOptions & DETAILS_MARKER_FLAG) {
                /*Yes, use platform-specific function and unpack char*/
                gScanFile.pSelectFunc = gScanFile.pHighlightOut;
                gScanFile.Details.MarkerChar =
                                DETAILS_MARKER_UNPACK(ReportingOptions);
        }

        /*Dos the client want to see nonmatch files?*/
        if (ReportingOptions & DETAILS_NONMATCH_FILES) {
                /*Yes, abandon file as soon as match found*/
                gScanFile.pSelectFunc = ScanFile_MatchedAbandon;
        }

        /*Configured module successfully*/
        return TRUE;

} /*Configure*/

/************************************************************************/
/*                                                                      */
/*      MatchedAny -- Report if any files matched search criteria       */
/*                                                                      */
/************************************************************************/
public_scope BOOL
ScanFile_MatchedAny(void)
{
        /*Was internal flag set?*/
        if (gScanFile.MatchedAny) {
                /*Yes, clear it and then report state to caller*/
                gScanFile.MatchedAny = FALSE;
                return TRUE;
        }

        /*Internal flag was not set*/
        return FALSE;

} /*MatchedAny*/

/************************************************************************/
/*                                                                      */
/*      MatchFunction -- Define routine to perform match                */
/*                                                                      */
/*      ScanFile wishes to provide extreme high-performance searches    */
/*      to do so in a very portable fashion.  This function is the      */
/*      result: ScanFile receives the address of the function that      */
/*      implements the match from an outsider (usually Platform).       */
/*      This function must be called before Pattern.                    */
/*                                                                      */
/************************************************************************/
public_scope void
ScanFile_MatchFunction(MATCH_FUNCTION pMatchFunc)
{
        /*Remember function for operation*/
        gScanFile.pExternMatchFunc = pMatchFunc;

} /*MatchFunction*/

/************************************************************************/
/*                                                                      */
/*      NoMatchFunction -- Place-holder to warn of incorrect config     */
/*                                                                      */
/*      This function is called if ScanFile does not receive a          */
/*      match function appropriate to the platform.                     */
/*                                                                      */
/************************************************************************/
public_scope BOOL PAGE_ALIGNED
ScanFile_NoMatchFunction(MATCHENG_SPEC *pTable, CHAR *pText,
                         MATCHENG_DETAILS *pDetails)
{
        fprintf(stderr, "Scanfile: No match function provided!");

        return FALSE;

} /*NoMatchFunction*/

/************************************************************************/
/*                                                                      */
/*      OutputFunctions -- Specify functions to perform match output    */
/*                                                                      */
/*      In order to keep ScanFile as portable as possible, the          */
/*      match and filename display functions are provided by            */
/*      an external party, since generation and display of              */
/*      output (and especially highlighting) is platform-specific.      */
/*                                                                      */
/*      This function must be called after Init but before any          */
/*      RE specification or module configuration.                       */
/*                                                                      */
/************************************************************************/
public_scope void
ScanFile_OutputFunctions(SELECTION_FUNCTION *pNormal,
                         SELECTION_FUNCTION *pHighlight,
                         SELECTION_FUNCTION *pFilenameOut)
{
        /*Record the functions to use for later*/
        gScanFile.pNormalOut = pNormal;
        gScanFile.pHighlightOut = pHighlight;
        gScanFile.pFilenameOut = pFilenameOut;

} /*OutputFunctions*/

/************************************************************************/
/*                                                                      */
/*      Debug -- Set debugging options                                  */
/*                                                                      */
/************************************************************************/
public_scope void
ScanFile_Debug(LWORD OffBits, LWORD OnBits)
{
        /*Edit debug options according to request*/
        gScanFile.Debug = (gScanFile.Debug & ~OffBits) | OnBits;

} /*Debug*/

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
ScanFile_Init(void)
{
        /*Initialise selection reporting details*/
        gScanFile.Details.LineMatchCount = 0;

        /*No match function provided initially*/
        gScanFile.pMatchFunc = ScanFile_NoMatchFunction;
        gScanFile.pExpandedRE = NIL;

        /*File scanner defaults to line-based version*/
        gScanFile.pFileScanner = ScanFile_ScanFileLines;

        /*Default end option is minimal effort*/
        gScanFile.EndCondition = ENDWITH_ANYTHING;
        gScanFile.TrailingLiteral = NUL;

        /*Configure default scanning options (?)*/
        gScanFile.MatchSense = TRUE;

        /*No platform-specific display functions provided yet*/
        gScanFile.pNormalOut = NULLFUNC;
        gScanFile.pHighlightOut = NULLFUNC;
        gScanFile.pFilenameOut = NULLFUNC;
        gScanFile.pSelectFunc = NULLFUNC;

        /*Don't display any debug information*/
        gScanFile.Debug = 0;

} /*Init*/
