/************************************************************************/
/*                                                                      */
/*      ScanFile -- Manage RE search for a single file                  */
/*                                                                      */
/*      This module accepts the regular expression to be used           */
/*      plus various search and display options, and performs           */
/*      the requested search on each file given.  It uses LineFile      */
/*      to bring the file in a suitably-edited memory buffer,           */
/*      and RETable to execute the search on each buffer.               */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#ifndef SCANFILE_H
#define SCANFILE_H

#include <compdef.h>
#include "matcheng.h"
#include "regexp.h"

#ifdef __cplusplus
extern "C" {
#endif /*__cplusplus*/

/*Configuration options*/
#define SCANFILE_NUMBER_LINES           BIT0
#define SCANFILE_INVERT_MATCH_SENSE     BIT1
#define SCANFILE_MATCH_WORDS            BIT2
#define SCANFILE_IGNORE_CASE            BIT3
#define SCANFILE_DEBUG_DISPLAY          BIT4
#define SCANFILE_OPT_APPROXIMATE        BIT5
#define SCANFILE_OPT_EASIEST_FIRST      BIT6
#define SCANFILE_OPT_BUFFER             BIT7
#define SCANFILE_OPT_SKIP               BIT9

/*Debugging options*/
#define SCANFILE_D_CONTROL              BIT0
#define SCANFILE_D_TABLE                BIT1

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
void
ScanFile_Init(void);

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/************************************************************************/
BOOL
ScanFile_Start(void);

/************************************************************************/
/*                                                                      */
/*      OutputFunctions -- Specify functions to perform match output    */
/*                                                                      */
/*      In order to keep ScanFile as portable as possible, the          */
/*      match and filename display functions are provided by            */
/*      an external party, since generation and display of              */
/*      output (and especially highlighting) is platform-specific.      */
/*                                                                      */
/*      This function must be called after Init but before any          */
/*      RE specification or module configuration.                       */
/*                                                                      */
/************************************************************************/
void
ScanFile_OutputFunctions(SELECTION_FUNCTION *pNormal,
                         SELECTION_FUNCTION *pHighlight,
                         SELECTION_FUNCTION *pFilenameOnly);

/************************************************************************/
/*                                                                      */
/*      MatchFunction -- Define routine to perform match                */
/*                                                                      */
/*      ScanFile wishes to provide very high-performance searches       */
/*      but do so in a very portable fashion.  This function is the     */
/*      result: ScanFile receives the address of the function that      */
/*      implements the match from an outsider (usually Platform).       */
/*      This function must be called before Pattern.                    */
/*                                                                      */
/************************************************************************/
void
ScanFile_MatchFunction(MATCH_FUNCTION pMatchFunc);

/************************************************************************/
/*                                                                      */
/*      Pattern -- Specify RE to be searched                            */
/*                                                                      */
/*      pPattern is the "compiled" version created by RegExp.           */
/*      ScanOptions allows modifications to the pattern such as         */
/*      case insensitivity, word match and inverted match sense         */
/*      to be specified.                                                */
/*                                                                      */
/*      Pattern expands the RE into a version optimised for speed,      */
/*      and return FALSE if it is unable to handle the RE               */
/*      (for example, if it runs out of RAM).                           */
/*                                                                      */
/*      This function takes control of the RE specification             */
/*      supplied and deletes it when it is no longer required.          */
/*                                                                      */
/************************************************************************/
BOOL
ScanFile_Pattern(REGEXP_SPEC *pPattern, LWORD ScanOptions);

/************************************************************************/
/*                                                                      */
/*      Configure -- Define how the module searches and reports matches */
/*                                                                      */
/*      ReportingOptions defines the way matching lines are displayed.  */
/*                                                                      */
/************************************************************************/
BOOL
ScanFile_Configure(LWORD ReportingOptions);

/************************************************************************/
/*                                                                      */
/*      Search -- Perform specified search on a file                    */
/*                                                                      */
/*      This function searches the specified file (or stdin if          */
/*      pFilename is NULL) using the search options specified           */
/*      by Configure.  The function is intended to be called via        */
/*      Glob_EnumerateFiles, and returns FALSE if multi-file            */
/*      searches are to be skipped.                                     */
/*                                                                      */
/************************************************************************/
BOOL
ScanFile_Search(CHAR *pFilename);

/************************************************************************/
/*                                                                      */
/*      MatchedAny -- Report if any files matched search criteria       */
/*                                                                      */
/************************************************************************/
BOOL
ScanFile_MatchedAny(void);

/************************************************************************/
/*                                                                      */
/*      Debug -- Set debugging options                                  */
/*                                                                      */
/************************************************************************/
void
ScanFile_Debug(LWORD OffBits, LWORD OnBits);

#ifdef __cplusplus
}
#endif /*__cplusplus*/

#endif /*SCANFILE_H*/
