/************************************************************************/
/*                                                                      */
/*      TblDisp -- Describe state tables in concise, readable format    */
/*                                                                      */
/*      This module presents the state tables created for a             */
/*      search in a human-readable format.  This is useful for          */
/*      understanding, debugging and testing the software.              */
/*                                                                      */
/*      Describe was created very late in the development --            */
/*      over two years after the project started.  It has               */
/*      quickly become the main tool for measuring the program's        */
/*      implementation of the RE, and has highlighted a number          */
/*      of (minor) cases where the search could be optimised            */
/*      further.                                                        */
/*                                                                      */
/*      Copyright (C) 1995-7 Grouse Software.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include "bakslash.h"
#include "compdef.h"
#include <string.h>
#include <stdio.h>
#include "tbldisp.h"

/*Summary of action entries for Description*/
typedef struct {
        BYTE Code;
        UINT8 NrEntries;
        WORD RangeSpecs[TABLE_SIZE / 2];

} ActionDescription;

/*Table of action entries to complete the description for each state*/
typedef struct {
        UINT NrActions;
        ActionDescription Action[MATCHENG_ACTIONS_MAX];

} TableDescription;

/*List of action codes and labels*/

typedef struct {
        BYTET Code;
        CHAR *pLabel;
} ActionLabelItem;

module_scope ActionLabelItem gTblDisp_Labels[] = {

        /*Actions for start-match state*/
        SKIP_MAX,               "SkipMax",
        SKIP17,                 "Skip17",
        SKIP16,                 "Skip16",
        SKIP15,                 "Skip15",
        SKIP14,                 "Skip14",
        SKIP13,                 "Skip13",
        SKIP12,                 "Skip12",
        SKIP11,                 "Skip11",
        SKIP10,                 "Skip10",
        SKIP9,                  "Skip9",
        SKIP8,                  "Skip8",
        SKIP7,                  "Skip7",
        SKIP6,                  "Skip6",
        SKIP5,                  "Skip5",
        SKIP4,                  "Skip4",
        SKIP3,                  "Skip3",
        SKIP2,                  "Skip2",
        AGAIN,                  "Again",
        BYTE_SEARCH,            "ByteSearch",
        START_MATCH,            "StartMatch",
        START_MATCH_PUSH,       "StartMatchPush",
        START_MATCH_PUSH_ADVANCE, "StartMatchPushAdvance",
        START_OFFSET_MATCH,     "StartOffsetMatch",
        MARK_LINE_START,        "MarkLineStart",
        MARK_LINE_START_PUSH_ADVANCE, "MarkLineStartPushAdvance",

        /*Actions that advance the match*/
        ADVANCE,                "Advance",
        AGAIN_PUSH_ADVANCE,     "AgainPushAdvance",
        BACK_AND_ADVANCE,       "BackAndAdvance",
        SKIP_NEXT_TABLE,        "SkipNext",
        ADVANCE_PUSH_ZERO,      "AdvancePushZero",

        /*Actions that complete search*/
        COMPLETED,              "Completed",
        COMPLETE_IF_WORD_START, "CompleteIfWordStart",

        /*Entries that don't match*/
        NO_MATCH,               "NoMatch",
        ABANDON,                "Abandon",

        /*And finally, action reserved for future special effects*/
        SPECIAL,                "Special",

        0, NULL

};

/************************************************************************/
/*                                                                      */
/*      DispFlag -- Report table summary described by flag variable     */
/*                                                                      */
/************************************************************************/
module_scope void
TblDisp_DispFlag(BYTET Flag)
{
        /*Display flag summary*/
        switch (TABLE_TYPE(Flag)) {
        case FAIL_TYPE:
                fprintf(stdout, "Fail");
                break;
        case START_TYPE:
                fprintf(stdout, "Start");
                break;
        case LITERAL_TYPE:
                fprintf(stdout, "Literal");
                break;
        case CLASS_TYPE:
                fprintf(stdout, "Class");
                break;
        case ALL_BUT_TYPE:
                fprintf(stdout, "AllBut");
                break;
        case MATCH_ANY_TYPE:
                fprintf(stdout, "MatchAny");
                break;
        case SUCCESS_TYPE:
                fprintf(stdout, "Success");
                break;
        default:
                fprintf(stdout, "?? unknown type");
                break;

        }

        /*Report options included in flag*/
        if (Flag & ITERATION_FLAG) {
                fprintf(stdout, "+iter");
        }
        if (Flag & ZERO_TRIP_FLAG) {
                fprintf(stdout, "+0trip");
        }

} /*DispFlag*/

/************************************************************************/
/*                                                                      */
/*      Display -- Show table contents for debugging                    */
/*                                                                      */
/************************************************************************/
public_scope void
TblDisp_Display(MATCHENG_SPEC *pSpec)
{
        BYTET *pTab = (BYTET *) &pSpec->Tables[0];
        BYTET *pFlags = ((BYTET *) pSpec) - 1;
        UINT Row;
        UINT Col;
        UINT TableNr = 0;
        CHAR Actions[TABLE_SIZE];

        /*Prepare 1-char labels for actions*/
        memset(Actions, '?', sizeof(Actions));

        Actions[SKIP_MAX]                     = 'M';
        Actions[SKIP17]                       = '+';
        Actions[SKIP16]                       = '+';
        Actions[SKIP15]                       = '+';
        Actions[SKIP14]                       = '+';
        Actions[SKIP13]                       = '+';
        Actions[SKIP12]                       = '+';
        Actions[SKIP11]                       = '+';
        Actions[SKIP10]                       = '+';
        Actions[SKIP9]                        = '9';
        Actions[SKIP8]                        = '8';
        Actions[SKIP7]                        = '7';
        Actions[SKIP6]                        = '6';
        Actions[SKIP5]                        = '5';
        Actions[SKIP4]                        = '4';
        Actions[SKIP3]                        = '3';
        Actions[SKIP2]                        = '2';
        Actions[AGAIN]                        = 'a';
        Actions[ADVANCE]                      = 'v';
        Actions[START_MATCH]                  = 'S';
        Actions[START_MATCH_PUSH]             = 'P';
        Actions[START_MATCH_PUSH_ADVANCE]     = 'A';
        Actions[AGAIN_PUSH_ADVANCE]           = 'r';
        Actions[BACK_AND_ADVANCE]             = 'b';
        Actions[SKIP_NEXT_TABLE]              = 'j';
        Actions[ADVANCE_PUSH_ZERO]            = 'o';
        Actions[NO_MATCH]                     = 'x';
        Actions[MARK_LINE_START]              = 'l';
        Actions[MARK_LINE_START_PUSH_ADVANCE] = 'L';
        Actions[START_OFFSET_MATCH]           = 'O';
        Actions[BYTE_SEARCH]                  = 'B';
        Actions[ABANDON]                      = 'X';
        Actions[COMPLETE_IF_WORD_START]       = 'W';
        Actions[SPECIAL]                      = '~';
        Actions[COMPLETED]                    = '!';

        for (;;) {
                printf("\n\nTable %u: ", TableNr++);
                TblDisp_DispFlag(*pFlags);

                for (Row = 0; Row < TABLE_SIZE; Row += 64) {
                        fprintf(stdout, "\n%1c%1c ", Row, Row + 15);
                        for (Col = 0; Col < 16; Col++) {
                                fprintf(stdout, "%c", Actions[*pTab++]);
                        }
                        fprintf(stdout, " %1c%1c ", Row+16, Row + 31);
                        for (Col = 16; Col < 32; Col++) {
                                fprintf(stdout, "%c", Actions[*pTab++]);
                        }
                        fprintf(stdout, " %1c%1c ", Row+32, Row + 47);
                        for (Col = 32; Col < 48; Col++) {
                                fprintf(stdout, "%c", Actions[*pTab++]);
                        }
                        fprintf(stdout, " %1c%1c ", Row+48, Row + 63);
                        for (Col = 48; Col < 64; Col++) {
                                fprintf(stdout, "%c", Actions[*pTab++]);
                        }
                }

                /*Is this the last table of the expression?*/
                if (TABLE_TYPE(*pFlags--) == SUCCESS_TYPE) {
                        /*Yes, exit from loop now*/
                        break;
                }

        }

        /*Add line termination to output*/
        fprintf(stdout, "\n");

} /*Display*/

/************************************************************************/
/*                                                                      */
/*      ActionSummary -- Display summary of entries matching action     */
/*                                                                      */
/*      If the action named here is present in the description          */
/*      table, the bytes matching the action are displayed.             */
/*      C-style escapes are used to describe nonprinting values.        */
/*                                                                      */
/************************************************************************/
module_scope void
TblDisp_ActionSummary(BYTET Code, CHAR *pLabel, TableDescription *pDesc)
{
        ActionDescription *pAct;
        ActionDescription *pEnd;
        UINT Range;
        BYTE High;
        BYTE Low;
        UINT TextLen;
        CHAR Encoding[8];

        /*Search for code within description table*/
        pEnd = &pDesc->Action[pDesc->NrActions];
        pAct = &pDesc->Action[0];
        for (;;) {
                /*Have we exhausted the list?*/
                if (pAct == pEnd) {
                       /*Yes, code isn't present in list*/
                       return;
                }

                /*Have we found an entry for this code?*/
                if (pAct->Code == Code) {
                        /*Yes, found entry*/
                        break;
                }

                /*Move to the next action, if any*/
                pAct++;

        }

        /*Found entry, report it*/
        TextLen = 20;
        fprintf(stdout, "%s: ", pLabel);
        if (strlen(pLabel) < 18) {
                /*Pad label to improve alignment of reports*/
                fprintf(stdout, "%*s", 18 - strlen(pLabel), "");

        } else {
                /*Keep track of line length*/
                TextLen += strlen(pLabel) + 2;

        }

        for (Range = 0; Range < pAct->NrEntries; Range++) {
                /*Do we have space for the next entry?*/
                if (TextLen > 68) {
                        /*Possibly not... move to a new line*/
                        fprintf(stdout, "\n%28s", "");
                        TextLen = 29;
                }

                /*Report entry as one of "'a'", "'a', 'b'" or "'a'-'b'"*/
                High = HIBYTE(pAct->RangeSpecs[Range]);
                Low =  LOBYTE(pAct->RangeSpecs[Range]);
                if (High == Low) {
                        /*Range represents a single item*/
                        BakSlash_EncodeByte(Low, Encoding);
                        TextLen += fprintf(stdout, "'%s'", Encoding);
                } else if (High == (Low + 1)) {
                        BakSlash_EncodeByte(Low, Encoding);
                        TextLen += fprintf(stdout, "'%s', ", Encoding);
                        BakSlash_EncodeByte(High, Encoding);
                        TextLen += fprintf(stdout, "'%s'", Encoding);
                } else {
                        BakSlash_EncodeByte(Low, Encoding);
                        TextLen += fprintf(stdout, "'%s'-", Encoding);
                        BakSlash_EncodeByte(High, Encoding);
                        TextLen += fprintf(stdout, "'%s'", Encoding);
                }

                /*Add delimiter if not at the end of the list*/
                if ((Range + 1) < pAct->NrEntries) {
                        TextLen += fprintf(stdout, ", ");
                }

        }

        fprintf(stdout, "\n");

} /*ActionSummary*/

/************************************************************************/
/*                                                                      */
/*      DescribeTable -- Generate readable description of a table       */
/*                                                                      */
/*      The description is built by organising the entries by           */
/*      action code, and by identifying ranges within each action.      */
/*      The result is a table with variable-length rows: each row       */
/*      describes an action, and the entries in each row describe       */
/*      ranges of characters within the rows.                           */
/*                                                                      */
/************************************************************************/
module_scope void
TblDisp_DescribeTable(BYTET *pTab)
{
        TableDescription Desc;
        ActionDescription *pAction;
        ActionDescription *pAfterLastAction;
        UINT i;
        BYTET Act;
        WORD *pRange;
        ActionLabelItem *pItem;

        /*Initialise all entries to 0*/
        memset(&Desc, 0, sizeof(Desc));

        /*Work through each byte of table, recording action occurrences*/
        for (i = 0; i < 256; i++) {
                Act = *pTab++;
                pAction = &Desc.Action[0];
                pAfterLastAction = &Desc.Action[Desc.NrActions];
                for (;;) {
                        /*Have we run out of actions?*/
                        if (pAction == pAfterLastAction) {
                                /*Yes, need to create a new entry*/
                                Desc.NrActions++;
                                break;
                        }

                        /*Have we found the action?*/
                        if (Act == pAction->Code) {
                                /*Yes, finish search with desired entry*/
                                break;
                        }

                        /*Next entry, if any*/
                        pAction++;

                }

                /*Is this the first entry for this description?*/
                pRange = &pAction->RangeSpecs[pAction->NrEntries - 1];
                if (pAction->NrEntries == 0) {
                        /*Yes, create the first entry now*/
                        pAction->Code = Act;
                        pAction->NrEntries++;
                        *++pRange = BYTES2WORD(i, i);
                        continue;
                }

                /*Does this entry extend the last range?*/
                if ((i - 1) == HIBYTE(*pRange)) {
                        /*Yes, merely extend range*/
                        *pRange = BYTES2WORD(i, LOBYTE(*pRange));
                        continue;
                }

                /*Need to add another range to action's description*/
                pAction->Code = Act;
                pAction->NrEntries++;
                *++pRange = BYTES2WORD(i, i);

        }

        /*Report table entries in fairly coherent order*/
        for (pItem = &gTblDisp_Labels[0]; pItem->pLabel != NULL; pItem++) {
                TblDisp_ActionSummary(pItem->Code, pItem->pLabel, &Desc);
        }

} /*DescribeTable*/

/************************************************************************/
/*                                                                      */
/*      Describe -- Report state tables in coherent fashion             */
/*                                                                      */
/*      This routine is springing into existence as part of the         */
/*      white-box testing of the system.  It describes each state       */
/*      table in concise but readable format.  This output is           */
/*      checked by part of the automated test rig, so don't             */
/*      change this routine unless you're willing to update the         */
/*      tests.                                                          */
/*                                                                      */
/************************************************************************/
public_scope void
TblDisp_Describe(MATCHENG_SPEC *pSpec, CHAR *pTitle)
{
        BYTET *pTab = (BYTET *) &pSpec->Tables[0];
        BYTET *pFlags = ((BYTET *) pSpec) - 1;
        UINT TableNr = 0;

        /*Display the title for this specification*/
        fprintf(stdout, "\n========== %s ==========\n", pTitle);

        /*Work through each table of the spec in turn*/
        for (;;) {
                fprintf(stdout, "[Table %u, ", ++TableNr);
                TblDisp_DispFlag(*pFlags);
                fprintf(stdout, "]\n");

                /*Generate the description for this table*/
                TblDisp_DescribeTable(pTab);
                fprintf(stdout, "\n");

                /*Was that the final table?*/
                if (*pFlags-- == SUCCESS_TYPE) {
                        /*Yes, finished displaying*/
                        break;
                }

                /*Move to the next table*/
                pTab += TABLE_SIZE;

        }

} /*Describe*/

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/************************************************************************/
public_scope BOOL
TblDisp_Start(void)
{
        /*Started successfully*/
        return TRUE;

} /*Start*/

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
TblDisp_Init(void)
{
} /*Init*/
