/************************************************************************/
/*                                                                      */
/*      CDecode -- C version of table-driven decoder                    */
/*                                                                      */
/*      Copyright (C) Grouse Software 1997.  All rights reserved.       */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include <ascii.h>
#include "ccodec.h"
#include <compdef.h>
#include <hufftree.h>
#include <limits.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>

#define NRBYTES                         (UCHAR_MAX + 1)
#define TABLES_MAX                      254

/*Encode actions using values corresponding to assembly entry points*/

#define USE1                            0x00
#define USE2                            0x1c
#define USE3                            0x38
#define USE4                            0x54
#define USE5                            0x70
#define USE6                            0x8c
#define USE7                            0xa8
#define USE8                            0xc4
#define STOP                            0xe0
#define USEGT8                          0xe4

module_scope UINT CCodec_DecodeAct[] = {
              STOP, /*placeholder for index 0*/
              USE1, USE2, USE3, USE4, USE5, USE6, USE7, USE8
};

struct CCodec_TableStruct {
        BYTE Table[TABLES_MAX][NRBYTES];
};

typedef struct {
        /*Hacky allocation tracking variable for table creation*/
        UINT Index;

        /*Hacky storage for end code as we're too lazy to pass it round*/
        WORD EndCode;

        /*Extremely hacky storage to allow two binary nums to be displayed*/
        CHAR Buf1[64];
        CHAR Buf2[64];
        BOOL Toggle;

} CCODEC_MODULE_CONTEXT;

module_scope CCODEC_MODULE_CONTEXT gCCodec;

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
CCodec_Init(void)
{
        gCCodec.Toggle = 0;
} /*Init*/

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing THAT WHICH has to be managed            */
/*                                                                      */
/************************************************************************/
public_scope BOOL
CCodec_Start(void)
{
        /*Nothing much yet*/
        return TRUE;

} /*Start*/

/************************************************************************/
/*                                                                      */
/*      b -- Report binary value in a string for debugging              */
/*                                                                      */
/************************************************************************/
module_scope CHAR *
b(UINT Len, UINT32 bits)
{
        CHAR *pBuf;

        /*Trim length if unreasonable*/
        if ((Len < 1) || (Len > 16)) Len = 1;

        /*Toggle buffer used for storing output*/
        gCCodec.Toggle = ! gCCodec.Toggle;
        if (gCCodec.Toggle) {
                pBuf = &gCCodec.Buf1[32];
        } else {
                pBuf = &gCCodec.Buf2[32];
        }

        /*Expand code into digits*/
        *pBuf = NUL;
        sprintf(pBuf, "(%02x)", LOBYTE(bits));
        do {
                *--pBuf = '0' + (bits & BIT0);
                bits >>= 1;
        } while (--Len);

        return pBuf;

} /*b*/

/************************************************************************/
/*                                                                      */
/*      AddToTable -- Add symbol to table-driven decoder                */
/*                                                                      */
/************************************************************************/
module_scope void
CCodec_AddToTable(CCODEC_TABLE pTab, HuffTree_CodeItem *pEncoding)
{
        UINT NrXEntries;
        BYTE *pStateTable;
        UINT Index;
        BYTE *pBits;
        UINT NrBits;
        UINT Action;
        BYTE Value;

        /*Work through encoding a byte at a time*/
        pStateTable = &pTab->Table[0][0];
        NrBits = pEncoding->NrBits;
        pBits = pEncoding->BitBuf;

        /*Create tree of state tables for encodings > 8 bits long*/
        while (NrBits > 8) {
                /*Have we handled a code with this prefix before?*/
                if (pStateTable[*pBits] != USEGT8) {
                        /*No, allocate new decoding state context*/
                        gCCodec.Index += 2;
                        Index = gCCodec.Index;

                        /*Have we run out of table space in 64k segment?*/
                        if (Index == TABLES_MAX) {
                                /*Yes, regretfully inform caller can't do*/
                                printf("CCodec_AddToTable: Sorry, but initial code can't handle >64k tables, sigh.\n");
                                exit(27);
                        }

                        /*Write code > 8 action in current table*/
                        pStateTable[*pBits] = USEGT8;
                        pStateTable[(*pBits) + NRBYTES] = Index;

                } else {
                        /*Already coded another symbol: use same table*/
                        Index = pStateTable[(*pBits) + NRBYTES];
                }

                /*Descend into next state coding lower bits*/
                pStateTable = &pTab->Table[Index][0];
                NrBits -= 8;
                pBits++;

        }

        /*Set each table entry matching bits.xxxx to code length*/
        NrXEntries = 1 << (8 - NrBits);
        if (pEncoding->Symbol != gCCodec.EndCode) {
                Action = CCodec_DecodeAct[NrBits];
                Value = pEncoding->Symbol;
        } else {
                Action = STOP;
                Value = NrBits - 1;
        }

        /*Initialise group of table entries matching bit specification*/
        memset(&pStateTable[*pBits],             Action, NrXEntries);
        memset(&pStateTable[(*pBits) + NRBYTES], Value,  NrXEntries);

} /*AddToTable*/

/************************************************************************/
/*                                                                      */
/*      PrepDecode -- Prepare table-driven decoder given code list      */
/*                                                                      */
/************************************************************************/
public_scope BOOL
CCodec_PrepDecode(UINT NrCodes, WORD EndCode, HuffTree_CodeItem *pCodeList,
                                CCODEC_TABLE *pTable)
{
        CCODEC_TABLE pTab;

        /*Destroy return value to reduce chances of misinterpretation*/
        *pTable = NIL;

        /*Acquire memory for table*/
        pTab = malloc(sizeof(struct CCodec_TableStruct));
        if (pTab == NULL) {
                /*Sorry, unable to acquire memory to create table*/
                return FALSE;
        }

        /*Initialise entries with any value other than CODE_GT_8*/
        memset(pTab, STOP, sizeof(struct CCodec_TableStruct));

        /*Use rather ugly global variable to track table allocation*/
        gCCodec.Index = 0;

        /*Use another ugly global to remember end code value*/
        gCCodec.EndCode = EndCode;

        /*Process each code of list*/
        while (NrCodes--) {
                /*Add code to table*/
                CCodec_AddToTable(pTab, pCodeList);

                /*Move to next item of list, if any*/
                pCodeList++;

        }

        /*Write created table to caller*/
        *pTable = pTab;

} /*PrepDecode*/

/************************************************************************/
/*                                                                      */
/*      Trace -- Show state information to help trace operation         */
/*                                                                      */
/************************************************************************/
module_scope void
CCodec_Trace(UINT Act, BYTE Out, BYTE Next8, WORD Reg, UINT BitAlign, BYTE In0, BYTE In1)
{
        printf("Act %02x", Act);
        if ((Act != USEGT8) && (Act != STOP)) {
                printf(", emit %02x ", Out);
                if ((Out >= 0x20) && (Out < 0x7f)) {
                        printf("%c ", Out);
                } else {
                        printf("  ");
                }

        } else {
                 printf("            ");
        }

        printf("%s %d %s: ", b(8, HIBYTE(Reg)), BitAlign, b(8, Next8));
        printf("%s %s\n", b(8, In0), b(8, In1));

} /*Trace*/

/************************************************************************/
/*                                                                      */
/*      Destroy -- Get rid of table-driven coder                        */
/*                                                                      */
/************************************************************************/
public_scope void
CCodec_Destroy(CCODEC_TABLE pTab)
{
        (void) free(pTab);

} /*Destroy*/

#include "ccodec.inc"
