/************************************************************************/
/*                                                                      */
/*      GHuff -- Test program for testing behoffski's table-driven Huffman coder and decode using the by-now ubiquitous Grouse Finite State Architecture (yawn)       */
/*                                                                      */
/*      (Apologies for using such a flippant title above!)              */
/*                                                                      */
/*      This program provides a demonstration of a high-speed           */
/*      Huffman decoder implemented using the Grouse FSA.               */
/*      The program allows files to be compressed and decompressed,     */
/*      including repeated decoding runs so that the performance        */
/*      of the decoder can be evaluated.                                */
/*                                                                      */
/*      WARNING: You run this program entirely at your own risk.        */
/*                                                                      */
/*      The program is not entirely robust: the main weakness is        */
/*      that the decoder tries to be fast at the expense of being       */
/*      careful with memory.  This means that if an invalid file        */
/*      is decoded, the decoder will probably crash, and memory         */
/*      will probably be corrupted.  This could be handled in           */
/*      real situations by arranging buffers so the memory              */
/*      manager triggers an address exception if the program            */
/*      tries to step outside its allocated area, but this hasn't       */
/*      been attempted here since the program is built for a            */
/*      real-mode DOS environment only.                                 */
/*                                                                      */
/*      Copyright (C) Grouse Software 1997.  All rights reserved.       */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include "ccodec.h"
#include "acodec.h"
#include <compdef.h>
#include <filefreq.h>
#include "ghuff.h"
#include "hufftree.h"
#include <io.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define ELEMENTS                        256
#define CODE_ITEMS_MAX                  (ELEMENTS + 8)

#define RAW_SIZE                        16384u
#define SQUISHED_SIZE                   24576u

#define FLOGS_PER_READ                  100

typedef struct {
        UINT32 Freqs[ELEMENTS + 1];
        HuffTree_CodeItem CodeList[CODE_ITEMS_MAX];

        CCODEC_TABLE DecodeTable;

        BOOL Write;
        BOOL FlogInternals;

        BYTE MaxFlogs;

        BOOL UseAssembly;

        CHAR Name[FILENAME_MAX];

} GHUFF_MODULE_CONTEXT;

module_scope GHUFF_MODULE_CONTEXT gGHuff;

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
GHuff_Init(void)
{
        /*Initialise default parameters*/
        gGHuff.Write = TRUE;
        gGHuff.FlogInternals = FALSE;
        gGHuff.UseAssembly = TRUE;

} /*Init*/

/************************************************************************/
/*                                                                      */
/*      Usage -- Tell user how to use us and then abort                 */
/*                                                                      */
/************************************************************************/
module_scope void
GHuff_Usage(void)
{
        fprintf(stdout, "Usage: ghuff ok e  <source> <squish>  to encode a file\n");
        fprintf(stdout, "       ghuff ok d  <squish> <dest>    to decode a file (asm)\n");
        fprintf(stdout, "       ghuff ok D  <squish> <dest>    decode (C version)\n");
        fprintf(stdout, "       ghuff ok dt <squish>           time asm decoder\n");
        fprintf(stdout, "       ghuff ok Dt <squish>           time C decoder\n");
        fprintf(stdout, "\n");
        fprintf(stdout, "Where  <source> is data file to be encoded\n");
        fprintf(stdout, "       <squish> is name (no extension) for encoded output\n");
        fprintf(stdout, "       <dest>   is decoded file\n");
        fprintf(stdout, "\n");
        fprintf(stdout, "Add 't' to time decode: no file write, decodes %u times\n",
                        FLOGS_PER_READ);
        fprintf(stdout, "\n");
        fprintf(stdout, "ghuff v1.00 Copyright (C) 1997 by Grouse Software.  All rights reserved.\n");
        fprintf(stdout, "Written for Grouse by behoffski (Brenton Hoff).\n");

        exit(51);

} /*Usage*/

/************************************************************************/
/*                                                                      */
/*      Warn -- Tell user that this program might be unreliable         */
/*                                                                      */
/************************************************************************/
module_scope void
GHuff_Warn(void)
{
        fprintf(stdout, "\n");
        fprintf(stdout, "WARNING: This is an experimental/demonstration program.  It has the\n");
        fprintf(stdout, "potential to damage your system.  Do NOT use this program unless\n");
        fprintf(stdout, "you are in an environment where this program cannot damage any\n");
        fprintf(stdout, "valuable data.\n");
        fprintf(stdout, "\n");
        fprintf(stdout, "If you agree to this restriction, you may skip this warning by\n");
        fprintf(stdout, "adding the word \"ok\" after the program name but before any\n");
        fprintf(stdout, "arguments.  This indicates you agree to run the program\n");
        fprintf(stdout, "entirely at your own risk.  The commands become:\n");
        fprintf(stdout, "\n");
        fprintf(stdout, "        ghuff ok e <source> <squish>   to encode\n");
        fprintf(stdout, "        ghuff ok d <squish> <dest>     to decode (asm)\n");
        fprintf(stdout, "\n");
        fprintf(stdout, "Use \"ghuff ok ?\" to get a brief command summary.\n");
        fprintf(stdout, "\n");

} /*Warn*/

/************************************************************************/
/*                                                                      */
/*      CodeFile -- Encode file using Huffman tree                      */
/*                                                                      */
/************************************************************************/
module_scope BOOL
GHuff_CodeFile(HuffTree_CodeItem *pCodeList, CHAR *pOriginalName,
                                             FILE *g)
{
        FILE *f;
        BYTE *pInBuf;
        BYTE *pOutBuf;
        UINT NrBytes;
        UINT NrSquish;
        UINT32 Raw = 0;
        UINT32 Squished = 0;

        /*Open input file, and fail if we can't*/
        f = fopen(pOriginalName, "rb");
        if (f == NULL) {
                printf("GHuff: Can't open: %s\n", pOriginalName);
                return FALSE;
        }

        /*Allocate buffers for operation*/
        pInBuf = malloc(RAW_SIZE);
        pOutBuf = malloc(SQUISHED_SIZE);

        /*Work through file in largish blocks*/
        for (;;) {
                /*Get next slab of data from file, if any*/
                NrBytes = _read(fileno(f), pInBuf, RAW_SIZE);

                /*Run data through encoder*/
                Raw += NrBytes;
                HuffTree_Encode(pCodeList, NrBytes, ELEMENTS, pInBuf,
                                &NrSquish, pOutBuf);
                Squished += NrSquish;

                /*Did we overrun the output buffer?*/
                if (NrSquish > SQUISHED_SIZE) {
                        /*Yes, panic: system integrity is suspect*/
                        printf("Squished data too big!  (%d)\n", NrSquish);
                        exit(88);
                }

                /*Write squished data to output file*/
                if (gGHuff.Write) {
                        fwrite(pOutBuf, 1, NrSquish, g);
                }

                /*Have we processed all the file's contents?*/
                if (eof(fileno(f))) {
                        /*Yes, finished encoding*/
                        break;
                }

        }

        /*Close input and output files*/
        fclose(f);
        fclose(g);

        /*Deallocate buffers*/
        free(pOutBuf);
        free(pInBuf);

        printf("File %s: Size %lu, squished %lu",
                     pOriginalName, Raw, Squished);
        if (Raw != 0) {
                printf(" (%4.2f bits/symbol)\n", (Squished * 8.0) / Raw);
        } else {
                printf("\n");
        }

        return TRUE;

} /*CodeFile*/

/************************************************************************/
/*                                                                      */
/*      DecodeFile -- Decode what Encode encoded                        */
/*                                                                      */
/************************************************************************/
module_scope BOOL
GHuff_DecodeFile(FILE *f, CHAR *pOutputName)
{
        FILE *g = NULL;
        BYTE *pInBuf;
        BYTE *pOutBuf;
        UINT NrBytes;
        UINT NrRaw;
        UINT SquishUsed;
        UINT32 Pos;
        UINT32 TotalUnsquish = 0;

        /*Create output file*/
        if (gGHuff.Write) {
                g = fopen(pOutputName, "wb");
                if (g == NULL) {
                        printf("GHuff: Unable to write %s\n", pOutputName);
                        exit(6);
                }
        }

        /*Allocate input/output buffers (both must be in same segment)*/
        pInBuf = malloc(SQUISHED_SIZE + SQUISHED_SIZE);
        pOutBuf = pInBuf + SQUISHED_SIZE;

        if (pInBuf == NULL) {
                printf("GHuff: Unable to allocate buffer for unsquish\n");
                exit(49);
        }

        if (gGHuff.FlogInternals) {
                gGHuff.MaxFlogs = FLOGS_PER_READ;
                fprintf(stdout, " (%u times)...\n", gGHuff.MaxFlogs);
        } else {
                gGHuff.MaxFlogs = 1;
                fprintf(stdout, "...\n");
        }

        /*Refresh file position to ensure libraries are in sync*/
        Pos = (UINT32) ftell(f);
        fseek(f, Pos, SEEK_SET);

        /*Work through file in largish blocks*/
        for (;;) {
                /*Remember block start as decoder consumption is variable*/
                Pos = (UINT32) ftell(f);

                /*Get next slab of data from file, if any*/
                NrBytes = _read(fileno(f), pInBuf, SQUISHED_SIZE);

                /*Have we processed all the file's contents?*/
                if (NrBytes == 0) {
                        /*Yes, finished encoding*/
                        break;
                }

                /*Expand next section of file*/
                {UINT Flog; for (Flog = 0; Flog < gGHuff.MaxFlogs; Flog++) {
                        NrRaw = 0;
                        if (gGHuff.UseAssembly) {
                                SquishUsed = ACodec_Decode(gGHuff.DecodeTable,
                                                pInBuf, &NrRaw, pOutBuf);

                        } else {
                                SquishUsed = CCodec_Decode(gGHuff.DecodeTable,
                                                pInBuf, &NrRaw, pOutBuf);
                        }

                        TotalUnsquish += NrRaw;

                }}

                /*Write decoded data to output file*/
                if (gGHuff.Write) {
                        write(fileno(g), pOutBuf, NrRaw);
                }

                /*Did we overrun the output buffer?*/
                if (NrRaw > RAW_SIZE) {
                        /*Yes, panic: system integrity is suspect*/
                        printf("GHuff: Decoded data too big!  (%d)\n", NrRaw);
                        exit(88);
                }

                /*Reposition input to start of the next block*/
                fseek(f, Pos + SquishUsed - 1, SEEK_SET);

        }

        /*Close input and output files*/
        if (f != NULL) {
                fclose(f);
        }
        if (g != NULL) {
                fclose(g);
        }

        /*Deallocate buffers*/
        free(pOutBuf);
        free(pInBuf);

        printf("Total unsquished: %lu\n", TotalUnsquish);

        return TRUE;

} /*DecodeFile*/

/************************************************************************/
/*                                                                      */
/*      main -- Control program execution                               */
/*                                                                      */
/************************************************************************/
public_scope int
GHuff_main(int argc, char **argv)
{
        UINT NrCodes;
        void *pTree;
        BOOL DoneAnything = FALSE;
        FILE *f;

        /*Skip over program name argument*/
        argc--;
        argv++;

        /*Has caller indicated they're happy to risk executing the program?*/
        if (strcmp(*argv, "ok") != 0) {
                /*No, issue warning and exit*/
                GHuff_Warn();
                exit(254);
        }
        argc--;
        argv++;

        /*Have we been asked to decode, perhaps including timing?*/
        if ((strcmp(*argv, "d") == 0) ||
                        (strcmp(*argv, "D") == 0) ||
                        (strcmp(*argv, "dt") == 0) ||
                        (strcmp(*argv, "Dt") == 0)) {

                if ((*argv)[0] == 'D') {
                        gGHuff.UseAssembly = FALSE;
                }

                /*Have we been asked to time the decoder's operation?*/
                if ((*argv)[1] == 't') {
                        /*Yes, suppress writing and repeatedly decode*/
                        gGHuff.Write = FALSE;
                        gGHuff.FlogInternals = TRUE;

                        /*Have we been given the right number of parameters?*/
                        if (argc != 2) {
                                GHuff_Usage();
                        }

                } else {
                        /*No, actual decode: need squished and dest params*/
                        if (argc != 3) {
                                GHuff_Usage();
                        }

                }

                DoneAnything = TRUE;
                argv++;
                argc--;
                strcpy(gGHuff.Name, argv[0]);

                /*Open file containing encoded information*/
                f = fopen(gGHuff.Name, "rb");
                if (f == NULL) {
                        /*Unable to open file*/
                        fprintf(stdout, "GHuff: Can't open %s\n",
                                        gGHuff.Name);
                        return 35;
                }

                /*Read codes from file*/
                if (! HuffTree_ReadList(f, gGHuff.CodeList,
                                                      &NrCodes)) {
                        fprintf(stdout, "Can't read from %s\n", gGHuff.Name);
                        return 28;
                }

                /*Convert code list into high-speed table format*/
                CCodec_PrepDecode(NrCodes, ELEMENTS, gGHuff.CodeList,
                                                     &gGHuff.DecodeTable);

                fprintf(stdout, "Decoding %s", gGHuff.Name);
                if (gGHuff.Write) {
                        fprintf(stdout, " into %s", (CHAR *) argv[1]);
                }

                /*Decode file data*/
                if (! GHuff_DecodeFile(f, (CHAR *) argv[1])) {
                        fprintf(stdout, "GHuff: Can't decode %s into %s\n",
                                       argv[0], argv[1]);
                        return 90;
                }

        }

        /*Have we been asked to encode?*/
        if (strcmp(*argv, "e") == 0) {
                DoneAnything = TRUE;
                argv++;
                argc--;

                /*Have we been given the right number of parameters?*/
                if (argc != 2) {
                        GHuff_Usage();
                }

                /*Can we cope with specified filename?*/
                if (strlen(argv[1]) > (sizeof(gGHuff.Name) - 5)) {
                        /*No, cannot manipulate name properly*/
                        printf("GHuff: Cannot cope with filename %s\n",
                                       argv[1]);
                        return 35;
                }
                strcpy(gGHuff.Name, argv[1]);

                memset(gGHuff.Freqs, 0, sizeof(gGHuff.Freqs));

                printf("\nCalculating byte frequencies...");
                if (! FileFreq_Analyse(argv[0], gGHuff.Freqs)) {
                        fprintf(stderr, "GHuff: Unable to process: %s\n",
                                                argv[0]);
                        return 62;
                }

                printf("\nCreating Huff tree");
                if (! HuffTree_Create(ELEMENTS + 1, &gGHuff.Freqs[0],
                                &pTree)) {
                        printf("Unable to create tree.\n");
                        return 88;
                }

                /*Extract codings as a list and prepare codec*/
                printf("\nConvert tree to symbol/code list...");
                NrCodes = HuffTree_SymList(pTree, gGHuff.CodeList);

                /*Open specified file to receive list*/
                f = fopen(gGHuff.Name, "wb");
                if (f == NULL) {
                        /*Unable to open file*/
                        return 73;
                }

                /*Save coding table for decompression*/
                if (! HuffTree_WriteList(f, NrCodes, gGHuff.CodeList)) {
                        /*Unable to save codes for reconstuction*/
                        fclose(f);
                        fprintf(stdout, "Unable to write %s\n", gGHuff.Name);
                        return 72;
                }

                /*Encode file data using coding table*/
                printf("\nEncoding file...\n");
                strcpy(gGHuff.Name, argv[1]);
                GHuff_CodeFile(gGHuff.CodeList, argv[0], f);

                fclose(f);

        }

        /*Has the caller asked for a summary of the commands?*/
        if (strcmp(*argv, "?") == 0) {
                /*Yes, give some help*/
                DoneAnything = TRUE;
                GHuff_Usage();
        }

        /*Did we comprehend the parameters successfully?*/
        if (! DoneAnything) {
                /*No, complain bitterly*/
                printf("GHuff: Didn't comprehend parameters\n");
                printf("\n");

                GHuff_Usage();
        }

        return 54;

} /*main*/
