/************************************************************************/
/*                                                                      */
/*      HuffTree -- Manipulate Huffman trees                            */
/*                                                                      */
/*      In the same way that "regexp" was a warm-up for REs, this       */
/*      module is a warm-up for trying the Grouse FSA as a way of       */
/*      getting ultra-fast table-driven Huffman coders/decoders.        */
/*                                                                      */
/*      Copyright (C) Grouse Software 1997.  All rights reserved.       */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include <ascii.h>
#include <compdef.h>
#include "hufftree.h"
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DEBUG                           0
#define DEBUG_ENCODE                    0

#define END_FREQ                        0xffffffffuL
#define SYM_POINTERS_MAX                256

typedef struct NodeStruct {
        struct NodeStruct *pLeft;
        struct NodeStruct *pRight;
        UINT32 Freq;
        UINT Symbol;
        BYTE Depth;
} Node;

typedef void NodeFn(Node *pNode, UINT NrBits);

typedef struct {
        /*VERY ugly variables to help with tree traversals*/
        NodeFn *pNodeFn;
        HuffTree_CodeItem *pCodeList;
        UINT NrCodes;
        BYTE BitBuf[HUFFTREE_BIT_BYTES_MAX];

        /*Another ugly global buffer to optimise encoding speed*/
        HuffTree_CodeItem *pSymCodes[SYM_POINTERS_MAX];


} HUFFTREE_MODULE_CONTEXT;

module_scope HUFFTREE_MODULE_CONTEXT gHuffTree;

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
HuffTree_Init(void)
{
        gHuffTree.pNodeFn = (NodeFn *) NULL;
} /*Init*/

/************************************************************************/
/*                                                                      */
/*      CompareNode -- QSort comparison function to provide freq order  */
/*                                                                      */
/************************************************************************/
module_scope int
HuffTree_CompareNode(const void *e1, const void *e2)
{
        Node *pA = *((Node **) e1);
        Node *pB = *((Node **) e2);

        if (pA->Freq < pB->Freq) {
                return -1;
        } else if (pA->Freq > pB->Freq) {
                return 1;
        } else {
                return 0;
        }

} /*CompareNode*/

/************************************************************************/
/*                                                                      */
/*      Create -- Form tree, given byte frequencies                     */
/*                                                                      */
/************************************************************************/
public_scope BOOL
HuffTree_Create(UINT NrSymbols, UINT32 *pFreqs, HuffTree_Tree **ppTree)
{
        UINT Sym;
        Node *pStart;
        Node *pNode;
        Node *Pool[1024];
        Node **pSyms;
        Node **pTrees;
        Node **pTreesEnd;
        Node *pSmall1;
        Node *pSmall2;
        Node *pTemp;

        /*Destroy return value to reduce chances of misinterpretation*/
        *ppTree = NIL;

        /*Grab memory to store nodes*/
        pStart = malloc(sizeof(Node) * 1024);

        /*Did we get the memory for the node list?*/
        if (pStart == NULL) {
                /*No, cannot create*/
                return FALSE;
        }

        memset(Pool, 0x80, sizeof(Pool));
        memset(pStart, 0xe0, sizeof(Node) * 512);

        pNode = pStart;

        /*Create pool of nodes containing each symbol*/
        for (Sym = 0; Sym < NrSymbols; Sym++) {
                /*Create node containing symbol*/
                pNode->pLeft = NULL;
                pNode->pRight = NULL;
                pNode->Symbol = Sym;
                pNode->Freq = *pFreqs++;
                pNode->Depth = 0;

                Pool[Sym] = pNode++;

        }

        /*Sort list into ascending frequency order*/
        qsort(&Pool[0], NrSymbols, sizeof(Pool[0]), HuffTree_CompareNode);

        #if DEBUG
        printf("\nAfter sorting:");
        for (Sym = 0; Sym < NrSymbols; Sym++) {
                printf("\nLeaf node %p, L %p, R %p, Freq %8lu  Sym: %3d",
                            Pool[Sym], Pool[Sym]->pLeft, Pool[Sym]->pRight,
                            Pool[Sym]->Freq, Pool[Sym]->Symbol);

        }
        #endif /*DEBUG*/

        /*Mark the end of the symbol list with a special node*/
        pNode->Freq = END_FREQ;
        pNode->pLeft = NULL;
        pNode->pRight = NULL;
        Pool[NrSymbols] = pNode;

        /*Limit the tree list using the same endmarker node*/
        pTrees = &Pool[NrSymbols + 1];
        *pTrees = pNode++;
        pTreesEnd = pTrees;
        pSyms = &Pool[0];
        for (;;) {
                /*Find next two smallest nodes from trees and/or sym lists*/
                if ((*pSyms)->Freq <= (*pTrees)->Freq) {
                        pSmall1 = *pSyms++;
                } else {
                        pSmall1 = *pTrees++;
                }
                if ((*pSyms)->Freq <= (*pTrees)->Freq) {
                        pSmall2 = *pSyms++;
                } else {
                        pSmall2 = *pTrees++;
                }

                #if DEBUG
                printf("\nSmall1 %lu Small2 %lu pSyms %p %lu pTrees %p %lu",
                    pSmall1->Freq, pSmall2->Freq,
                    pSyms, (*pSyms)->Freq, pTrees, (*pTrees)->Freq);
                #endif /*DEBUG*/

                /*Have we reached the end of the tree list?*/
                if (pSmall2->Freq == END_FREQ) {
                        /*Yes, finished constructing*/
                        break;
                }

                /*Is the first subtree nested deeper than the second?*/
                if (pSmall1->Depth > pSmall2->Depth) {
                        /*Yes, rearrange to collect deep trees together*/
                        pTemp = pSmall1;
                        pSmall1 = pSmall2;
                        pSmall2 = pTemp;
                }

                /*Combine two smallest nodes into bigger tree*/
                pNode->pLeft = pSmall1;
                pNode->pRight = pSmall2;
                pNode->Freq = pSmall1->Freq + pSmall2->Freq;
                pNode->Depth = pSmall2->Depth + 1;

                /*Insert new largest node at end of list plus new endmarker*/
                *pTreesEnd++ = pNode++;
                *pTreesEnd = Pool[NrSymbols];

        }

        /*Last node created is the root*/
        *ppTree = (HuffTree_Tree *) *--pTreesEnd;

        /*Created tree successfully*/
        return TRUE;

} /*Create*/

/************************************************************************/
/*                                                                      */
/*      Find -- Return position of symbol in list of codes              */
/*                                                                      */
/************************************************************************/
module_scope HuffTree_CodeItem *
HuffTree_Find(HuffTree_CodeItem *pCodeList, WORD Symbol)
{
        /*Is the code finable from the array?*/
        if (Symbol < SYM_POINTERS_MAX) {
                /*Yes, use pointer directly, it's MUCH faster*/
                return gHuffTree.pSymCodes[Symbol];
        }

        /*Use stupid, primitive linear search to find symbol in list*/
        while (pCodeList->Symbol != Symbol) {
                pCodeList++;
        }

        /*Does not even check for end of list?!*/

        return pCodeList;

} /*Find*/

/************************************************************************/
/*                                                                      */
/*      Encode -- Encode data using coding tree                         */
/*                                                                      */
/************************************************************************/
public_scope void
HuffTree_Encode(HuffTree_CodeItem *pCodeList,
                               UINT NrItems, WORD EndSymbol,
                               BYTE *pInput,
                               UINT *pNrSquished, BYTE *pSquished)
{
        WORD Buf = 0;
        UINT BitsFilled = 0;
        HuffTree_CodeItem *pNode;
        BYTE *pBits;
        UINT BitShift;
        BYTE *pSquishStart = pSquished;
        WORD NextSym;

        #if DEBUG_ENCODE
        printf("\nEmit:");
        #endif /*DEBUG_ENCODE*/
        NrItems++;
        while (NrItems--) {
                if (NrItems != 0) {
                        /*Code next byte of input*/
                        NextSym = (WORD) *pInput++;
                } else {
                        /*Finished buffer data -- encode endmarker*/
                        NextSym = EndSymbol;
                }
                pNode = HuffTree_Find(pCodeList, NextSym);
                pBits = pNode->BitBuf;

                #if DEBUG_ENCODE
                printf("\nCode, Len: %02x, %2d ", NextSym, pNode->NrBits);
                #endif /*DEBUG_ENCODE*/

                /*Add encoded bits to output stream*/
                BitShift = 8 - BitsFilled;
                BitsFilled += pNode->NrBits;

                /*Add code to bitstream and emit completed bytes*/
                for (;;) {
                        /*Add next part of bitstream to output buffer*/
                        Buf |= ((WORD) *pBits++) << BitShift;
                        #if DEBUG_ENCODE
                        printf("(%02x %04x)", pBits[-1], Buf);
                        #endif /*DEBUG_ENCODE*/

                        /*Have we emitted all completed bytes?*/
                        if (BitsFilled < 8) {
                                /*Yes, exit with partially-filled byte*/
                                break;
                        }

                        /*Write byte to output buffer*/
                        *pSquished++ = HIBYTE(Buf);
                        #if DEBUG_ENCODE
                        printf(" %02x ", HIBYTE(Buf));
                        #endif /*DEBUG_ENCODE*/

                        BitsFilled -= 8;
                        Buf <<= 8;

                }

        }

        #if DEBUG_ENCODE
        printf("\n");
        printf("Bits filled after encode: %u\n", BitsFilled);
        #endif /*DEBUG_ENCODE*/

        /*Do we have any bits lying around waiting to be sent?*/
        if (BitsFilled != 0) {
                /*Emit final bits not yet transmitted*/
                *pSquished++ = HIBYTE(Buf);
                #if DEBUG_ENCODE
                printf("emit %02x\n", HIBYTE(Buf));
                #endif /*DEBUG_ENCODE*/
        }

        /*Write count of compressed bytes to caller*/
        *pNrSquished = pSquished - pSquishStart;

} /*Encode*/

/************************************************************************/
/*                                                                      */
/*      Traverse -- Traverse tree, execute function on each leaf        */
/*                                                                      */
/************************************************************************/
module_scope void
HuffTree_Traverse(Node *pNode, UINT NrBits)
{
        BYTE *pCurr = &gHuffTree.BitBuf[NrBits / 8];
        BYTE Bitmask = 1 << (7 - (NrBits % 8));

        /*Traverse left node, if any*/
        if (pNode->pLeft != NULL) {
                HuffTree_Traverse(pNode->pLeft, NrBits + 1);
        }

        if (pNode->pRight != NULL) {
                *pCurr |= Bitmask;
                HuffTree_Traverse(pNode->pRight, NrBits + 1);
                *pCurr &= ~Bitmask;
        } else if (pNode->pLeft == NULL) {
                gHuffTree.pNodeFn(pNode, NrBits);
        }

} /*Traverse*/

/************************************************************************/
/*                                                                      */
/*      ShowBits -- Show variable-length bit code                       */
/*                                                                      */
/************************************************************************/
module_scope void
HuffTree_ShowBits(BYTE Bits, INT Digits)
{
        CHAR s[34];
        CHAR *pS = &s[0];

        do {
              *pS++ = '0' + (Bits >> 7);
              Bits <<= 1;
        } while (--Digits > 0);
        *pS = NUL;

        printf("%s", s);

} /*ShowBits*/

UINT32 SquishBits;
UINT32 UnsquishBits;

/************************************************************************/
/*                                                                      */
/*      DisplayNode -- Display node (as helper to tree traversal)       */
/*                                                                      */
/************************************************************************/
module_scope void
HuffTree_DisplayNode(Node *pNode, UINT NrBits)
{
        BYTE *pBits;

        /*Accumulate size of compressed bitstream and original stream*/
        SquishBits += NrBits * pNode->Freq;
        UnsquishBits += pNode->Freq;

        /*Display symbol and encoding*/
        pBits = gHuffTree.BitBuf;
        printf("\nSym %5d(%2d): ", pNode->Symbol, NrBits);
        while (NrBits > 8) {
                HuffTree_ShowBits(*pBits++, 8);
                NrBits -= 8;
        }

        HuffTree_ShowBits(*pBits, NrBits);

} /*DisplayNode*/

/************************************************************************/
/*                                                                      */
/*      Show -- Show tree in text format                                */
/*                                                                      */
/************************************************************************/
public_scope void
HuffTree_Show(HuffTree_Tree *pTree)
{
        gHuffTree.pNodeFn = HuffTree_DisplayNode;
        SquishBits = 0;
        UnsquishBits = 0;

        memset(gHuffTree.BitBuf, 0, sizeof(gHuffTree.BitBuf));
        HuffTree_Traverse((Node *) pTree, 0);

        printf("\n\nBit count unsquished, squished: %8lu %8lu\n",
                        UnsquishBits * 8, SquishBits);

} /*Show*/

/************************************************************************/
/*                                                                      */
/*      BuildList -- Describe tree codings in symbol order as a list    */
/*                                                                      */
/************************************************************************/
module_scope void
HuffTree_BuildList(Node *pNode, UINT NrBits)
{
        HuffTree_CodeItem *pItem = gHuffTree.pCodeList++;

        /*Create list entry based on node plus traversal information*/
        pItem->Flags = 0;
        pItem->NrBits = NrBits;
        pItem->Symbol = pNode->Symbol;
        memcpy(pItem->BitBuf, gHuffTree.BitBuf, sizeof(pItem->BitBuf));

        /*Is this node likely to be encoded often?*/
        if (pItem->Symbol < SYM_POINTERS_MAX) {
                /*Yes, remember direct reference for speed*/
                gHuffTree.pSymCodes[pItem->Symbol] = pItem;
        }

        gHuffTree.NrCodes++;

} /*BuildList*/

/************************************************************************/
/*                                                                      */
/*      SymList -- Provide description of tree as list                  */
/*                                                                      */
/************************************************************************/
public_scope UINT
HuffTree_SymList(HuffTree_Tree *pTree, HuffTree_CodeItem *pCodeList)
{
        /*Traverse tree, writing details of each node into list*/
        gHuffTree.pNodeFn = HuffTree_BuildList;
        gHuffTree.pCodeList = pCodeList;
        gHuffTree.NrCodes = 0;
        memset(gHuffTree.BitBuf, 0, sizeof(gHuffTree.BitBuf));

        /*Clear out buffer to optimise byte references*/
        /*(Strictly speaking, should assign 0 to pointer, not zero memory)*/
        memset(gHuffTree.pSymCodes, 0, sizeof(gHuffTree.pSymCodes));


        HuffTree_Traverse((Node *) pTree, 0);

        #if DEBUG
        /*Display list for debugging*/
        printf("\n");
        {UINT i,j; for (i = 0; i < gHuffTree.NrCodes; i++) {
                printf("%5d(%2d):", pCodeList->Symbol, pCodeList->NrBits);
                for (j = 0; (j * 8) < pCodeList->NrBits ;j++) {
                        printf(" %02x", pCodeList->BitBuf[j]);
                }
                printf("\n");

                pCodeList++;
        }}
        #endif /*DEBUG*/

        return gHuffTree.NrCodes;

} /*SymList*/

/************************************************************************/
/*                                                                      */
/*      WriteList -- Save symbol list in file                           */
/*                                                                      */
/************************************************************************/
public_scope BOOL
HuffTree_WriteList(FILE *f, UINT NrItems, HuffTree_CodeItem *pCodeList)
{
        UINT i;
        UINT j;

        /*Write list details to file*/
        fprintf(f, "%d\n", NrItems);
        for (i = 0; i < NrItems; i++) {
                fprintf(f, "%5d (%2d):", pCodeList->Symbol, pCodeList->NrBits);
                for (j = 0; (j * 8) < pCodeList->NrBits ;j++) {
                        fprintf(f, " %02x", pCodeList->BitBuf[j]);
                }
                if (0 == fprintf(f, "\n")) {
                        /*Error writing to output file*/
                        return FALSE;
                }

                pCodeList++;
        }

        return TRUE;

} /*WriteList*/

/************************************************************************/
/*                                                                      */
/*      ReadList -- Read symbol list from file                          */
/*                                                                      */
/************************************************************************/
public_scope BOOL
HuffTree_ReadList(FILE *f, HuffTree_CodeItem *pCodeList,
                                   UINT *pNrCodes)
{
        UINT j;
        UINT NrCodes = 0;
        UINT NrBits;
        UINT NrItems;

        /*Read list details from file*/
        fscanf(f, "%d", &NrItems);
        for (;;) {
                /*Read symbol value and nr bits in the code*/
                if (fscanf(f, "%5d (%2d):",
                                &pCodeList->Symbol, &NrBits) == 0) {
                        /*Error reading from file*/
                        return FALSE;
                }


                /*Read the bits of the code*/
                pCodeList->NrBits = (UINT8) NrBits;
                for (j = 0; (j * 8) < pCodeList->NrBits ;j++) {
                        fscanf(f, " %02x", &pCodeList->BitBuf[j]);
                }

                pCodeList++;

                /*Have we read all the codes?*/
                if (++NrCodes == NrItems) {
                        /*Yes, stop looking at file*/
                        break;
                }


        }

        /*Skip LF after final code*/
        (void) fgetc(f);

        /*Report how many codes were found*/
        *pNrCodes = NrCodes;
        return TRUE;

} /*ReadList*/
