/************************************************************************/
/*                                                                      */
/*      Glob -- Perform file globbing, with Unix/DOS flavours           */
/*                                                                      */
/*      This module helps non-Unix programs handle files in a           */
/*      similar fashion to programs executed under a Unix shell.        */
/*      Wildcard specifications are collected by repeated calls to      */
/*      AddSpec.  Configuation options such as including                */
/*      subdirectories may be specified before or during the            */
/*      collection.                                                     */
/*                                                                      */
/*      Once all specs have been described, call Expand which           */
/*      finds matching files and sorts them into alphabetical           */
/*      order.  The number of files found by Expand is reported.        */
/*      However, if subdirectories are included, this count is          */
/*      reported as 9999.                                               */
/*                                                                      */
/*      After expanding, call EnumerateFiles to step through            */
/*      the files found.  EnumerateFiles will also descend into         */
/*      subdirectories if this option was requested, but will not       */
/*      try to perform any file name sorting.                           */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#ifndef GLOB_H
#define GLOB_H

#include <compdef.h>

#ifdef __cplusplus
extern "C" {
#endif /*__cplusplus*/

/*Define configuration options as a set of bits*/

#define GLOB_SUPPRESS_ERRORS            BIT0
#define GLOB_HALT_ON_ERROR              BIT1
#define GLOB_INCLUDE_SUBDIRECTORIES     BIT2

/*Function type for client to receive file names: return FALSE to abandon*/
typedef BOOL GLOB_ENUM_FUNCTION(CHAR *pFileName);

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
void
Glob_Init(void);

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/*      pProgName allows Glob to produce messages prefixed by the       */
/*      program name.  SpecsMax and ExpandBytesMax allow the client     */
/*      to have some control over memory consumption, but allow         */
/*      the implementation to choose very simple memory management      */
/*      strategies.                                                     */
/*                                                                      */
/*      The function returns FALSE if the requested resources were      */
/*      not available.                                                  */
/*                                                                      */
/************************************************************************/
BOOL
Glob_Start(CHAR *pProgName, UINT SpecsMax, UINT ExpandBytesMax);

/************************************************************************/
/*                                                                      */
/*      Configure -- Set how the module reacts to important events      */
/*                                                                      */
/*      Some behaviours of the module may change from run to run,       */
/*      or may be set differently from one application to another.      */
/*      This function allows these behaviours to be specified by        */
/*      the client.  This helps isolate the module from other parts     */
/*      of the application so that it is easier to reuse.  The          */
/*      default configuration has all options off (all bits 0).         */
/*                                                                      */
/*      The configuration is specified as an off/on pair so that        */
/*      each call can alter some settings without disrupting others.    */
/*                                                                      */
/************************************************************************/
void
Glob_Configure(UINT32 OffConfigs, UINT32 OnConfigs);

/************************************************************************/
/*                                                                      */
/*      Clear -- Erase spec and expanded lists and start again          */
/*                                                                      */
/************************************************************************/
void
Glob_Clear(void);

/************************************************************************/
/*                                                                      */
/*      AddSpec -- Collect pointer to file specification for globbing   */
/*                                                                      */
/*      The client should call this function for each file specific-    */
/*      ation to be globbed.  The function returns FALSE if it was      */
/*      unable to store the specification.                              */
/*                                                                      */
/*      The memory containing the string text must be preserved         */
/*      by the client, as Glob merely remembers the pointer.            */
/*                                                                      */
/************************************************************************/
BOOL
Glob_AddSpec(CHAR *pFileSpec);

/************************************************************************/
/*                                                                      */
/*      Expand -- Compile list of files from specification(s)           */
/*                                                                      */
/*      Once the list of file specifications is complete, the client    */
/*      must call expand to expand the list into file names.  The       */
/*      function returns the number of files found.                     */
/*                                                                      */
/************************************************************************/
UINT
Glob_Expand(void);

/************************************************************************/
/*                                                                      */
/*      EnumerateFiles -- Call client's function for each file found    */
/*                                                                      */
/*      During the enumeration, the client function specified here      */
/*      should return TRUE if the enumeration is to continue or         */
/*      FALSE if the enumeration is to be abandoned.                    */
/*                                                                      */
/************************************************************************/
void
Glob_EnumerateFiles(GLOB_ENUM_FUNCTION *pClientFunc);

#ifdef __cplusplus
}
#endif /*__cplusplus*/

#endif /*GLOB_H*/
