/************************************************************************/
/*                                                                      */
/*      wc -- Count words, lines and bytes of (text) files              */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include <ascii.h>
#include <compdef.h>
#include <fcntl.h>
#include <glob.h>
#include <io.h>
#include <limits.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include "wc.h"

#define PAGE_ALIGNED            far
#define ENDMARKER               0xEE

#define EE_NOP                  0x00    /*Buf end check, NOP if not*/
#define NOP                     0x04    /*Stay in same state*/
#define EE_WORD                 0x08    /*Buf end check, WORD if not*/
#define WORD                    0x0c    /*Words++, change to word*/
#define WHITE                   0x14    /*Change to whitespace*/
#define LF_WHITE                0x1a    /*Lines++, change to whitespace*/
#define LF_NOP                  0x1c    /*Lines++, stay in same state*/

#define BUF_SIZE                (28*2048u)
#define BYTE_SIZE               (UCHAR_MAX + 1)

/*Module options*/
#define WC_O_LINES              BIT0
#define WC_O_WORDS              BIT1
#define WC_O_BYTES              BIT2

/*Global variables updated directly by counting loop*/
unsigned long Lines;
unsigned long Words;
BYTE *pCurrentTable;

typedef struct {

        /*State tables for finite-state machine*/
        BYTE Tables[BYTE_SIZE * 2];

        /*Buffer to hold file in memory*/
        BYTE Buf[BUF_SIZE + 2];

        /*Grand totals*/
        UINT NrFiles;
        UINT32 BytesGrand;
        UINT32 WordsGrand;
        UINT32 LinesGrand;

        /*Reporting options*/
        UINT32 Opt;

        /*File globbing options*/
        UINT32 GlobOpt;

        /*Function to read bytes from file*/
        Wc_ReadFn *pReadFn;

} WC_MODULE_CONTEXT;

module_scope WC_MODULE_CONTEXT gWc;

#define WhiteTable      (&gWc.Tables[0])
#define WordTable       (&gWc.Tables[BYTE_SIZE])

/*C and assembly versions of word counter*/
void PAGE_ALIGNED CountC(UINT16 NrBytes, BYTE *pText);
void PAGE_ALIGNED CountAsm(UINT16 NrBytes, BYTE *pText);

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
Wc_Init(void)
{
        /*Initialise whitespace state table*/
        memset(WhiteTable, WORD, BYTE_SIZE);
        WhiteTable[ 9] = WhiteTable[11] = WhiteTable[12] = NOP;
        WhiteTable[13] = WhiteTable[32] = NOP;
        WhiteTable[10] = LF_NOP;
        WhiteTable[ENDMARKER] = EE_WORD;

        /*Initialise word state table*/
        memset(WordTable, NOP, BYTE_SIZE);
        WordTable[ 9]  = WordTable[11]  = WordTable[12]  = WHITE;
        WordTable[13]  = WordTable[32]  = WHITE;
        WordTable[10]  = LF_WHITE;
        WordTable[ENDMARKER] = EE_NOP;

        /*Default to reporting nothing*/
        gWc.Opt = 0;

        /*Default to simple globbing*/
        gWc.GlobOpt = 0;

        /*Default to ussing ANSI-compatible read function*/
        gWc.pReadFn = read;

} /*Init*/

/************************************************************************/
/*                                                                      */
/*      Start -- Begin managing what has to be managed                  */
/*                                                                      */
/************************************************************************/
public_scope void
Wc_Start(void)
{
} /*Start*/

/************************************************************************/
/*                                                                      */
/*      Usage -- Tell client how to use program                         */
/*                                                                      */
/************************************************************************/
module_scope void
Wc_Usage(void)
{
        fprintf(stderr, "Usage: wc [-cwldV] <files>...\n");
        exit(1);

} /*Usage*/

/************************************************************************/
/*                                                                      */
/*      Version -- Report program version to stdout                     */
/*                                                                      */
/************************************************************************/
module_scope void
Wc_Version(void)
{
        printf(
"wc 1.01 Copyright (C) 1995-7 Grouse Software.  All rights reserved.\n");
        printf(
"Written for Grouse by behoffski (Brenton Hoff).\n");

        exit(1);

} /*Version*/

/************************************************************************/
/*                                                                      */
/*      Report -- Display statistics according to selection             */
/*                                                                      */
/************************************************************************/
module_scope void
Wc_Report(CHAR *pName, UINT32 Bytes, UINT32 Words, UINT32 Lines)
{
        if (gWc.Opt & WC_O_LINES) {
                printf("%7lu ", Lines);
        }
        if (gWc.Opt & WC_O_WORDS) {
                printf("%7lu ", Words);
        }
        if (gWc.Opt & WC_O_BYTES) {
                printf("%7lu ", Bytes);
        }

        printf("%s\n", pName);

} /*Report*/

/************************************************************************/
/*                                                                      */
/*      File -- Perform count on specified file                         */
/*                                                                      */
/************************************************************************/
module_scope BOOL
Wc_File(char *pFilename) {
        UINT NrBytes;
        int Handle;
        UINT32 Bytes;

        /*Open file or stdin if specified*/
        if (strcmp(pFilename, "-") == 0) {
                /*Open stdin for counting*/
                Handle = fileno(stdin);

        } else {
                /*Attempt to open specified file*/
                Handle = open(pFilename, O_BINARY | O_RDONLY);
                if (Handle == -1) {
                        printf("\nCan't open: %s", pFilename);
                        return 1;
                }

        }

        Lines = 0;
        Words = 0;
        Bytes = 0;
        pCurrentTable = WhiteTable;

        for (;;) {
                /*Read a slab of the file into memory*/
                /*(_read used because read in TSC v3.10 is poor on pipes)*/

                NrBytes = (UINT) gWc.pReadFn(Handle, gWc.Buf, BUF_SIZE);

                /*Exit loop if there's an error or if we're finished*/
                if ((NrBytes == 0) || (((INT) NrBytes) == -1)) {
                        break;
                }

                /*Process the slab using CountC or CountAsm*/
                CountAsm(NrBytes, gWc.Buf);
                /*CountC(NrBytes, gWc.Buf);*/

                Bytes += (UINT32) NrBytes;

        }

        Wc_Report(pFilename, Bytes, Words, Lines);
        (void) close(Handle);

        /*Accumulate grand totals*/
        gWc.NrFiles++;
        gWc.BytesGrand += Bytes;
        gWc.WordsGrand += Words;
        gWc.LinesGrand += Lines;

        /*Globbing should keep going*/
        return 1;

} /*File*/

/************************************************************************/
/*                                                                      */
/*      main -- Supervise parsing and serving request                   */
/*                                                                      */
/************************************************************************/
public_scope int
Wc_main(int argc, char **argv)
{
        CHAR *pArg;
        CHAR *pOpt;
        BOOL CountStdin = FALSE;

        /*Skip past exe name*/
        argc--;
        argv++;

        for (;;) {
                /*Have we run out of options?*/
                if (argc-- == 0) {
                       /*Yes, finished interpreting parameters*/
                       break;
                }

                pArg = *argv++;

                /*Is this parameter an option?*/
                if (pArg[0] != '-') {
                        /*No, it's a filespec to be globbed*/
                        Glob_AddSpec(pArg);
                        continue;
                }

                /*Decode options*/
                pOpt = &pArg[1];

                /*Was '-' given by itself?*/
                if (*pOpt == NUL) {
                        /*Yes, remember to look at stdin*/
                        CountStdin = TRUE;
                        continue;
                }

                for (;;) {
                        /*Have we found the end of the current word?*/
                        if (*pOpt == NUL) {
                                /*Yes, move to next word*/
                                break;
                        }

                        /*Decode option*/
                        switch (*pOpt++) {
                        case 'c':
                                gWc.Opt |= WC_O_BYTES;
                                break;

                        case 'd':
                                gWc.GlobOpt |= GLOB_INCLUDE_SUBDIRECTORIES;
                                break;

                        case 'l':
                                gWc.Opt |= WC_O_LINES;
                                break;

                        case 'V':
                                Wc_Version();
                                break;

                        case 'w':
                                gWc.Opt |= WC_O_WORDS;
                                break;

                        default:
                                fprintf(stderr,
                                              "wc: Unrecognised option: %c\n",
                                              pOpt[-1]);
                                Wc_Usage();
                                break;

                        }

                }

        }

        /*Were any counts specified by parameters?*/
        if (gWc.Opt == 0) {
                /*No, default to reporting lines/words/bytes*/
                gWc.Opt = WC_O_LINES | WC_O_WORDS | WC_O_BYTES;
        }

        /*Configure globbing*/
        Glob_Configure(0, gWc.GlobOpt);

        /*Expand specs and complain if none found*/
        if (Glob_Expand() == 0) {
                /*No files found for counting: search stdin*/
                CountStdin = TRUE;
        }

        /*Was stdin specified?*/
        if (CountStdin) {
                /*Yes, count it first*/
                Wc_File("-");
        }

        Glob_EnumerateFiles(Wc_File);

        /*Were there multiple files processed?*/
        if (gWc.NrFiles > 1) {
                /*Yes, include a grand total*/
                Wc_Report("total", gWc.BytesGrand,
                                   gWc.WordsGrand, gWc.LinesGrand);
        }

        return 0;

} /*main*/

/************************************************************************/
/*                                                                      */
/*      CountC -- C version of finite-state machine (for port/debug)    */
/*                                                                      */
/************************************************************************/
void PAGE_ALIGNED
CountC(UINT16 NrBytes, BYTE *pText)
{
        BYTE NextCh;
        BYTE Action;
        BYTE *pTab = pCurrentTable;
        BYTE *pEnd = &pText[NrBytes];

        /*Add endmarker to buffer to trigger end test below*/
        *pEnd++ = ENDMARKER;

        /*Loop through each byte*/
        for (;;) {
            NextCh = *pText++;
            Action = pTab[NextCh];
            switch (Action) {
            case EE_NOP:   if (pText == pEnd) break;
                           /*FALLTHROUGH*/
            case NOP:      continue;

            case EE_WORD:  if (pText == pEnd) break;
                           /*FALLTHROUGH*/
            case WORD:     Words++;
                           pTab = WordTable;
                           continue;

            case WHITE:    pTab = WhiteTable;
                           continue;

            case LF_WHITE: pTab = WhiteTable;
                           /*FALLTHROUGH*/
            case LF_NOP:   Lines++;
                           continue;
            }

            /*End of buffer found: remember in-word context and exit*/
            pCurrentTable = pTab;
            return;
        }
    
} /*CountC*/

/************************************************************************/
/*                                                                      */
/*      ReadFunction -- Receive function to read bytes from a file      */
/*                                                                      */
/*      wc assumes that the ANSI read function is sufficient            */
/*      and correct to fetch file lines.  If this assumption is         */
/*      incorrect for some reason, the client may supply an             */
/*      alternative function via this interface.                        */
/*                                                                      */
/*      This interface was included when behoffski found that           */
/*      the operation of read for pipes was incorrect,                  */
/*      whereas _read worked correctly under TopSpeed C v3.10.          */
/*                                                                      */
/************************************************************************/
public_scope void
Wc_ReadFunction(Wc_ReadFn *pReadFn)
{
        /*Record function for use while working*/
        gWc.pReadFn = pReadFn;

} /*ReadFunction*/
