/************************************************************************/
/*                                                                      */
/*      MatchC -- ANSI C version of match engine                        */
/*                                                                      */
/*      This module implements the table-driven search using ANSI C     */
/*      constructs.  It is invaluable for debugging state table         */
/*      generation and for prototyping match engine actions.            */
/*      The performance will usually be very good for a C version,      */
/*      but will not come close to an implementation exploiting         */
/*      the Grouse Finite State Architecture in assembly.               */
/*                                                                      */
/*      Copyright (C) Grouse Software 1995-7.  All rights reserved.     */
/*      Written for Grouse by behoffski (Brenton Hoff).                 */
/*                                                                      */
/*      Free software: no warranty; use anywhere is ok; spread the      */
/*      sources; note any mods; share variations and derivatives        */
/*      (including sending to behoffski@grouse.com.au).                 */
/*                                                                      */
/************************************************************************/

#include <ctype.h>
#include "ascii.h"
#include <compdef.h>
#include "matcheng.h"
#include "matchc.h"
#include "memlchr.h"
#include <memory.h>
#include <stdio.h>

#define BACKTRACK_STACK_SIZE            200 /*pTab/pText pairs for C search*/
#define STROPT_STATES_MAX               17
#define RE_STATES_MAX                   200

typedef struct {
        /*Flag describing what information to report*/
        LWORD Debug;

} MATCHC_MODULE_CONTEXT;

module_scope MATCHC_MODULE_CONTEXT gMatchC;

#if DEBUGGING
#define DEBUG(Item, Msg)           {if (gMatchC.Debug & Item) {printf(Msg);}}
#define DEBUG1(Item, Msg, p1)      {if (gMatchC.Debug & Item) {printf(Msg, p1);}}
#define DEBUG2(Item, Msg, p1, p2)  {if (gMatchC.Debug & Item) {printf(Msg, p1, p2);}}
#else
#define DEBUG(Item, Msg)
#define DEBUG1(Item, Msg, p1)
#define DEBUG2(Item, Msg, p1, p2)
#endif /*DEBUGGING*/

/************************************************************************/
/*                                                                      */
/*      Init -- Prepare module for operation                            */
/*                                                                      */
/************************************************************************/
public_scope void
MatchC_Init(void)
{
        /*Don't display any debug information*/
        gMatchC.Debug = 0;

} /*Init*/

/************************************************************************/
/*                                                                      */
/*      Match -- Perform search and report success                      */
/*                                                                      */
/************************************************************************/
public_scope BOOL PAGE_ALIGNED
MatchC_Match(MATCHENG_SPEC *pSpec, CHAR *pTextStart,
                MATCHENG_DETAILS *pDetails)
{
        /*Stack area for storing search context*/
        CHAR *pText = pTextStart;
        BYTE *pTab = pSpec->Tables[0];
        void *Stack[BACKTRACK_STACK_SIZE * 2];
        void **pStack = &Stack[(BACKTRACK_STACK_SIZE * 2) - 1];
        CHAR *pStart = pText;
        UINT LineNr = 0;
        UINT StrOptOffset;

        /*Pick up FixedString search offset in case optimisation used*/
        StrOptOffset = pSpec->SearchSkipOffset;

        /*Push failure in case search attempts to backtrack past start*/
        *--pStack = pTab;
        *--pStack = pTextStart;
        pTab += TABLE_SIZE;

        /*Unrolled_switch approximates what the assembly is doing*/

        #define unrolled_switch                 do { switch
        #define end_unrolled_switch             } while (TRUE)
        #define continue_switch                 continue

        /*Get next buffer character and act according to current state*/
        unrolled_switch (pTab[(UCHAR) *pText++]) { /*lodsb; xlat; jmp ax*/
        case SKIP_MAX: pText += StrOptOffset;
                         DEBUG(MATCHC_D_ACTION, " _SkipMax"); continue_switch;
        case SKIP17: pText += 16;
                         DEBUG(MATCHC_D_ACTION, " _Skip17"); continue_switch;
        case SKIP16: pText += 15;
                         DEBUG(MATCHC_D_ACTION, " _Skip16"); continue_switch;
        case SKIP15: pText += 14;
                         DEBUG(MATCHC_D_ACTION, " _Skip15"); continue_switch;
        case SKIP14: pText += 13;
                         DEBUG(MATCHC_D_ACTION, " _Skip14"); continue_switch;
        case SKIP13: pText += 12;
                         DEBUG(MATCHC_D_ACTION, " _Skip13"); continue_switch;
        case SKIP12: pText += 11;
                         DEBUG(MATCHC_D_ACTION, " _Skip12"); continue_switch;
        case SKIP11: pText += 10;
                         DEBUG(MATCHC_D_ACTION, " _Skip11"); continue_switch;
        case SKIP10: pText +=  9;
                         DEBUG(MATCHC_D_ACTION, " _Skip10"); continue_switch;
        case SKIP9:  pText +=  8;
                         DEBUG(MATCHC_D_ACTION, " _Skip9"); continue_switch;
        case SKIP8:  pText +=  7;
                         DEBUG(MATCHC_D_ACTION, " _Skip8"); continue_switch;
        case SKIP7:  pText +=  6;
                         DEBUG(MATCHC_D_ACTION, " _Skip7"); continue_switch;
        case SKIP6:  pText +=  5;
                         DEBUG(MATCHC_D_ACTION, " _Skip6"); continue_switch;
        case SKIP5:  pText +=  4;
                         DEBUG(MATCHC_D_ACTION, " _Skip5"); continue_switch;
        case SKIP4:  pText +=  3;
                         DEBUG(MATCHC_D_ACTION, " _Skip4"); continue_switch;
        case SKIP3:  pText +=  2;
                         DEBUG(MATCHC_D_ACTION, " _Skip3"); continue_switch;
        case SKIP2:  pText +=  1;
                         DEBUG(MATCHC_D_ACTION, " _Skip2"); continue_switch;
        case AGAIN:
                /*Keep trying with same table*/
                DEBUG(MATCHC_D_ACTION, " _Again");

                /*Continue with (unrolled) next do{switch component*/
                continue_switch;

        case ADVANCE:
                /*Matched; move to next table*/
                DEBUG(MATCHC_D_ACTION, " _Advance");
                pTab += TABLE_SIZE;

                continue_switch;

        case START_MATCH:
                /*Save match start and attempt RE match*/
                DEBUG(MATCHC_D_ACTION, " _Start");
                pStart = pText;
                pText--;
                pTab += TABLE_SIZE;

                continue_switch;

        case START_MATCH_PUSH:
                /*Save match start and push backtrack of next char*/
                DEBUG(MATCHC_D_ACTION, " _StartPush");
                *--pStack = pTab;
                *--pStack = pText;

                pStart = pText;
                pText--;
                pTab += TABLE_SIZE;

                continue_switch;

        case START_MATCH_PUSH_ADVANCE:
                /*START_MATCH_PUSH and include AGAIN to 2nd state*/
                DEBUG(MATCHC_D_ACTION, " _StartPushAdvance");
                *--pStack = pTab;
                *--pStack = pText;

                pStart = pText;
                pTab += TABLE_SIZE * 2;

                continue_switch;

        case AGAIN_PUSH_ADVANCE:
                /*Save "backtrack" of next state, and repeat current state*/
                DEBUG(MATCHC_D_ACTION, " _AgainPushAdvance");
                *--pStack = pTab + TABLE_SIZE;
                *--pStack = pText - 1;

                continue_switch;

        case BACK_AND_ADVANCE:
                /*Move text back 1 char and advance to next table*/
                DEBUG(MATCHC_D_ACTION, " _BackAndAdvance");
                pText--;
                pTab += TABLE_SIZE;

                continue_switch;

        case ADVANCE_PUSH_ZERO:
                /*Next table but allow match to be skipped via backtrack*/
                DEBUG(MATCHC_D_ACTION, " _AdvancePushZero");
                pTab += TABLE_SIZE;
                *--pStack = pTab;
                *--pStack = pText - 1;

                continue_switch;

        case NO_MATCH:
                /*No match: pop to previous state*/
                DEBUG(MATCHC_D_ACTION, " _NoMatch");
                pText = *pStack++;
                pTab = *pStack++;

                continue_switch;

        case MARK_LINE_START:
                /*Remember line start for buffer search*/
                DEBUG(MATCHC_D_ACTION, " _MarkLineStart");
                pDetails->pLineStart = pText;
                LineNr++;

                continue_switch;

        case MARK_LINE_START_PUSH_ADVANCE:
                /*Remember line start and attempt anchored match*/
                DEBUG(MATCHC_D_ACTION, " _MarkLineStartPushAdvance");
                pDetails->pLineStart = pText;
                pStart = pText + 1;
                LineNr++;
                *--pStack = pTab;
                *--pStack = pText;
                pTab += TABLE_SIZE;

                continue_switch;

        case START_OFFSET_MATCH:
                /*Save otimisation search point and attempt RE match*/
                DEBUG(MATCHC_D_ACTION, " _StartOffsetMatch");
                *--pStack = pTab;
                *--pStack = pText;
                pText -= StrOptOffset + 1;
                pStart = pText + 1;
                pTab += TABLE_SIZE;

                continue_switch;

        case BYTE_SEARCH:
                /*Use inbuilt search function as it's faster*/
                DEBUG(MATCHC_D_ACTION, " _ByteSearch");
                pText = ((CHAR *) memchr(pText, StrOptOffset, 0xffff)) + 1;
                if (pText == pDetails->pAfterEndOfBuffer) {
                        /*Abandon search (line count is not valid)*/
                        return FALSE;
                }

                /*First byte found, try to match remainder*/
                pStart = pText;
                *--pStack = pTab;
                *--pStack = pText;
                pTab += TABLE_SIZE * 2;

                continue_switch;

        case ABANDON:
                /*Buffer exhausted without finding match*/
                DEBUG(MATCHC_D_ACTION, " _Abandon");

                /*Write count of lines seen so these are tracked correctly*/
                pDetails->BufLineNr = LineNr - 1;

                return FALSE;
                break;

        case COMPLETE_IF_WORD_START:
                /*Check for word start, and exit with success if found*/
                DEBUG(MATCHC_D_ACTION, " ChkWordStart");

                /*Is there a word character before the match start?*/
                if ((pStart[-2] == '_') || isalnum(pStart[-2]) ) {
                        /*Yes, match cannot start in middle of word*/
                        pText = *pStack++;
                        pTab = *pStack++;

                        continue_switch;

                }

                if (pStart[-2] <  '0') {
                        goto Completed;
                }
                if (pStart[-2] <= '9') {
                        continue_switch;
                }
                if (tolower(pStart[-2]) <  'a') {
                        goto Completed;
                }
                if (tolower(pStart[-2]) <= 'z') {
                        continue_switch;
                }

                /*Match is complete: write to caller and exit*/
                goto Completed;
                break;

        case COMPLETED:
Completed:
                DEBUG(MATCHC_D_ACTION, "\nYay! completed");

                /*Write details of match to caller*/
                pDetails->pMatchStart = pStart - 1;
                pDetails->pMatchEnd = pText - 1;
                pDetails->BufLineNr = LineNr;

                /*Did we search for match using start optimisation?*/
                if (StrOptOffset != 0) {
                        /*Yes, find start of matching line*/
                        pDetails->pLineStart = ((CHART *) memlchr(pStart,
                                        LF, 0xffff)) + 1;
                }

                return TRUE;
                break;

        default:
                printf("\nRetable_Match: Unknown action: %02x",
                                pTab[*(pText-1)]);
                return FALSE;
                break;

        /*End of unrolled switch statement*/
        } end_unrolled_switch;

        /*Add a return (not reached) to keep compiler happy*/
        return FALSE;

} /*Match*/

/************************************************************************/
/*                                                                      */
/*      Debug -- Set debugging options                                  */
/*                                                                      */
/************************************************************************/
public_scope void
MatchC_Debug(LWORD OffBits, LWORD OnBits)
{
        /*Edit debugging profile according to request*/
        gMatchC.Debug = (gMatchC.Debug & ~OffBits) | OnBits;

} /*Debug*/
